﻿using pdftron.PDF.Tools.Controls.ControlBase;
using System;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;

namespace pdftron.PDF.Tools.Controls
{
    public enum AnnotationToolbarOrientation
    {
        Horizontal,
        Vertical,
    }

    public class AnnotationToolbarBase : UserControl, ICloseableControl
    {
        #region Tools

        /// <summary>
        /// Allows user to externally control whether the StickyNote button is visible or collapsed (since view model is not exposed)
        /// </summary>
        /// 
        public static DependencyProperty IsStickyNoteToolButtonVisibleProperty =
          DependencyProperty.Register("IsStickyNoteToolButtonVisible", typeof(bool),
          typeof(AnnotationCommandBar), new PropertyMetadata(false));

        public bool IsStickyNoteToolButtonVisible
        {
            get
            {
                return (bool)(GetValue(IsStickyNoteToolButtonVisibleProperty));
            }
            set
            {
                SetValue(IsStickyNoteToolButtonVisibleProperty, value);
                _ViewModel.IsStickyNoteToolButtonVisible = value;
            }
        }

        /// <summary>
        /// Allows user to externally control whether the Highlight button is visible or collapsed (since view model is not exposed)
        /// </summary>
        /// 
        public static DependencyProperty IsHighlightToolButtonVisibleProperty =
          DependencyProperty.Register("IsHighlightToolButtonVisible", typeof(bool),
          typeof(AnnotationCommandBar), new PropertyMetadata(false));

        public bool IsHighlightToolButtonVisible
        {
            get
            {
                return (bool)(GetValue(IsHighlightToolButtonVisibleProperty));
            }
            set
            {
                SetValue(IsHighlightToolButtonVisibleProperty, value);
                _ViewModel.IsHighlightToolButtonVisible = value;
            }
        }

        /// <summary>
        /// Allows user to externally control whether the Underline button is visible or collapsed (since view model is not exposed)
        /// </summary>
        /// 
        public static DependencyProperty IsUnderlineToolButtonVisibleProperty =
          DependencyProperty.Register("IsUnderlineToolButtonVisible", typeof(bool),
          typeof(AnnotationCommandBar), new PropertyMetadata(false));

        public bool IsUnderlineToolButtonVisible
        {
            get
            {
                return (bool)(GetValue(IsUnderlineToolButtonVisibleProperty));
            }
            set
            {
                SetValue(IsUnderlineToolButtonVisibleProperty, value);
                _ViewModel.IsUnderlineToolButtonVisible = value;
            }
        }

        /// <summary>
        /// Allows user to externally control whether the Strikeout button is visible or collapsed (since view model is not exposed)
        /// </summary>
        /// 
        public static DependencyProperty IsStrikeoutToolButtonVisibleProperty =
          DependencyProperty.Register("IsStrikeoutToolButtonVisible", typeof(bool),
          typeof(AnnotationCommandBar), new PropertyMetadata(false));

        public bool IsStrikeoutToolButtonVisible
        {
            get
            {
                return (bool)(GetValue(IsStrikeoutToolButtonVisibleProperty));
            }
            set
            {
                SetValue(IsStrikeoutToolButtonVisibleProperty, value);
                _ViewModel.IsStrikeoutToolButtonVisible = value;
            }
        }

        /// <summary>
        /// Allows user to externally control whether the Squiggly button is visible or collapsed (since view model is not exposed)
        /// </summary>
        /// 
        public static DependencyProperty IsSquigglyToolButtonVisibleProperty =
          DependencyProperty.Register("IsSquigglyToolButtonVisible", typeof(bool),
          typeof(AnnotationCommandBar), new PropertyMetadata(false));

        public bool IsSquigglyToolButtonVisible
        {
            get
            {
                return (bool)(GetValue(IsSquigglyToolButtonVisibleProperty));
            }
            set
            {
                SetValue(IsSquigglyToolButtonVisibleProperty, value);
                _ViewModel.IsSquigglyToolButtonVisible = value;
            }
        }

        /// <summary>
        /// Allows user to externally control whether the Stamper button is visible or collapsed (since view model is not exposed)
        /// </summary>
        /// 
        public static DependencyProperty IsStamperToolButtonVisibleProperty =
          DependencyProperty.Register("IsStamperToolButtonVisible", typeof(bool),
          typeof(AnnotationCommandBar), new PropertyMetadata(false));

        public bool IsStamperToolButtonVisible
        {
            get
            {
                return (bool)(GetValue(IsStamperToolButtonVisibleProperty));
            }
            set
            {
                SetValue(IsStamperToolButtonVisibleProperty, value);
                _ViewModel.IsStamperToolButtonVisible = value;
            }
        }

        /// <summary>
        /// Allows user to externally control whether the Signature button is visible or collapsed (since view model is not exposed)
        /// </summary>
        /// 
        public static DependencyProperty IsSignatureToolButtonVisibleProperty =
          DependencyProperty.Register("IsSignatureToolButtonVisible", typeof(bool),
          typeof(AnnotationCommandBar), new PropertyMetadata(false));

        public bool IsSignatureToolButtonVisible
        {
            get
            {
                return (bool)(GetValue(IsSignatureToolButtonVisibleProperty));
            }
            set
            {
                SetValue(IsSignatureToolButtonVisibleProperty, value);
                _ViewModel.IsSignatureToolButtonVisible = value;
            }
        }

        /// <summary>
        /// Allows user to externally control whether the Ink button is visible or collapsed (since view model is not exposed)
        /// </summary>
        /// 
        public static DependencyProperty IsInkToolButtonVisibleProperty =
          DependencyProperty.Register("IsInkToolButtonVisible", typeof(bool),
          typeof(AnnotationCommandBar), new PropertyMetadata(false));

        public bool IsInkToolButtonVisible
        {
            get
            {
                return (bool)(GetValue(IsInkToolButtonVisibleProperty));
            }
            set
            {
                SetValue(IsInkToolButtonVisibleProperty, value);
                _ViewModel.IsInkToolButtonVisible = value;
            }
        }

        /// <summary>
        /// Allows user to externally control whether the Eraser button is visible or collapsed (since view model is not exposed)
        /// </summary>
        /// 
        public static DependencyProperty IsEraserToolButtonVisibleProperty =
          DependencyProperty.Register("IsEraserToolButtonVisible", typeof(bool),
          typeof(AnnotationCommandBar), new PropertyMetadata(false));

        public bool IsEraserToolButtonVisible
        {
            get
            {
                return (bool)(GetValue(IsEraserToolButtonVisibleProperty));
            }
            set
            {
                SetValue(IsEraserToolButtonVisibleProperty, value);
                _ViewModel.IsEraserToolButtonVisible = value;
            }
        }

        /// <summary>
        /// Allows user to externally control whether the Marker button is visible or collapsed (since view model is not exposed)
        /// </summary>
        /// 
        public static DependencyProperty IsMarkerToolButtonVisibleProperty =
          DependencyProperty.Register("IsMarkerToolButtonVisible", typeof(bool),
          typeof(AnnotationCommandBar), new PropertyMetadata(false));

        public bool IsMarkerToolButtonVisible
        {
            get
            {
                return (bool)(GetValue(IsMarkerToolButtonVisibleProperty));
            }
            set
            {
                SetValue(IsMarkerToolButtonVisibleProperty, value);
                _ViewModel.IsMarkerToolButtonVisible = value;
            }
        }

        /// <summary>
        /// Allows user to externally control whether the TextBox button is visible or collapsed (since view model is not exposed)
        /// </summary>
        /// 
        public static DependencyProperty IsTextBoxToolButtonVisibleProperty =
          DependencyProperty.Register("IsTextBoxToolButtonVisible", typeof(bool),
          typeof(AnnotationCommandBar), new PropertyMetadata(false));

        public bool IsTextBoxToolButtonVisible
        {
            get
            {
                return (bool)(GetValue(IsTextBoxToolButtonVisibleProperty));
            }
            set
            {
                SetValue(IsTextBoxToolButtonVisibleProperty, value);
                _ViewModel.IsTextBoxToolButtonVisible = value;
            }
        }

        /// <summary>
        /// Allows user to externally control whether the Shape tools button is visible or collapsed (since view model is not exposed)
        /// </summary>
        /// 
        public static DependencyProperty IsShapeToolsButtonVisibleProperty =
          DependencyProperty.Register("IsShapeToolsButtonVisible", typeof(bool),
          typeof(AnnotationCommandBar), new PropertyMetadata(false));

        public bool IsShapeToolsButtonVisible
        {
            get
            {
                return (bool)(GetValue(IsShapeToolsButtonVisibleProperty));
            }
            set
            {
                SetValue(IsShapeToolsButtonVisibleProperty, value);
                _ViewModel.IsShapeToolsButtonVisible = value;
            }
        }

        /// <summary>
        /// Allows user to externally control whether the Pan button is visible or collapsed (since view model is not exposed)
        /// </summary>
        /// 
        public static DependencyProperty IsPanToolsButtonVisibleProperty =
          DependencyProperty.Register("IsPanToolsButtonVisible", typeof(bool),
          typeof(AnnotationCommandBar), new PropertyMetadata(false));

        public bool IsPanToolsButtonVisible
        {
            get
            {
                return (bool)(GetValue(IsPanToolsButtonVisibleProperty));
            }
            set
            {
                SetValue(IsPanToolsButtonVisibleProperty, value);
                _ViewModel.IsPanToolsButtonVisible = value;
            }
        }

        /// <summary>
        /// Allows user to externally control whether the Multi-Select tool button is visible or collapsed (since view model is not exposed)
        /// </summary>
        /// 
        public static DependencyProperty IsMultiSelectToolButtonVisibleProperty =
          DependencyProperty.Register("IsMultiSelectToolButtonVisible", typeof(bool),
          typeof(AnnotationCommandBar), new PropertyMetadata(false));

        public bool IsMultiSelectToolButtonVisible
        {
            get
            {
                return (bool)(GetValue(IsMultiSelectToolButtonVisibleProperty));
            }
            set
            {
                SetValue(IsMultiSelectToolButtonVisibleProperty, value);
                _ViewModel.IsMultiSelectToolButtonVisible = value;
            }
        }

        /// <summary>
        /// Allows user to externally control whether the Undo and Redo buttons are visible or collapsed (since view model is not exposed)
        /// </summary>
        /// 
        public static DependencyProperty IsUndoRedoToolButtonVisibleProperty =
          DependencyProperty.Register("IsUndoRedoToolButtonVisible", typeof(bool),
          typeof(AnnotationCommandBar), new PropertyMetadata(false));

        public bool IsUndoRedoToolButtonVisible
        {
            get
            {
                return (bool)(GetValue(IsUndoRedoToolButtonVisibleProperty));
            }
            set
            {
                SetValue(IsUndoRedoToolButtonVisibleProperty, value);
                _ViewModel.IsUndoRedoToolButtonVisible = value;
            }
        }

        /// <summary>
        /// Allows user to externally control whether the Square button is visible or collapsed (since view model is not exposed)
        /// </summary>
        /// 
        public static DependencyProperty IsSquareToolButtonVisibleProperty =
          DependencyProperty.Register("IsSquareToolButtonVisible", typeof(bool),
          typeof(AnnotationCommandBar), new PropertyMetadata(false));

        public bool IsSquareToolButtonVisible
        {
            get
            {
                return (bool)(GetValue(IsSquareToolButtonVisibleProperty));
            }
            set
            {
                SetValue(IsSquareToolButtonVisibleProperty, value);
                _ViewModel.IsSquareToolButtonVisible = value;
            }
        }

        /// <summary>
        /// Allows user to externally control whether the Circle button is visible or collapsed (since view model is not exposed)
        /// </summary>
        /// 
        public static DependencyProperty IsCircleToolButtonVisibleProperty =
          DependencyProperty.Register("IsCircleToolButtonVisible", typeof(bool),
          typeof(AnnotationCommandBar), new PropertyMetadata(false));

        public bool IsCircleToolButtonVisible
        {
            get
            {
                return (bool)(GetValue(IsCircleToolButtonVisibleProperty));
            }
            set
            {
                SetValue(IsCircleToolButtonVisibleProperty, value);
                _ViewModel.IsCircleToolButtonVisible = value;
            }
        }

        /// <summary>
        /// Allows user to externally control whether the Line button is visible or collapsed (since view model is not exposed)
        /// </summary>
        /// 
        public static DependencyProperty IsLineToolButtonVisibleProperty =
          DependencyProperty.Register("IsLineToolButtonVisible", typeof(bool),
          typeof(AnnotationCommandBar), new PropertyMetadata(false));

        public bool IsLineToolButtonVisible
        {
            get
            {
                return (bool)(GetValue(IsLineToolButtonVisibleProperty));
            }
            set
            {
                SetValue(IsLineToolButtonVisibleProperty, value);
                _ViewModel.IsLineToolButtonVisible = value;
            }
        }

        /// <summary>
        /// Allows user to externally control whether the Arrow button is visible or collapsed (since view model is not exposed)
        /// </summary>
        /// 
        public static DependencyProperty IsArrowToolButtonVisibleProperty =
          DependencyProperty.Register("IsArrowToolButtonVisible", typeof(bool),
          typeof(AnnotationCommandBar), new PropertyMetadata(false));

        public bool IsArrowToolButtonVisible
        {
            get
            {
                return (bool)(GetValue(IsArrowToolButtonVisibleProperty));
            }
            set
            {
                SetValue(IsArrowToolButtonVisibleProperty, value);
                _ViewModel.IsArrowToolButtonVisible = value;
            }
        }

        /// <summary>
        /// Allows user to externally control whether the Polyline button is visible or collapsed (since view model is not exposed)
        /// </summary>
        /// 
        public static DependencyProperty IsPolylineToolButtonVisibleProperty =
          DependencyProperty.Register("IsPolylineToolButtonVisible", typeof(bool),
          typeof(AnnotationCommandBar), new PropertyMetadata(false));

        public bool IsPolylineToolButtonVisible
        {
            get
            {
                return (bool)(GetValue(IsPolylineToolButtonVisibleProperty));
            }
            set
            {
                SetValue(IsPolylineToolButtonVisibleProperty, value);
                _ViewModel.IsPolylineToolButtonVisible = value;
            }
        }

        /// <summary>
        /// Allows user to externally control whether the Polygon button is visible or collapsed (since view model is not exposed)
        /// </summary>
        /// 
        public static DependencyProperty IsPolygonToolButtonVisibleProperty =
          DependencyProperty.Register("IsPolygonToolButtonVisible", typeof(bool),
          typeof(AnnotationCommandBar), new PropertyMetadata(false));

        public bool IsPolygonToolButtonVisible
        {
            get
            {
                return (bool)(GetValue(IsPolygonToolButtonVisibleProperty));
            }
            set
            {
                SetValue(IsPolygonToolButtonVisibleProperty, value);
                _ViewModel.IsPolygonToolButtonVisible = value;
            }
        }

        /// <summary>
        /// Allows user to externally control whether the Cloud button is visible or collapsed (since view model is not exposed)
        /// </summary>
        /// 
        public static DependencyProperty IsCloudToolButtonVisibleProperty =
          DependencyProperty.Register("IsCloudToolButtonVisible", typeof(bool),
          typeof(AnnotationCommandBar), new PropertyMetadata(false));

        public bool IsCloudToolButtonVisible
        {
            get
            {
                return (bool)(GetValue(IsCloudToolButtonVisibleProperty));
            }
            set
            {
                SetValue(IsCloudToolButtonVisibleProperty, value);
                _ViewModel.IsCloudToolButtonVisible = value;
            }
        }

        #endregion

        public static readonly DependencyProperty OrientationProperty =
        DependencyProperty.RegisterAttached("Orientation", typeof(AnnotationToolbarOrientation),
        typeof(AnnotationToolbarBase), new PropertyMetadata(null, OnOrientationPropertyChanged));

        public static void SetOrientation(DependencyObject d, AnnotationToolbarOrientation value)
        {
            d.SetValue(OrientationProperty, value);
        }

        public static AnnotationToolbarOrientation GetOrientation(DependencyObject d)
        {
            return (AnnotationToolbarOrientation)d.GetValue(OrientationProperty);
        }

        private static void OnOrientationPropertyChanged(DependencyObject d,
            DependencyPropertyChangedEventArgs e)
        {
            AnnotationToolbarOrientation oldValue = (AnnotationToolbarOrientation)e.OldValue;
            AnnotationToolbarOrientation newValue = (AnnotationToolbarOrientation)e.NewValue;
            if (newValue != oldValue)
            {
                
            }
        }

        public static readonly DependencyProperty SmallScreenLayoutProperty = 
            DependencyProperty.RegisterAttached("SmallScreenLayout", typeof(bool),
                typeof(AnnotationToolbarBase), new PropertyMetadata(false));

        public bool SmallScreenLayout
        {
            get { return (bool)GetValue(SmallScreenLayoutProperty); }
            set { SetValue(SmallScreenLayoutProperty, value); }
        }


        ViewModels.AnnotationToolbarViewModel _ViewModel;
        public AnnotationToolbarBase() 
        {
            _ViewModel =  new ViewModels.AnnotationToolbarViewModel(this);
            this.DataContext =  _ViewModel;
            this.SizeChanged += AnnotationCommandBar_SizeChanged;
        }

        /// <summary>
        /// Gets or sets whether or not the buttons stay down once pressed.
        /// When set to true, the toolbar will not switch back to the pan tool after each 
        /// annotation is drawn.
        /// </summary>
        public ToolManager ToolManager
        {
            get { return _ViewModel.ToolManager; }
            set { _ViewModel.ToolManager = value; }
        }

        /// <summary>
        /// Gets or sets whether or not the buttons stay down once pressed.
        /// When set to true, the toolbar will not switch back to the pan tool after each 
        /// annotation is drawn.
        /// </summary>
        public bool ButtonsStayDown
        {
            get { return _ViewModel.ButtonsStayDown; }
            set { _ViewModel.ButtonsStayDown = value; }
        }

        public event pdftron.PDF.Tools.Controls.ControlBase.ControlClosedDelegate ControlClosed = delegate { };

        public void CloseControl()
        {
            _ViewModel.IsMainAppBarOpen = false;
            _ViewModel.ShowInkToolbar = false;
            _ViewModel.CurrentTool = ToolType.e_pan;
            ControlClosed();
        }

        protected bool IsButtonHidden(Windows.UI.Xaml.Controls.Primitives.ButtonBase currButtonBase)
        {
            if ((currButtonBase.Name == "AnnotStickyNoteButton" && _ViewModel.IsStickyNoteToolButtonVisible == false)
                || (currButtonBase.Name == "AnnotHighlightButton" && _ViewModel.IsHighlightToolButtonVisible == false)
                || (currButtonBase.Name == "AnnotUnderlineButton" && _ViewModel.IsUnderlineToolButtonVisible == false)
                || (currButtonBase.Name == "AnnotStrikeoutButton" && _ViewModel.IsStrikeoutToolButtonVisible == false)
                || (currButtonBase.Name == "AnnotSquigglyButton" && _ViewModel.IsSquigglyToolButtonVisible == false)
                || (currButtonBase.Name == "AnnotStamperButton" && _ViewModel.IsStamperToolButtonVisible == false)
                || (currButtonBase.Name == "AnnotSignatureButton" && _ViewModel.IsSignatureToolButtonVisible == false)
                || (currButtonBase.Name == "AnnotFreeHandButton" && _ViewModel.IsInkToolButtonVisible == false)
                || (currButtonBase.Name == "AnnotEraserButton" && _ViewModel.IsEraserToolButtonVisible == false)
                || (currButtonBase.Name == "AnnotMarkerButton" && _ViewModel.IsMarkerToolButtonVisible == false)
                || (currButtonBase.Name == "AnnotFreeTextButton" && _ViewModel.IsTextBoxToolButtonVisible == false)
                || (currButtonBase.Name == "AnnotPolygonButton" && _ViewModel.IsShapeToolsButtonVisible == false)
                || (currButtonBase.Name == "AnnotGroupCreateButton" && _ViewModel.IsMultiSelectToolButtonVisible == false)
                || (currButtonBase.Name == "AnnotPanButton" && _ViewModel.IsPanToolsButtonVisible == false)
                || (currButtonBase.Name == "UndoButton" && _ViewModel.IsUndoRedoToolButtonVisible == false)
                || (currButtonBase.Name == "RedoButton" && _ViewModel.IsUndoRedoToolButtonVisible == false)
                || (currButtonBase.Name == "AnnotSquareButton" && _ViewModel.IsSquareToolButtonVisible == false)
                || (currButtonBase.Name == "AnnotCircleButton" && _ViewModel.IsCircleToolButtonVisible == false)
                || (currButtonBase.Name == "AnnotLineButton" && _ViewModel.IsLineToolButtonVisible == false)
                || (currButtonBase.Name == "AnnotArrowButton" && _ViewModel.IsArrowToolButtonVisible == false)
                || (currButtonBase.Name == "AnnotPolylineButton" && _ViewModel.IsPolylineToolButtonVisible == false)
                || (currButtonBase.Name == "AnnotPolygonButton2" && _ViewModel.IsPolygonToolButtonVisible == false)
                || (currButtonBase.Name == "AnnotCloudButton" && _ViewModel.IsCloudToolButtonVisible == false))
            {
                return true;
            }

            return false;
        }

        public event pdftron.PDF.Tools.Controls.ControlBase.ControlClosedDelegate ControlAppBarOpened = delegate { };

        #region resizing

        protected double MINIMUM_BUTTON_WIDTH = 60;
        protected double MINIMUM_BUTTON_WIDTH_SMALL_SCREEN = 50;
        protected Windows.UI.Xaml.Controls.Primitives.ButtonBase[] ToolbarButtonRanking;
        protected System.Collections.Generic.List<Windows.UI.Xaml.Controls.Primitives.ButtonBase> ResizableInkButtons;
        protected System.Collections.Generic.List<Windows.UI.Xaml.Controls.Primitives.ButtonBase> ResizablePolygonSaveButtons;
        protected double ExtraPaddingForInkToolbar = 0;
        protected double ExtraPaddingForPolygonSaveToolbar = 0;
        protected virtual void AnnotationCommandBar_SizeChanged(object sender, SizeChangedEventArgs e)
        {
            ResizeButtons(e.NewSize);
        }

        protected virtual void ResizeButtons(Windows.Foundation.Size newSize)
        {
            if (ToolbarButtonRanking == null)
            {
                return;
            }

            int numButtons = ToolbarButtonRanking.Length;
            double minButtonWidth = MINIMUM_BUTTON_WIDTH;
            if (SmallScreenLayout)
            {
                minButtonWidth = MINIMUM_BUTTON_WIDTH_SMALL_SCREEN;
            }

            int maxButtons = Math.Min((int)((newSize.Width) / minButtonWidth), numButtons);

            double buttonwidth = (newSize.Width) / (maxButtons);
            double remainder = 0;

            for (int i = 0; i < maxButtons; i++)
            {
                double trunkWidth = (int)buttonwidth;
                remainder += buttonwidth - trunkWidth;
                if (remainder >= 1)
                {
                    ++trunkWidth;
                    --remainder;
                }
                var currButtonBase = ToolbarButtonRanking[i];
                if (currButtonBase != null)
                {
                    if (!IsButtonHidden(currButtonBase))
                    {
                        currButtonBase.Visibility = Windows.UI.Xaml.Visibility.Visible;
                    }
                    currButtonBase.Width = (int)trunkWidth;
                }

            }
            for (int i = maxButtons; i < numButtons; i++)
            {
                var currButtonBase = ToolbarButtonRanking[i];
                if (currButtonBase != null)
                {
                    currButtonBase.Visibility = Windows.UI.Xaml.Visibility.Collapsed;
                }
            }
            if (ResizableInkButtons != null)
            {
                int numInkButtons = 6;
                if (!SmallScreenLayout)
                {
                    numInkButtons = 8;
                }
                double totalAvailableWidth = Math.Max(100, newSize.Width - ExtraPaddingForInkToolbar);
                double maxButtonWidth = totalAvailableWidth / numInkButtons;
                if (buttonwidth > maxButtonWidth)
                {
                    buttonwidth = maxButtonWidth;
                }
                foreach (Windows.UI.Xaml.Controls.Primitives.ButtonBase inkButton in ResizableInkButtons)
                {
                    inkButton.Width = buttonwidth;
                }
            }
            if (ResizablePolygonSaveButtons != null)
            {
                int numPolygonSaveButtons = 5;
                double totalAvailableWidth = Math.Max(100, newSize.Width - ExtraPaddingForPolygonSaveToolbar);
                double maxButtonWidth = totalAvailableWidth / numPolygonSaveButtons;
                if (buttonwidth > maxButtonWidth)
                {
                    buttonwidth = maxButtonWidth;
                }
                foreach (Windows.UI.Xaml.Controls.Primitives.ButtonBase polygonSaveButton in ResizablePolygonSaveButtons)
                {
                    polygonSaveButton.Width = buttonwidth;
                }
            }
        }

        #endregion resizing
    }
}
