﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices.WindowsRuntime;
using Windows.Foundation;
using Windows.Foundation.Collections;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Controls.Primitives;
using Windows.UI.Xaml.Data;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Navigation;

using UIPoint = Windows.Foundation.Point;
using UIRect = Windows.Foundation.Rect;

// The User Control item template is documented at http://go.microsoft.com/fwlink/?LinkId=234236

namespace pdftron.PDF.Tools.Controls
{
    public sealed partial class MagnifyingHost : UserControl
    {

        #region Dependency Properties

        public static readonly DependencyProperty MagnifyingElementProperty = DependencyProperty.Register(
          "MagnifyingElement",
          typeof(FrameworkElement),
          typeof(MagnifyingHost),
          new PropertyMetadata(null, new PropertyChangedCallback(OnMagnifyingElementChanged))
        );

        public FrameworkElement MagnifyingElement
        {
            get { return (FrameworkElement)GetValue(MagnifyingElementProperty); }
            set { SetValue(MagnifyingElementProperty, value); }
        }

        private static void OnMagnifyingElementChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            MagnifyingHost host = d as MagnifyingHost;
            if (host != null)
            {
                FrameworkElement newChild = e.NewValue as FrameworkElement;
                host.MagnifyingElementHost.Child = newChild;
            }
        }


        public static readonly DependencyProperty MagnifiedElementProperty = DependencyProperty.Register(
          "MagnifiedElement",
          typeof(FrameworkElement),
          typeof(MagnifyingHost),
          new PropertyMetadata(null, new PropertyChangedCallback(OnMagnifiedElementChanged))
        );

        public FrameworkElement MagnifiedElement
        {
            get { return (FrameworkElement)GetValue(MagnifiedElementProperty); }
            set { SetValue(MagnifiedElementProperty, value); }
        }

        private static void OnMagnifiedElementChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            MagnifyingHost host = d as MagnifyingHost;
            if (host != null)
            {
                FrameworkElement newChild = e.NewValue as FrameworkElement;
                host.MagnifiedElementHost.Child = newChild;
            }
        }


        public static readonly DependencyProperty MagnifyingToolbarDataContextProperty = DependencyProperty.Register(
          "MagnifyingToolbarDataContext",
          typeof(object),
          typeof(MagnifyingHost),
          new PropertyMetadata(null, new PropertyChangedCallback(OnMagnifyingToolbarDataContextChanged))
        );

        public object MagnifyingToolbarDataContext
        {
            get { return GetValue(MagnifyingToolbarDataContextProperty); }
            set { SetValue(MagnifyingToolbarDataContextProperty, value); }
        }

        private static void OnMagnifyingToolbarDataContextChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            MagnifyingHost host = d as MagnifyingHost;
            if (host != null)
            {
                host.MagnifyingToolbarElement.DataContext = e.NewValue;
            }
        }

        public static readonly DependencyProperty MaxMagnifiedHeightRatioProperty = DependencyProperty.Register(
          "MaxMagnifiedHeightRatio",
          typeof(double),
          typeof(MagnifyingHost),
          new PropertyMetadata(0.4)
        );

        /// <summary>
        /// Gets or sets the maximum ratio of the MagnifyingHost that can be occupied by the magnified view 
        /// (Excludes MagnifyingToolbar)
        /// </summary>
        public double MaxMagnifiedHeightRatio
        {
            get { return (double)GetValue(MaxMagnifiedHeightRatioProperty); }
            set { SetValue(MaxMagnifiedHeightRatioProperty, value); }
        }


        public static readonly DependencyProperty MinMagnifiedHeightRatioProperty = DependencyProperty.Register(
            "MinMagnifiedHeightRatio",
            typeof(double),
            typeof(MagnifyingHost),
            new PropertyMetadata(0.0)
        );

        /// <summary>
        /// Gets or sets the minimum ratio of the MagnifyingHost that can be occupied by the magnified view 
        /// (Excludes MagnifyingToolbar)
        /// </summary>
        public double MinMagnifiedHeightRatio
        {
            get { return (double)GetValue(MinMagnifiedHeightRatioProperty); }
            set { SetValue(MinMagnifiedHeightRatioProperty, value); }
        }

        public static readonly DependencyProperty MagnifiedHeightProperty = DependencyProperty.Register(
            "MagnifiedHeight",
            typeof(double),
            typeof(MagnifyingHost),
            new PropertyMetadata(double.NaN, new PropertyChangedCallback(OnMagnifiedHeightChanged))
        );

        /// <summary>
        /// Gets or sets the height of the Magnified part of the MagnifiedHost
        /// </summary>
        public double MagnifiedHeight
        {
            get { return (double)GetValue(MagnifiedHeightProperty); }
            set { SetValue(MagnifiedHeightProperty, value); }
        }

        private static void OnMagnifiedHeightChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            MagnifyingHost host = d as MagnifyingHost;
            if (host != null)
            {
                try
                {
                    double h = (double)e.NewValue;
                    host.MagnifiedSideHost.Height = h;
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
            }
        }

        public static readonly DependencyProperty IsCenterMovableProperty = DependencyProperty.Register(
            "IsCenterMovable",
            typeof(bool),
            typeof(MagnifyingHost),
            new PropertyMetadata(false, new PropertyChangedCallback(OnIsCenterMovableChanged))
        );

        /// <summary>
        /// Gets or sets the height of the Magnified part of the MagnifiedHost
        /// </summary>
        public bool IsCenterMovable
        {
            get { return (bool)GetValue(IsCenterMovableProperty); }
            set { SetValue(IsCenterMovableProperty, value); }
        }

        private static void OnIsCenterMovableChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            MagnifyingHost host = d as MagnifyingHost;
            if (host != null)
            {
                try
                {
                    bool newValue = (bool)e.NewValue;
                    if (newValue)
                    {
                        host.SubscribeToDragEvents();
                    }
                    else
                    {
                        host.UnsubscribeToDragEvents();
                    }
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
            }
        }

        #endregion Dependency Properties

        private bool _IsSubscribedToDragEvents = false;
        private void UnsubscribeToDragEvents()
        {
            if (_IsSubscribedToDragEvents)
            {
                MagnifyingToolbarElement.DragTarget.PointerEntered -= DragTarget_PointerEntered;
                MagnifyingToolbarElement.DragTarget.PointerExited -= DragTarget_PointerExited;
                MagnifyingToolbarElement.DragTarget.PointerPressed -= DragTarget_PointerPressed;
                MagnifyingToolbarElement.DragTarget.PointerMoved -= DragTarget_PointerMoved;
                MagnifyingToolbarElement.DragTarget.PointerReleased -= DragTarget_PointerReleased;
                MagnifyingToolbarElement.DragTarget.PointerCanceled -= DragTarget_PointerCanceled;
                MagnifyingToolbarElement.DragTarget.PointerCaptureLost -= DragTarget_PointerCaptureLost;
            }
            _IsSubscribedToDragEvents = false;
        }

        private void SubscribeToDragEvents()
        {
            MagnifyingToolbarElement.DragTarget.PointerEntered += DragTarget_PointerEntered;
            MagnifyingToolbarElement.DragTarget.PointerExited += DragTarget_PointerExited;
            MagnifyingToolbarElement.DragTarget.PointerPressed += DragTarget_PointerPressed;
            MagnifyingToolbarElement.DragTarget.PointerMoved += DragTarget_PointerMoved;
            MagnifyingToolbarElement.DragTarget.PointerReleased += DragTarget_PointerReleased;
            MagnifyingToolbarElement.DragTarget.PointerCanceled += DragTarget_PointerCanceled;
            MagnifyingToolbarElement.DragTarget.PointerCaptureLost += DragTarget_PointerCaptureLost;
            _IsSubscribedToDragEvents = true;
        }


        public MagnifyingHost()
        {
            this.InitializeComponent();
            _ContactPoints = new Dictionary<uint, Pointer>();
            _EnteredMousePointers = new Dictionary<uint, Pointer>();
            _PointersWithDragShape = new Dictionary<uint, Pointer>();
        }


        #region Interaction

        private Dictionary<uint, Pointer> _ContactPoints;
        private Dictionary<uint, Pointer> _EnteredMousePointers;
        private Dictionary<uint, Pointer> _PointersWithDragShape;
        private bool _InteractionOngoing = false;
        private UIPoint _DownPoint;
        private double _HeightWhenPressing = 0;


        private void DragTarget_PointerEntered(object sender, PointerRoutedEventArgs e)
        {
            if (e.Pointer.PointerDeviceType == Windows.Devices.Input.PointerDeviceType.Mouse)
            {
                _EnteredMousePointers[e.Pointer.PointerId] = e.Pointer;
            }
            UpdatePointerAppearance(e.Pointer);
        }

        private void DragTarget_PointerExited(object sender, PointerRoutedEventArgs e)
        {
            if (e.Pointer.PointerDeviceType == Windows.Devices.Input.PointerDeviceType.Mouse && _EnteredMousePointers.ContainsKey(e.Pointer.PointerId))
            {
                _EnteredMousePointers.Remove(e.Pointer.PointerId);
            }
            UpdatePointerAppearance(e.Pointer);
        }

        private void DragTarget_PointerPressed(object sender, PointerRoutedEventArgs e)
        {
            if (_ContactPoints.Count > 0)
            {
                _InteractionOngoing = false;
            }
            _ContactPoints[e.Pointer.PointerId] = e.Pointer;
            MagnifyingToolbarElement.DragTarget.CapturePointer(e.Pointer);
            if (_ContactPoints.Count == 1)
            {
                _InteractionOngoing = true;
                _HeightWhenPressing = MagnifiedHeight;
                _DownPoint = e.GetCurrentPoint(this).Position;
            }
        }

        private void DragTarget_PointerMoved(object sender, PointerRoutedEventArgs e)
        {
            if (_InteractionOngoing && _ContactPoints.Count == 1 && _ContactPoints.ContainsKey(e.Pointer.PointerId))
            {
                double newHeight = _HeightWhenPressing + _DownPoint.Y - e.GetCurrentPoint(this).Position.Y;
                if (newHeight > MaxMagnifiedHeightRatio * ActualHeight)
                {
                    newHeight = MaxMagnifiedHeightRatio * ActualHeight;
                }
                if (newHeight < Math.Max(0, MinMagnifiedHeightRatio * ActualHeight))
                {
                    newHeight = Math.Max(0, MinMagnifiedHeightRatio * ActualHeight);
                }
                MagnifiedHeight = newHeight;
            }
        }

        private void DragTarget_PointerReleased(object sender, PointerRoutedEventArgs e)
        {
            _InteractionOngoing = false;
            if (_ContactPoints.ContainsKey(e.Pointer.PointerId))
            {
                _ContactPoints.Remove(e.Pointer.PointerId);
            }
            MagnifyingToolbarElement.DragTarget.ReleasePointerCapture(e.Pointer);
            UpdatePointerAppearance(e.Pointer);
        }

        private void DragTarget_PointerCanceled(object sender, PointerRoutedEventArgs e)
        {
            _InteractionOngoing = false;
            if (_ContactPoints.ContainsKey(e.Pointer.PointerId))
            {
                _ContactPoints.Remove(e.Pointer.PointerId);
            }
            MagnifyingToolbarElement.DragTarget.ReleasePointerCapture(e.Pointer);
            UpdatePointerAppearance(e.Pointer);
        }

        private void DragTarget_PointerCaptureLost(object sender, PointerRoutedEventArgs e)
        {
            _InteractionOngoing = false;
            if (_ContactPoints.ContainsKey(e.Pointer.PointerId))
            {
                _ContactPoints.Remove(e.Pointer.PointerId);
            }
            UpdatePointerAppearance(e.Pointer);
        }

        private void UpdatePointerAppearance(Pointer pointer)
        {
            if (pointer.PointerDeviceType == Windows.Devices.Input.PointerDeviceType.Mouse)
            {
                bool shouldBeDragShape = _EnteredMousePointers.ContainsKey(pointer.PointerId) || (_ContactPoints.ContainsKey(pointer.PointerId) && _InteractionOngoing);
                if (shouldBeDragShape)
                {
                    Windows.UI.Xaml.Window.Current.CoreWindow.PointerCursor = 
                        new Windows.UI.Core.CoreCursor(Windows.UI.Core.CoreCursorType.SizeNorthSouth, pointer.PointerId);
                }
                else
                {
                    Windows.UI.Xaml.Window.Current.CoreWindow.PointerCursor = 
                        new Windows.UI.Core.CoreCursor(Windows.UI.Core.CoreCursorType.Arrow, pointer.PointerId);
                }
            }
        }

        #endregion Interaction


    }
}
