﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Media;

namespace pdftron.PDF.Tools.Controls.Resources
{
    public class TabButton : Button
    {
        public TabButton()
            : base()
        {
        }

        public static readonly DependencyProperty PointerOverBackgroundBrushProperty =
        DependencyProperty.RegisterAttached("PointerOverBackgroundBrush", typeof(Brush),
        typeof(TabButton), new PropertyMetadata(null));

        private Brush _PointerOverBackgroundBrush;
        public Brush PointerOverBackgroundBrush
        {
            get { return _PointerOverBackgroundBrush; }
            set { _PointerOverBackgroundBrush = value; }
        }

        public static readonly DependencyProperty PointerOverForegroundBrushProperty =
        DependencyProperty.RegisterAttached("PointerOverForegroundBrush", typeof(Brush),
        typeof(TabButton), new PropertyMetadata(null));

        private Brush _PointerOverForegroundBrush;
        public Brush PointerOverForegroundBrush
        {
            get { return _PointerOverForegroundBrush; }
            set { _PointerOverForegroundBrush = value; }
        }

        public static readonly DependencyProperty PointerPressedBackgroundBrushProperty =
        DependencyProperty.RegisterAttached("PointerPressedBackgroundBrush", typeof(Brush),
        typeof(TabButton), new PropertyMetadata(null));

        private Brush _PointerPressedBackgroundBrush;
        public Brush PointerPressedBackgroundBrush
        {
            get { return _PointerPressedBackgroundBrush; }
            set { _PointerPressedBackgroundBrush = value; }
        }

        public static readonly DependencyProperty PointerPressedForegroundBrushProperty =
        DependencyProperty.RegisterAttached("PointerPressedForegroundBrush", typeof(Brush),
        typeof(TabButton), new PropertyMetadata(null));

        private Brush _PointerPressedForegroundBrush;
        public Brush PointerPressedForegroundBrush
        {
            get { return _PointerPressedForegroundBrush; }
            set { _PointerPressedForegroundBrush = value; }
        }

        public static readonly DependencyProperty CornerRadiusProperty =
        DependencyProperty.RegisterAttached("CornerRadius", typeof(CornerRadius),
        typeof(TabButton), new PropertyMetadata(null));

        private CornerRadius _CornerRadius;
        public CornerRadius CornerRadius
        {
            get { return _CornerRadius; }
            set { _CornerRadius = value; }
        }

        public static readonly DependencyProperty PropertiesIndicatorProperty =
            DependencyProperty.RegisterAttached("PropertiesIndicator", typeof(Visibility),
            typeof(TabButton), new PropertyMetadata(Visibility.Collapsed));

        public Visibility PropertiesIndicator
        {
            get { return (Visibility)GetValue(PropertiesIndicatorProperty); }
            set { SetValue(PropertiesIndicatorProperty, value); }
        }
    }

    /// <summary>
    /// The main use of this class is to make it possible to change the opacity of the label,
    /// which is guaranteed to work even if the IsColapsed state doesn't get entered 
    /// (Seems to be an UWP bug).
    /// </summary>
    public class ExtendedAppBarButton : AppBarButton
    {
        public ExtendedAppBarButton()
            : base() { }

        public static readonly DependencyProperty LabelOpacityProperty =
            DependencyProperty.RegisterAttached("LabelOpacity", typeof(double),
                typeof(ExtendedAppBarButton), new PropertyMetadata(1.0));

        public double LabelOpacity
        {
            get { return (double)GetValue(LabelOpacityProperty); }
            set { SetValue(LabelOpacityProperty, value); }
        }
    }
    public class AppBarTabButton : ExtendedAppBarButton
    {
        public AppBarTabButton()
            : base()
        {
        }

        public static readonly DependencyProperty PointerOverBackgroundBrushProperty =
        DependencyProperty.RegisterAttached("PointerOverBackgroundBrush", typeof(Brush),
        typeof(AppBarTabButton), new PropertyMetadata(null));

        private Brush _PointerOverBackgroundBrush;
        public Brush PointerOverBackgroundBrush
        {
            get { return _PointerOverBackgroundBrush; }
            set { _PointerOverBackgroundBrush = value; }
        }

        public static readonly DependencyProperty PointerOverForegroundBrushProperty =
        DependencyProperty.RegisterAttached("PointerOverForegroundBrush", typeof(Brush),
        typeof(AppBarTabButton), new PropertyMetadata(null));

        private Brush _PointerOverForegroundBrush;
        public Brush PointerOverForegroundBrush
        {
            get { return _PointerOverForegroundBrush; }
            set { _PointerOverForegroundBrush = value; }
        }

        public static readonly DependencyProperty PointerPressedBackgroundBrushProperty =
        DependencyProperty.RegisterAttached("PointerPressedBackgroundBrush", typeof(Brush),
        typeof(AppBarTabButton), new PropertyMetadata(null));

        private Brush _PointerPressedBackgroundBrush;
        public Brush PointerPressedBackgroundBrush
        {
            get { return _PointerPressedBackgroundBrush; }
            set { _PointerPressedBackgroundBrush = value; }
        }

        public static readonly DependencyProperty PointerPressedForegroundBrushProperty =
        DependencyProperty.RegisterAttached("PointerPressedForegroundBrush", typeof(Brush),
        typeof(AppBarTabButton), new PropertyMetadata(null));

        private Brush _PointerPressedForegroundBrush;
        public Brush PointerPressedForegroundBrush
        {
            get { return _PointerPressedForegroundBrush; }
            set { _PointerPressedForegroundBrush = value; }
        }

        public static readonly DependencyProperty PropertiesIndicatorProperty =
            DependencyProperty.RegisterAttached("PropertiesIndicator", typeof(Visibility),
            typeof(AppBarTabButton), new PropertyMetadata(Visibility.Collapsed));

        public Visibility PropertiesIndicator
        {
            get { return (Visibility)GetValue(PropertiesIndicatorProperty); }
            set { SetValue(PropertiesIndicatorProperty, value); }
        }
    }
}
