﻿using pdftron.Common;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Xml;
using System.Xml.Serialization;
using Windows.Foundation;
using Windows.Storage;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Media.Animation;
using Windows.UI.Xaml.Shapes;

using PDFDouble = pdftron.Common.DoubleRef;

namespace pdftron.PDF.Tools.Controls.ViewModels
{
    public class AnnotationItem : ViewModelBase
    {
        /// <summary>
        ///  Should not be relied upon to stay consistent.
        /// </summary>
        public IAnnot Annot { get; set; }

        private string _Author = string.Empty;
        public string Author
        {
            get { return _Author; }
            set
            {
                if (Set(ref _Author, value))
                {
                    RaisePropertyChanged("AuthorAndContent");
                }
            }
        }

        private string _Content = string.Empty;
        public string Content
        {
            get { return _Content; }
            set
            {
                if (Set(ref _Content, value))
                {
                    RaisePropertyChanged("AuthorAndContent");
                }
            }
        }
        public string AuthorAndContent
        {
            get
            {
                if (!string.IsNullOrWhiteSpace(Author))
                {
                    return Author + ": " + Content;
                }
                else if (!string.IsNullOrWhiteSpace(Content))
                {
                    return Content;
                }
                return string.Empty;
            }
        }
        private int _PageNumber = 0;
        public int PageNumber
        {
            get { return _PageNumber; }
            set { Set(ref _PageNumber, value); }
        }
        public Rect AnnotRect { get; set; }

        private string _AdditionalInformation = string.Empty;
        public string AdditionalInformation
        {
            get { return _AdditionalInformation; }
            set
            {
                if (Set(ref _AdditionalInformation, value))
                {
                    RaisePropertyChanged("AnnotationSymbol");
                }
            }
        }

        private AnnotType _AnnotationType;
        public AnnotType AnnotationType
        {
            get { return _AnnotationType; }
            set
            {
                if (Set(ref _AnnotationType, value))
                {
                    RaisePropertyChanged("AnnotationSymbol");
                }
            }
        }
        public string AnnotationSymbol
        {
            get
            {
                string annotGlyph = new string((char)0x270F, 1); // default
                switch (AnnotationType)
                {
                    case AnnotType.e_Line:
                        annotGlyph = new string((char)0xE098, 1);
                        if (!string.IsNullOrWhiteSpace(AdditionalInformation) && AdditionalInformation.Equals("Arrow", StringComparison.OrdinalIgnoreCase))
                        {
                            annotGlyph = new string((char)0x0059, 1);
                        }
                        break;
                    case AnnotType.e_Square:
                        annotGlyph = new string((char)0xE095, 1);
                        break;
                    case AnnotType.e_Circle:
                        annotGlyph = new string((char)0xE096, 1);
                        break;
                    case AnnotType.e_Ink:
                        annotGlyph = new string((char)0xE235, 1);
                        if (!string.IsNullOrWhiteSpace(AdditionalInformation) && AdditionalInformation.Equals("Marker", StringComparison.OrdinalIgnoreCase))
                        {
                            annotGlyph = new string((char)0x00AE, 1);
                        }
                        break;
                    case AnnotType.e_Polygon:
                        annotGlyph = new string((char)0xE097, 1);
                        break;
                    case AnnotType.e_Polyline:
                        annotGlyph = new string((char)0x004E, 1);
                        break;
                    case AnnotType.e_FreeText:
                        annotGlyph = new string((char)0x0050, 1);
                        if (!string.IsNullOrWhiteSpace(AdditionalInformation) && AdditionalInformation.Equals("Callout", StringComparison.OrdinalIgnoreCase))
                        {
                            annotGlyph = new string((char)0x0049, 1);
                        }
                        break;
                    case AnnotType.e_Text:
                        annotGlyph = new string((char)0xE310, 1);
                        break;
                    case AnnotType.e_Highlight:
                        annotGlyph = new string((char)0x0057, 1);
                        break;
                    case AnnotType.e_Underline:
                        annotGlyph = new string((char)0xE104, 1);
                        break;
                    case AnnotType.e_StrikeOut:
                        annotGlyph = new string((char)0xE105, 1);
                        break;
                    case AnnotType.e_Squiggly:
                        annotGlyph = new string((char)0x005A, 1);
                        break;
                    case AnnotType.e_Caret:
                        annotGlyph = new string((char)0x004F, 1);
                        break;
                    case AnnotType.e_Stamp:
                        annotGlyph = new string((char)0x004A, 1);
                        if (!string.IsNullOrWhiteSpace(AdditionalInformation))
                        {
                            if (AdditionalInformation.Equals("Signature", StringComparison.OrdinalIgnoreCase))
                            {
                                annotGlyph = new string((char)0x004D, 1);
                            }
                            else if (AdditionalInformation.Equals("Image", StringComparison.OrdinalIgnoreCase))
                            {
                                annotGlyph = new string((char)0xE139, 1);
                            }
                        }
                        break;
                    case AnnotType.e_Redact:
                        annotGlyph = new string((char)0xE208, 1);
                        break;
                }


                return annotGlyph;
            }
        }

        public AnnotationItem(string author, string content, AnnotType annotationSymbol, int pageNumber, Rect annotRect, string additionalInformation = null)
        {
            _Author = author;
            _Content = content;
            _AnnotationType = annotationSymbol;
            _PageNumber = pageNumber;
            AnnotRect = annotRect;
            _AdditionalInformation = additionalInformation;
        }

        public void CopyFrom(AnnotationItem other)
        {
            Author = other.Author;
            Content = other.Content;
            AnnotationType = other.AnnotationType;
            PageNumber = other.PageNumber;
            AnnotRect = other.AnnotRect;
            AdditionalInformation = other.AdditionalInformation;
        }
    }

    public class PageHeaderItem
    {
        public int PageNumber { get; set; }
        public string PageHeader
        {
            get
            {
                return string.Format(ResourceHandler.GetString("AnnotationList_Page"), PageNumber);
            }
        }

        public ObservableCollection<AnnotationItem> Annotations { get; set; }

        public PageHeaderItem(IList<AnnotationItem> items)
        {
            PageNumber = 0;
            if (items != null && items.Count > 0)
            {
                PageNumber = items[0].PageNumber;
            }
            Annotations = new ObservableCollection<AnnotationItem>();
            foreach (AnnotationItem item in items)
            {
                Annotations.Add(item);
            }
        }
    }

    /// <summary>
    /// Note that if the document is being modified while this list is being generated
    /// it could cause errors. Ensure that ClearViewModel() followed by awaiting
    /// WaitForAnnotationListToCancel() happens before the document is modified.
    /// </summary>
    public class AnnotationListViewModel : ViewModelBase
    {
        private static List<AnnotType> WORKABLE_ANNOTATIONS = new List<AnnotType>()
        {
            AnnotType.e_Circle,
            AnnotType.e_Square,
            AnnotType.e_Text,
            AnnotType.e_Line,
            AnnotType.e_Polygon, 
            AnnotType.e_Underline, 
            AnnotType.e_StrikeOut,
            AnnotType.e_Ink,
            AnnotType.e_Highlight,
            AnnotType.e_FreeText,
            AnnotType.e_Squiggly,
            AnnotType.e_Stamp,
            AnnotType.e_Caret,
            AnnotType.e_Polyline,
            AnnotType.e_Redact
        };

        private bool _IsFinishedLoading = false;
        public bool IsFinishedLoading
        {
            get { return _IsFinishedLoading; }
            private set { Set(ref _IsFinishedLoading, value); }
        }

        private PDFViewCtrl _PDFViewCtrl;
        public PDFViewCtrl PDFViewCtrl
        {
            get { return _PDFViewCtrl; }
            set 
            { 
                _PDFViewCtrl = value;
                Init();
            }
        }

        private ToolManager _ToolManager = null;
        public ToolManager ToolManager
        {
            get { return _ToolManager; }
            set
            {
                if (_ToolManager != value)
                {
                    if (_ToolManager != null)
                    {
                        _ToolManager.SelectionChanged -= ToolManager_SelectionChanged;
                        _ToolManager.AnnotationAdded -= ToolManager_AnnotationAdded;
                        _ToolManager.AnnotationEdited -= ToolManager_AnnotationEdited;
                        _ToolManager.AnnotationPreRemoved -= ToolManager_AnnotationPreRemoved;

                        _ToolManager.SelectionGroupChanged -= ToolManager_SelectionGroupChanged;
                        _ToolManager.AnnotationGroupAdded -= ToolManager_AnnotationGroupAdded;
                        _ToolManager.AnnotationGroupEdited -= ToolManager_AnnotationGroupEdited;
                        _ToolManager.AnnotationGroupPreRemoved -= ToolManager_AnnotationGroupPreRemoved;
                        _ToolManager.AnnotationGroupPreEdited -= ToolManager_AnnotationGroupPreEdited;
                    }
                    _ToolManager = value;
                    if (_ToolManager != null)
                    {
                        _ToolManager.SelectionChanged += ToolManager_SelectionChanged;
                        _ToolManager.AnnotationAdded += ToolManager_AnnotationAdded;
                        _ToolManager.AnnotationEdited += ToolManager_AnnotationEdited;
                        _ToolManager.AnnotationPreRemoved += ToolManager_AnnotationPreRemoved;

                        _ToolManager.SelectionGroupChanged += ToolManager_SelectionGroupChanged;
                        _ToolManager.AnnotationGroupAdded += ToolManager_AnnotationGroupAdded;
                        _ToolManager.AnnotationGroupEdited += ToolManager_AnnotationGroupEdited;
                        _ToolManager.AnnotationGroupPreRemoved += ToolManager_AnnotationGroupPreRemoved;
                        _ToolManager.AnnotationGroupPreEdited += ToolManager_AnnotationGroupPreEdited;
                    }
                }
            }
        }

        private NavigationStack _NavigationStack = null;
        public NavigationStack NavigationStack
        {
            get { return _NavigationStack; }
            set { _NavigationStack = value; }
        }

        public AnnotationListViewModel()
        {
            AnnotationGroups = new ObservableCollection<PageHeaderItem>();
            InitInteraction();
        }

        public AnnotationListViewModel(PDFViewCtrl ctrl)
        {
            AnnotationGroups = new ObservableCollection<PageHeaderItem>();
            InitInteraction();
            PDFViewCtrl = ctrl;
        }

        private void Init()
        {
            CreateAnnotationList();
        }

        public void ClearViewModel()
        {
            _Cancel = true;
            ToolManager = null;
        }

        /// <summary>
        /// Wait for this before disposing the PDFDoc, after you call ClearViewModel().
        /// This guarantees that annots are not being extracted.
        /// </summary>
        /// <returns></returns>
        public async Task WaitForAnnotationListToFinish()
        {
            System.Diagnostics.Debug.WriteLine("Waiting for annot list to cancel");
            await AnnotListSemaphore.WaitAsync();
            System.Diagnostics.Debug.WriteLine("Annot list is cancelled or finished");
            AnnotListSemaphore.Release();
        }

        private bool _Cancel = false;
        public SemaphoreSlim _AnnotListSemaphore;
        public SemaphoreSlim AnnotListSemaphore
        {
            get
            {
                if (_AnnotListSemaphore == null)
                {
                    _AnnotListSemaphore = new SemaphoreSlim(1);
                }
                return _AnnotListSemaphore;
            }
        }


        #region Properties
        public ObservableCollection<PageHeaderItem> AnnotationGroups { get; private set; }

        private bool _FoundAnnotations = true;
        public bool FoundAnnotations
        {
            get { return _FoundAnnotations; }
            set
            {
                if (_FoundAnnotations != value)
                {
                    _FoundAnnotations = value;
                    RaisePropertyChanged();
                }
            }
        }

        private Rectangle _FlashingRectangle;

        private Brush _FlashBrush = new SolidColorBrush(Windows.UI.Colors.Transparent);
        public Brush FlashBrush
        {
            get { return _FlashBrush; }
            set
            {
                _FlashBrush = value;
                if (_FlashingRectangle != null)
                {
                    _FlashingRectangle.Fill = value;
                }
            }
        }

        private Storyboard _FlashAnimation = null;
        public Storyboard FlashAnimation
        {
            get { return _FlashAnimation; }
            set 
            { 
                _FlashAnimation = value;
                Storyboard.SetTarget(FlashAnimation, _FlashingRectangle);
                Storyboard.SetTargetName(FlashAnimation, _FlashingRectangle.Name);
                _FlashAnimation.Completed += FlashAnimation_Completed;
            }
        }

        public delegate void PageChangedDelegate(int prevPage, int newPage);
        public event PageChangedDelegate PageChanged;

        #endregion Properties

        private int _NextPageToProcess = 0;
        private async void CreateAnnotationList()
        {
            PDFDoc doc = _PDFViewCtrl.GetDoc();

            if (doc != null)
            {
                bool gotItems = false;
                _NextPageToProcess = 1;
                IList<AnnotationItem> items = await GetNextPageWithAnnotsAsync(_NextPageToProcess, doc);
                while (items != null && !_Cancel)
                {
                    if (items.Count > 0)
                    {
                        _NextPageToProcess = items[0].PageNumber + 1;
                    }
                    gotItems = true;
                    foreach (AnnotationItem item in items)
                    {
                        item.AnnotRect = _PDFViewCtrl.GetPageRectForAnnot(item.Annot, item.PageNumber);
                        item.Annot = null;
                    }
                    PageHeaderItem header = new PageHeaderItem(items);
                    AnnotationGroups.Add(header);
                    items = await GetNextPageWithAnnotsAsync(_NextPageToProcess, doc);
                }
                _NextPageToProcess = 0;
                IsFinishedLoading = true;
                if (!gotItems)
                {
                    FoundAnnotations = false;
                }
            }
        }

        private IAsyncOperation<IList<AnnotationItem>> GetNextPageWithAnnotsAsync(int pageNumber, PDFDoc doc)
        {
            Task<IList<AnnotationItem>> t = new Task<IList<AnnotationItem>>(() =>
            {
                return GetNextPageWithAnnots(pageNumber, doc);
            });
            t.Start();
            return t.AsAsyncOperation<IList<AnnotationItem>>();
        }

        private IList<AnnotationItem> GetNextPageWithAnnots(int pageNumber, PDFDoc doc)
        {
            try
            {
                AnnotListSemaphore.Wait();
                if (_Cancel)
                {
                    return null;
                }
                try
                {
                    doc.LockRead();
                    PageIterator pageIter = doc.GetPageIterator(pageNumber);
                    while (pageIter.HasNext() && !_Cancel)
                    {
                        pdftron.PDF.Page page = pageIter.Current();
                        int numAnnots = page.GetNumAnnots();
                        if (numAnnots > 0)
                        {
                            IList<AnnotationItem> items = GetAnnotationItemsOnPage(page);
                            if (items.Count > 0)
                            {
                                pageIter.Next();
                                return items;
                            }
                        }

                        pageIter.Next();
                    }
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
                finally
                {
                    doc.UnlockRead();
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                AnnotListSemaphore.Release();
            }
            return null;
        }


        private IList<AnnotationItem> GetAnnotationItemsOnPage(pdftron.PDF.Page page)
        {
            List<AnnotationItem> items = new List<AnnotationItem>();
            int numAnnots = page.GetNumAnnots();
            for (int i = 0; i < numAnnots; i++)
            {
                IAnnot annot = page.GetAnnot(i);
                if (annot == null || !annot.IsValid()) continue;

                AnnotType annotType = annot.GetAnnotType();
                if (WORKABLE_ANNOTATIONS.Contains(annotType) && annot.IsMarkup())
                {
                    items.Add(MakeAnnotationItem(annot, annotType, page));
                }
            }

            return items;
        }

        private AnnotationItem MakeAnnotationItem(IAnnot annot, AnnotType annotType, Page page)
        {
            string author = "";
            string contents = "";
            string additonalInfo = "";

            if (annot is pdftron.PDF.Annots.IMarkup)
            {
                Annots.IMarkup markup = (Annots.IMarkup)annot;

                author = markup.GetTitle();

                pdftron.PDF.Annots.Popup popup = markup.GetPopup();
                if (popup != null && popup.IsValid())
                {
                    contents = popup.GetContents();
                }
                else
                {
                    contents = markup.GetContents();
                }

                if (string.IsNullOrWhiteSpace(contents) && annot is Annots.ITextMarkup)
                {
                    pdftron.PDF.TextExtractor textExtractor = new pdftron.PDF.TextExtractor();

                    try
                    {
                        int indexy = page.GetIndex();
                        textExtractor.Begin(page);
                        String text = textExtractor.GetTextUnderAnnot(annot);
                        if (!string.IsNullOrWhiteSpace(text))
                        {
                            contents = text;
                        }
                    }
                    catch (Exception ex)
                    {
                        Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                    }
                }
            }

            // if stamp, see if it's a signature
            if (annotType == AnnotType.e_Stamp)
            {
                pdftron.SDF.Obj sigObj = annot.GetSDFObj();
                pdftron.SDF.Obj stampObj = annot.GetSDFObj();
                sigObj = sigObj.FindObj(ToolManager.SignatureAnnotationIdentifyingString);
                stampObj = stampObj.FindObj(ToolManager.StamperAnnotationIdentifyingString);
                if (sigObj != null)
                {
                    additonalInfo = "Signature";
                }
                else if (stampObj != null)
                {
                    additonalInfo = "Image";
                }
            }

            // if a line, see if it is an arrow
            if (annotType == AnnotType.e_Line)
            {
                Annots.Line line = (Annots.Line)annot;
                if (UtilityFunctions.IsArrow(line))
                {
                    additonalInfo = "arrow";
                }
            }

            // if a free text, see if it's a callout
            if (annotType == AnnotType.e_FreeText)
            {
                pdftron.SDF.Obj obj = annot.GetSDFObj();
                obj = obj.FindObj("CL");
                if (obj != null)
                {
                    additonalInfo = "Callout";
                }
            }

            if (annotType == AnnotType.e_Ink)
            {
                Annots.Ink ink = (Annots.Ink)annot;
                // check if it is a marker
                if (ink.GetBlendMode() == GStateBlendMode.e_bl_multiply)
                {
                    additonalInfo = "Marker";
                }
                else
                {
                    additonalInfo = "Ink";
                }
            }

            AnnotationItem item = new AnnotationItem(author, contents, annotType, page.GetIndex(), annot.GetRect(), additonalInfo);
            item.Annot = annot;

            return item;
        }


        #region Interaction

        private void InitInteraction()
        {
            ItemClickCommand = new RelayCommand(ItemClickCommandImpl);
            _FlashingRectangle = new Rectangle();
            _FlashingRectangle.Opacity = 0;
            _FlashingRectangle.Fill = FlashBrush;
        }

        private enum SelectionChangeSources
        {
            None,
            AnnotationList,
            ToolManager,
        }
        private SelectionChangeSources _SelectionChangedSource = SelectionChangeSources.None;
        private AnnotationItem _SelectedItem = null;
        public AnnotationItem SelectedItem
        {
            get { return _SelectedItem; }
            set
            {
                if(Set(ref _SelectedItem, value))
                {
                    if (_SelectionChangedSource == SelectionChangeSources.None && ToolManager != null)
                    {
                        _SelectionChangedSource = SelectionChangeSources.AnnotationList;

                        ToolManager.UnselectAllAnnotations();

                        _SelectionChangedSource = SelectionChangeSources.None;
                    }
                    if (_SelectionChangedSource != SelectionChangeSources.ToolManager && SelectedItem != null)
                    {
                        ShowAnnotation(SelectedItem);
                    }
                }
            }
        }

        private List<AnnotationItem> _SelectedGroupItems = null;
        public List<AnnotationItem> SelectedGroupItems
        {
            get { return _SelectedGroupItems; }
            set
            {
                if (Set(ref _SelectedGroupItems, value))
                {
                    if (_SelectionChangedSource == SelectionChangeSources.None && ToolManager != null)
                    {
                        _SelectionChangedSource = SelectionChangeSources.AnnotationList;

                        ToolManager.UnselectAllAnnotations();

                        _SelectionChangedSource = SelectionChangeSources.None;
                    }
                    if (_SelectionChangedSource != SelectionChangeSources.ToolManager && SelectedGroupItems != null)
                    {
                        ShowAnnotation(SelectedGroupItems[0]);
                    }
                }
            }
        }

        public RelayCommand ItemClickCommand { get; set; }

        private void ItemClickCommandImpl(object param)
        {
            ItemClickEventArgs args = param as ItemClickEventArgs;
            if (args != null)
            {
                AnnotationItem item = args.ClickedItem as AnnotationItem;
                if (item != null && item == SelectedItem)
                {
                    ShowAnnotation(item);
                }
            }
        }

        public void ToolManager_SelectionChanged(IAnnot annotation, int pageNumber)
        {
            try
            {
                if (_SelectionChangedSource == SelectionChangeSources.None)
                {
                    _SelectionChangedSource = SelectionChangeSources.ToolManager;
                    if (annotation == null)
                    {
                        SelectedItem = null;
                    }
                    else
                    {
                        AnnotationItem item = FindItemWithAnnot(annotation, pageNumber);
                        SelectedItem = item;
                    }
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                _SelectionChangedSource = SelectionChangeSources.None;
            }
        }

        public void ToolManager_SelectionGroupChanged(Dictionary<IAnnot, int> annotationGroup)
        {
            try
            {
                if (_SelectionChangedSource == SelectionChangeSources.None)
                {
                    _SelectionChangedSource = SelectionChangeSources.ToolManager;
                    if (annotationGroup == null)
                    {
                        SelectedGroupItems = null;
                    }
                    else
                    {
                        SelectedGroupItems.Clear();
                        foreach (var annotationItem in annotationGroup)
                        {
                            AnnotationItem item = FindItemWithAnnot(annotationItem.Key, annotationItem.Value);
                            SelectedGroupItems.Add(item);
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                _SelectionChangedSource = SelectionChangeSources.None;
            }
        }

        private void ShowAnnotation(AnnotationItem item)
        {
            NavigationState beforeState = NavigationState.TakeSnapshot(PDFViewCtrl);
            SwitchPageIfNecessary(item.PageNumber);
            int nowPage = PDFViewCtrl.GetCurrentPage();

            Rect canvasRect = null;
            try
            {
                _PDFViewCtrl.DocLockRead();
                IAnnot annot = FindAnnotationFromItem(item);
                if (annot != null && annot.IsValid())
                {
                    Rect annotScreenRect = _PDFViewCtrl.GetScreenRectForAnnot(annot, item.PageNumber);
                    canvasRect = UtilityFunctions.ConvertFromScreenRectToAnnotationCanvasRect(_PDFViewCtrl, annotScreenRect);
                    canvasRect.Normalize();

                    ScrollRectIntoView(item.PageNumber, canvasRect);

                    annotScreenRect = _PDFViewCtrl.GetScreenRectForAnnot(annot, item.PageNumber);
                    canvasRect = UtilityFunctions.ConvertFromScreenRectToAnnotationCanvasRect(_PDFViewCtrl, annotScreenRect);
                    canvasRect.Normalize();
                }
            }
            catch (Exception e)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(e));
            }
            finally
            {
                _PDFViewCtrl.DocUnlockRead();
            }

            if (canvasRect == null)
            {
                return;
            }

            NavigationState afterState = NavigationState.TakeSnapshot(PDFViewCtrl);
            if (NavigationStack != null)
            {
                NavigationStack.RegisterJump(beforeState, afterState);
            }

            if (FlashAnimation != null)
            {
                FlashAnnotation(item.PageNumber, canvasRect);
            }
        }

        /// <summary>
        /// Will switch page if the current page can not be scrolled into view.
        /// </summary>
        /// <param name="pageNumber"></param>
        public void SwitchPageIfNecessary(int pageNumber)
        {
            // If in a non-continuous mode we want to first go to the page.
            PDFViewCtrlPagePresentationMode mode = this.PDFViewCtrl.GetPagePresentationMode();
            if (mode == PDFViewCtrlPagePresentationMode.e_single_page ||
                mode == PDFViewCtrlPagePresentationMode.e_facing ||
                mode == PDFViewCtrlPagePresentationMode.e_facing_cover)
            {
                int currentPage = PDFViewCtrl.GetCurrentPage();
                if (currentPage == pageNumber)
                {
                    return;
                }

                if (mode == PDFViewCtrlPagePresentationMode.e_facing && ((currentPage + 1) / 2 == (pageNumber + 1) / 2))
                {
                    return;
                }
                else if (mode == PDFViewCtrlPagePresentationMode.e_facing_cover && (currentPage / 2 == pageNumber / 2))
                {
                    return;
                }
                PageChanged?.Invoke(_PDFViewCtrl.GetCurrentPage(), pageNumber);
                this.PDFViewCtrl.SetCurrentPage(pageNumber);
            }
        }

        public void ScrollRectIntoView(int pageNumber, Rect rect)
        {
            // Now, compare the view's position on the canvas to that of the annotation

            PDFDouble px = new PDFDouble(0);
            PDFDouble py = new PDFDouble(0);
            _PDFViewCtrl.ConvScreenPtToAnnotationCanvasPt(px, py);

            Rect viewRect = new Rect(px.Value, py.Value, px.Value + this.PDFViewCtrl.ActualWidth, py.Value + this.PDFViewCtrl.ActualHeight);
            if (viewRect.Width() < 1 || viewRect.Height() < 1)
            {
                return;
            }

            Rect intersectRect = new Rect();
            bool intersect = intersectRect.IntersectRect(viewRect, rect);

            // If we don't intersect enough, scroll it into view.
            if (!intersect || (intersectRect.Width() < rect.Width() - 1) || (intersectRect.Height() < rect.Height() - 1))
            {
                double annotWidth = rect.Width();
                double annotHeight = rect.Height();
                double viewWidth = viewRect.Width();
                double viewHeight = viewRect.Height();
                double x = -1;
                double y = -1;
                double zoomToBeDone = 1.0;
                double MAX_MARGIN = 10.0;

                if (annotWidth > 0 && annotHeight > 0)
                {
                    if (viewWidth < annotWidth)
                    {
                        zoomToBeDone = viewWidth / annotWidth;
                        x = (rect.x1 + rect.x2 - viewWidth) / 2.0; // center
                    }
                    else
                    {
                        double margin = Math.Min(MAX_MARGIN, (viewWidth - annotWidth) / 2.0);
                        if (viewRect.x1 > rect.x1 - margin)
                        {
                            x = Math.Max(rect.x1 - margin, 0);
                        }
                        else if (viewRect.x2 < rect.x2 + margin)
                        {
                            x = Math.Max(rect.x2 + margin - viewWidth, 0);
                        }
                    }


                    if (viewHeight < annotHeight)
                    {
                        zoomToBeDone = Math.Min(viewHeight / annotHeight, zoomToBeDone);
                        y = (rect.y1 + rect.y2 - viewHeight) / 2.0;
                    }
                    else
                    {
                        double margin = Math.Min(MAX_MARGIN, (viewHeight - annotHeight) / 2.0);
                        if (viewRect.y1 > rect.y1 - margin)
                        {
                            y = Math.Max(rect.y1 - margin, 0);
                        }
                        else if (viewRect.y2 < rect.y2 + margin)
                        {
                            y = Math.Max(rect.y2 + margin - viewHeight, 0);
                        }
                    }
                }

                if (x >= 0)
                {
                    this.PDFViewCtrl.SetHScrollPos(x);
                }
                if (y >= 0)
                {
                    this.PDFViewCtrl.SetVScrollPos(y);
                }
                if (zoomToBeDone < 1.0)
                {
                    this.PDFViewCtrl.SetZoom(this.PDFViewCtrl.GetZoom() * zoomToBeDone);
                }
            }
        }

        private void FlashAnnotation(int pagenumber, Rect rect)
        {
            DetachFlashinfRectangle();

            Canvas canvas = _PDFViewCtrl.GetAnnotationCanvas();
            canvas.Children.Add(_FlashingRectangle);
            _FlashingRectangle.Tag = canvas;
            _FlashingRectangle.Opacity = 0;
            _FlashingRectangle.Width = rect.Width();
            _FlashingRectangle.Height = rect.Height();
            var px = new pdftron.Common.DoubleRef(rect.x1);
            var py = new pdftron.Common.DoubleRef(rect.y1);

            var mode = PDFViewCtrl.GetPagePresentationMode();

            _FlashingRectangle.SetValue(Canvas.LeftProperty, px.Value);
            _FlashingRectangle.SetValue(Canvas.TopProperty, py.Value);

            FlashAnimation.Begin();
        }

        private void FlashingAnimationCompleted(object sender, object e)
        {
            DetachFlashinfRectangle();
        }

        private void DetachFlashinfRectangle()
        {
            Canvas parent = _FlashingRectangle.Tag as Canvas;
            if (parent != null)
            {
                parent.Children.Remove(_FlashingRectangle);
            }
        }

        void FlashAnimation_Completed(object sender, object e)
        {
            DetachFlashinfRectangle();
        }

        private IAnnot FindAnnotationFromItem(AnnotationItem item)
        {
            foreach (PageHeaderItem header in AnnotationGroups)
            {
                if (header.PageNumber == item.PageNumber)
                {
                    if (header.Annotations.Contains(item))
                    {
                        return FindAnnotationFromIndex(header.Annotations.IndexOf(item), header);
                    }

                }
            }
            return null;
        }

        private IAnnot FindAnnotationFromIndex(int index, PageHeaderItem header)
        {
            try
            {
                PDFViewCtrl.DocLockRead();
                PDFDoc doc = PDFViewCtrl.GetDoc();

                Page page = doc.GetPage(header.PageNumber);
                int numAnnots = page.GetNumAnnots();
                int validIndexes = 0;
                for (int i = 0; i < numAnnots; ++i)
                {
                    IAnnot pageAnnot = page.GetAnnot(i);
                    if (pageAnnot == null || !pageAnnot.IsValid()) continue;

                    if (WORKABLE_ANNOTATIONS.Contains(pageAnnot.GetAnnotType()) && pageAnnot.IsMarkup())
                    {
                        if (index == validIndexes)
                        {
                            return pageAnnot;
                        }
                        ++validIndexes;
                    }
                }
            }
            catch (Exception e)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(e));
            }
            finally
            {
                PDFViewCtrl.DocUnlockRead();
            }
            return null;
        }

        private AnnotationItem FindItemWithAnnot(IAnnot annot, int pageNumber)
        {
            foreach (PageHeaderItem header in AnnotationGroups)
            {
                if (header.PageNumber == pageNumber)
                {
                    return FindItemWithAnnot(annot, header);
                }
            }
            return null;
        }

        private AnnotationItem FindItemWithAnnot(IAnnot annot, PageHeaderItem header)
        {
            try
            {
                PDFViewCtrl.DocLockRead();
                PDFDoc doc = PDFViewCtrl.GetDoc();

                Page page = doc.GetPage(header.PageNumber);
                int numAnnots = page.GetNumAnnots();
                int index = 0;
                for (int i = 0; i < numAnnots; ++i)
                {
                    IAnnot pageAnnot = page.GetAnnot(i);
                    if (pageAnnot == null || !pageAnnot.IsValid()) continue;

                    if (WORKABLE_ANNOTATIONS.Contains(pageAnnot.GetAnnotType()) && pageAnnot.IsMarkup())
                    {
                        if (pageAnnot.GetSDFObj().IsEqual(annot.GetSDFObj()))
                        {
                            if (header.Annotations != null && header.Annotations.Count > index)
                            {
                                return header.Annotations[index];
                            }
                            return null; // something is wrong, and the annotation is not in the list.
                        }
                        ++index;
                    }
                }
            }
            catch (Exception e)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(e));
            }
            finally
            {
                PDFViewCtrl.DocUnlockRead();
            }
            return null;
        }


        private void ToolManager_AnnotationAdded(IAnnot annotation, int pageNumber)
        {
            if (_NextPageToProcess <= 1 || _NextPageToProcess > pageNumber)
            {
                ModificationList.Add(new Tuple<AnnotationModifications, IAnnot, int>(AnnotationModifications.Add, annotation, pageNumber));
                HandleModifications();
            }
        }

        private void ToolManager_AnnotationGroupAdded(Dictionary<IAnnot, int> annotationGroup)
        {
            foreach (var annotationItem in annotationGroup)
            {
                ToolManager_AnnotationAdded(annotationItem.Key, annotationItem.Value);
            }
        }

        private void ToolManager_AnnotationEdited(IAnnot annotation, int pageNumber)
        {
            if (_NextPageToProcess <= 1 || _NextPageToProcess > pageNumber)
            {
                ModificationList.Add(new Tuple<AnnotationModifications, IAnnot, int>(AnnotationModifications.Edit, annotation, pageNumber));
                HandleModifications();
            }
        }

        private void ToolManager_AnnotationGroupEdited(Dictionary<IAnnot, int> annotationGroup)
        {
            foreach (var annotationItem in annotationGroup)
            {
                ToolManager_AnnotationEdited(annotationItem.Key, annotationItem.Value);
            }
        }

        private void ToolManager_AnnotationGroupPreEdited(Dictionary<IAnnot, int> annotationGroup)
        {
            foreach (var annotationItem in annotationGroup)
            {
                ToolManager_AnnotationEdited(annotationItem.Key, annotationItem.Value);
            }
        }

        private void ToolManager_AnnotationPreRemoved(IAnnot annotation, int pageNumber)
        {
            if (_NextPageToProcess <= 1 || _NextPageToProcess > pageNumber)
            {
                ModificationList.Add(new Tuple<AnnotationModifications, IAnnot, int>(AnnotationModifications.Remove, annotation, pageNumber));
                HandleModifications();
            }
        }

        private void ToolManager_AnnotationGroupPreRemoved(Dictionary<IAnnot, int> annotationGroup)
        {
            foreach (var annotationItem in annotationGroup)
            {
                ToolManager_AnnotationPreRemoved(annotationItem.Key, annotationItem.Value);
            }
        }

        private enum AnnotationModifications
        {
            Add,
            Edit,
            Remove,
        }

        private List<Tuple<AnnotationModifications, IAnnot, int>> _ModificationList;
        private List<Tuple<AnnotationModifications, IAnnot, int>> ModificationList
        {
            get
            {
                if (_ModificationList == null)
                {
                    _ModificationList = new List<Tuple<AnnotationModifications, IAnnot, int>>();
                }
                return _ModificationList;
            }
        }

        private void HandleModifications()
        {
            while (ModificationList.Count > 0)
            {
                _SelectionChangedSource = SelectionChangeSources.ToolManager;
                switch (ModificationList[0].Item1)
                {
                    case AnnotationModifications.Add:
                        AddAnnotation(ModificationList[0].Item2, ModificationList[0].Item3);
                        break;
                    case AnnotationModifications.Edit:
                        EditAnnotation(ModificationList[0].Item2, ModificationList[0].Item3);
                        break;
                    case AnnotationModifications.Remove:
                        RemoveAnnotation(ModificationList[0].Item2, ModificationList[0].Item3);
                        break;
                }
                _SelectionChangedSource = SelectionChangeSources.None;
                ModificationList.RemoveAt(0);
            }
        }

        private void AddAnnotation(IAnnot annot, int pageNumber)
        {
            AnnotType annotType = annot.GetAnnotType();
            if (!WORKABLE_ANNOTATIONS.Contains(annotType) || !annot.IsMarkup())
            {
                return;
            }

            PageHeaderItem header = null;
            foreach (PageHeaderItem head in AnnotationGroups)
            {
                if (head.PageNumber >= pageNumber)
                {
                    header = head;
                    break;
                }
            }
            try
            {
                PDFViewCtrl.DocLockRead();
                Page page = PDFViewCtrl.GetDoc().GetPage(pageNumber);

                AnnotationItem item = MakeAnnotationItem(annot, annotType, page);
                if (header == null || header.PageNumber != pageNumber)
                {
                    List<AnnotationItem> items = new List<AnnotationItem> { item };
                    PageHeaderItem newHeader = new PageHeaderItem(items);
                    if (header == null)
                    {
                        AnnotationGroups.Add(newHeader);
                        FoundAnnotations = true;
                    }
                    else
                    {
                        int insertionIndex = AnnotationGroups.IndexOf(header);
                        AnnotationGroups.Insert(insertionIndex, newHeader);
                    }
                }
                else
                {
                    int annotIndex = 0;
                    int numAnnots = page.GetNumAnnots();
                    for (int i = 0; i < numAnnots; ++i)
                    {
                        IAnnot pageAnnot = page.GetAnnot(i);
                        if (pageAnnot == null || !pageAnnot.IsValid()) continue;

                        if (WORKABLE_ANNOTATIONS.Contains(pageAnnot.GetAnnotType()) && pageAnnot.IsMarkup())
                        {
                            if (pageAnnot.GetSDFObj().IsEqual(annot.GetSDFObj()))
                            {
                                if (header.Annotations != null && header.Annotations.Count > annotIndex)
                                {
                                    header.Annotations.Insert(annotIndex, item);
                                    return;
                                }
                                header.Annotations.Add(item);
                                return;
                            }
                            ++annotIndex;
                        }
                    }
                }
            }
            catch (Exception e)
            {
            Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(e));
        }
        finally
            {
                PDFViewCtrl.DocUnlockRead();
            }
        }

        private void RemoveAnnotation(IAnnot annot, int pageNumber)
        {
            AnnotationItem item = null;
            PageHeaderItem header = null;
            foreach (PageHeaderItem headerItem in AnnotationGroups)
            {
                if (headerItem.PageNumber == pageNumber)
                {
                    header = headerItem;
                    item = FindItemWithAnnot(annot, header);
                    break;
                }
            }
            if (header != null && item != null && header.Annotations.Contains(item))
            {
                header.Annotations.Remove(item);
                if (header.Annotations.Count == 0)
                {
                    AnnotationGroups.Remove(header);
                }
            }
        }

        private void EditAnnotation(IAnnot annot, int pageNumber)
        {
            AnnotationItem item = null;
            PageHeaderItem header = null;
            foreach (PageHeaderItem headerItem in AnnotationGroups)
            {
                if (headerItem.PageNumber == pageNumber)
                {
                    header = headerItem;
                    item = FindItemWithAnnot(annot, header);
                    break;
                }
            }

            if (header != null && item != null && header.Annotations.Contains(item))
            {
                try
                {
                    PDFViewCtrl.DocLockRead();
                    Page page = PDFViewCtrl.GetDoc().GetPage(pageNumber);
                    AnnotType annotType = annot.GetAnnotType();
                    AnnotationItem newItem = MakeAnnotationItem(annot, annotType, page);
                    item.CopyFrom(newItem);
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
                finally
                {
                    PDFViewCtrl.DocUnlockRead();
                }
            }
        }

        #endregion Interaction

        #region Back Key
        public bool GoBack()
        {

            return false;
        }

        #endregion Back Key
    }
}
