﻿using System;
using System.Threading.Tasks;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using pdftron.PDF.Tools.Controls.ControlBase;
using Windows.UI.Xaml.Controls.Primitives;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;

using UIRect = Windows.Foundation.Rect;
using Windows.UI.Xaml.Media;

namespace pdftron.PDF.Tools.Controls.ViewModels 
{
    public enum SelectedTool
    {
        Pan,
        Line,
        Arrow,
        Square,
        Oval,
        Ink,
        Eraser,
        Marker,
        TextBox,
        Sticky,
        Signature,
        Highlight,
        Underline,
        Strikeout,
        Squiggly,
        Stamper,
        None,
    }
    

    public class AnnotationToolbarViewModel : ViewModelBase
    {
        ICloseableControl _View;

#region Commands
        public RelayCommand ToolButtonClickedCommand {get; private set; }
        public RelayCommand InkToolbarCommand { get; private set; }
        public RelayCommand InkPresetButtonCommand { get; private set; }
        public RelayCommand PolygonSaveToolbarCommand { get; private set; }
        public RelayCommand UndoButtonClickedCommand { get; private set; }
        public RelayCommand RedoButtonClickedCommand { get; private set; }

        #endregion Commands

        #region Tool Visibility

        private bool _isStickyNoteToolButtonVisible = true;
        public bool IsStickyNoteToolButtonVisible
        {
            get { return _isStickyNoteToolButtonVisible; }
            set
            {
                if (value != _isStickyNoteToolButtonVisible)
                {
                    _isStickyNoteToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isHighlightToolButtonVisible = true;
        public bool IsHighlightToolButtonVisible
        {
            get { return _isHighlightToolButtonVisible; }
            set
            {
                if (value != _isHighlightToolButtonVisible)
                {
                    _isHighlightToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isUnderlineToolButtonVisible = true;
        public bool IsUnderlineToolButtonVisible
        {
            get { return _isUnderlineToolButtonVisible; }
            set
            {
                if (value != _isUnderlineToolButtonVisible)
                {
                    _isUnderlineToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isStrikeoutToolButtonVisible = true;
        public bool IsStrikeoutToolButtonVisible
        {
            get { return _isStrikeoutToolButtonVisible; }
            set
            {
                if (value != _isStrikeoutToolButtonVisible)
                {
                    _isStrikeoutToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isSquigglyToolButtonVisible = true;
        public bool IsSquigglyToolButtonVisible
        {
            get { return _isSquigglyToolButtonVisible; }
            set
            {
                if (value != _isSquigglyToolButtonVisible)
                {
                    _isSquigglyToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isStamperToolButtonVisible = true;
        public bool IsStamperToolButtonVisible
        {
            get { return _isStamperToolButtonVisible; }
            set
            {
                if (value != _isStamperToolButtonVisible)
                {
                    _isStamperToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isSignatureToolButtonVisible = true;
        public bool IsSignatureToolButtonVisible
        {
            get { return _isSignatureToolButtonVisible; }
            set
            {
                if (value != _isSignatureToolButtonVisible)
                {
                    _isSignatureToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isInkToolButtonVisible = true;
        public bool IsInkToolButtonVisible
        {
            get { return _isInkToolButtonVisible; }
            set
            {
                if (value != _isInkToolButtonVisible)
                {
                    _isInkToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isEraserToolButtonVisible = true;
        public bool IsEraserToolButtonVisible
        {
            get { return _isEraserToolButtonVisible; }
            set
            {
                if (value != _isEraserToolButtonVisible)
                {
                    _isEraserToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isMarkerToolButtonVisible = true;
        public bool IsMarkerToolButtonVisible
        {
            get { return _isMarkerToolButtonVisible; }
            set
            {
                if (value != _isMarkerToolButtonVisible)
                {
                    _isMarkerToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isTextBoxToolButtonVisible = true;
        public bool IsTextBoxToolButtonVisible
        {
            get { return _isTextBoxToolButtonVisible; }
            set
            {
                if (value != _isTextBoxToolButtonVisible)
                {
                    _isTextBoxToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isShapeToolsButtonVisible = true;
        public bool IsShapeToolsButtonVisible
        {
            get { return _isShapeToolsButtonVisible; }
            set
            {
                if (value != _isShapeToolsButtonVisible)
                {
                    _isShapeToolsButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isMultiSelectToolButtonVisible = true;
        public bool IsMultiSelectToolButtonVisible
        {
            get { return _isMultiSelectToolButtonVisible; }
            set
            {
                if (value != _isMultiSelectToolButtonVisible)
                {
                    _isMultiSelectToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isPanToolButtonVisible = true;
        public bool IsPanToolsButtonVisible
        {
            get { return _isPanToolButtonVisible; }
            set
            {
                if (value != _isPanToolButtonVisible)
                {
                    _isPanToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isUndoRedoToolButtonVisible = true;
        public bool IsUndoRedoToolButtonVisible
        {
            get { return _isUndoRedoToolButtonVisible; }
            set
            {
                if (value != _isUndoRedoToolButtonVisible)
                {
                    _isUndoRedoToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isSquareToolButtonVisible = true;
        public bool IsSquareToolButtonVisible
        {
            get { return _isSquareToolButtonVisible; }
            set
            {
                if (value != _isSquareToolButtonVisible)
                {
                    _isSquareToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isCircleToolButtonVisible = true;
        public bool IsCircleToolButtonVisible
        {
            get { return _isCircleToolButtonVisible; }
            set
            {
                if (value != _isCircleToolButtonVisible)
                {
                    _isCircleToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isLineToolButtonVisible = true;
        public bool IsLineToolButtonVisible
        {
            get { return _isLineToolButtonVisible; }
            set
            {
                if (value != _isLineToolButtonVisible)
                {
                    _isLineToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isArrowToolButtonVisible = true;
        public bool IsArrowToolButtonVisible
        {
            get { return _isArrowToolButtonVisible; }
            set
            {
                if (value != _isArrowToolButtonVisible)
                {
                    _isArrowToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isPolylineToolButtonVisible = true;
        public bool IsPolylineToolButtonVisible
        {
            get { return _isPolylineToolButtonVisible; }
            set
            {
                if (value != _isPolylineToolButtonVisible)
                {
                    _isPolylineToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isPolygonToolButtonVisible = true;
        public bool IsPolygonToolButtonVisible
        {
            get { return _isPolygonToolButtonVisible; }
            set
            {
                if (value != _isPolygonToolButtonVisible)
                {
                    _isPolygonToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isCloudToolButtonVisible = true;
        public bool IsCloudToolButtonVisible
        {
            get { return _isCloudToolButtonVisible; }
            set
            {
                if (value != _isCloudToolButtonVisible)
                {
                    _isCloudToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        #endregion

        private ToolManager _ToolManager;
        /// <summary>
        /// Gets or sets the ToolManager
        /// </summary>
        internal ToolManager ToolManager
        {
            get { return _ToolManager; }
            set
            {
                if (_ToolManager != null)
                {
                    _ToolManager.NewToolCreated -= _ToolManager_NewToolCreated;

                    if (_ToolManager.UndoRedoAction != null)
                    {
                        _ToolManager.UndoRedoAction.OnUndoRedoStatusChanged -= UndoRedoAction_OnUndoRedoStatusChanged;
                    }
                }
                _ToolManager = value;
                if (_ToolManager != null)
                {
                    _ToolManager.NewToolCreated += _ToolManager_NewToolCreated;

                    if (_ToolManager.UndoRedoAction != null)
                    {
                        _ToolManager.UndoRedoAction.OnUndoRedoStatusChanged += UndoRedoAction_OnUndoRedoStatusChanged;

                        // NOTE: Make sure to update both CanUndo and CanRedo when a new ToolManager is loaded
                        CanUndo = _ToolManager.UndoRedoAction.CanUndo;
                        CanRedo = _ToolManager.UndoRedoAction.CanRedo;
                    }
                }
            }
        }

        private bool _ButtonsStayDown;
        /// <summary>
        /// Gets or sets whether or not the buttons stay down once pressed.
        /// When set to true, the toolbar will not switch back to the pan tool after each 
        /// annotation is drawn.
        /// </summary>
        internal bool ButtonsStayDown
        {
            get { return _ButtonsStayDown; }
            set
            {
                _ButtonsStayDown = value;
                if (_ToolManager != null)
                {
                    _ToolManager.CurrentTool.ReturnToPanModeWhenFinished = !value;
                }
            }
        }

        #region XAML properties

        // Controls whether the regular annotation bar or polygon tool bar is in view
        private bool _showPolygonSaveToolbar = false;
        public bool ShowPolygonSaveToolbar
        {
            get { return _showPolygonSaveToolbar; }
            set
            {
                if (value != _showPolygonSaveToolbar)
                {
                    _showPolygonSaveToolbar = value;
                    RaisePropertyChanged();
                }
            }
        }

        // Controls whether the regular annotation bar or ink bar is in view
        public bool _ShowInkToolbar = false;
        public bool ShowInkToolbar
        {
            get { return _ShowInkToolbar; }
            set
            {
                if (value != _ShowInkToolbar)
                {
                    _ShowInkToolbar = value;
                    RaisePropertyChanged();

                    if (value)
                    {
                        CurrentPresetLabel = Settings.GetPropertyLabel(ToolType.e_ink_create);
                        SetInkPropertiesFromLabel();
                    }
                    else
                    {
                        CurrentPresetLabel = -1;
                    }
                }
            }
        }

        // Control whethers the ink app bar or regular tool bar or polygon tool bar is expanded or compact
        public bool _IsMainAppBarOpen = false;
        public bool IsMainAppBarOpen
        {
            get { return _IsMainAppBarOpen; }
            set
            {
                if (value != _IsMainAppBarOpen)
                {
                    _IsMainAppBarOpen = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _CanUndo = false;
        public bool CanUndo
        {
            get { return _CanUndo; }
            set
            {
                if (value != _CanUndo)
                {
                    _CanUndo = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _CanRedo = false;
        public bool CanRedo
        {
            get { return _CanRedo; }
            set
            {
                if (value != _CanRedo)
                {
                    _CanRedo = value;
                    RaisePropertyChanged();
                }
            }
        }

        private string _UndoToolTip = "Undo";
        public string UndoToolTip
        {
            get { return _UndoToolTip; }
            set
            {
                if (value != _UndoToolTip)
                {
                    _UndoToolTip = value;
                    RaisePropertyChanged();
                }
            }
        }

        private string _RedoToolTip = "Redo";
        public string RedoToolTip
        {
            get { return _RedoToolTip; }
            set
            {
                if (value != _RedoToolTip)
                {
                    _RedoToolTip = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _CanRedoInk = false;
        public bool CanRedoInk
        {
            get { return _CanRedoInk; }
            set
            {
                if (value != _CanRedoInk)
                {
                    _CanRedoInk = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _CanSaveInk = false;
        public bool CanSaveInk
        {
            get { return _CanSaveInk; }
            set
            {
                if (value != _CanSaveInk)
                {
                    _CanSaveInk = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _canSavePolygon = true;
        public bool CanSavePolygon
        {
            get { return _canSavePolygon; }
            set
            {
                if (value != _canSavePolygon)
                {
                    _canSavePolygon = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _GotToolFromToolManager = false;
        private ToolType _CurrentTool = ToolType.e_pan;
        public ToolType CurrentTool
        {
            get { return _CurrentTool; }
            set
            {
                if (_CurrentTool != value)
                {
                    if (_CurrentTool == ToolType.e_ink_create && ShowInkToolbar)
                    {
                        ShowInkToolbar = false;
                    }
                    _CurrentTool = value;
                    if (!_GotToolFromToolManager)
                    {
                        SetCurrentTool();
                    }
                    RaisePropertyChanged();

                    RaisePropertyChanged("PanSelected");
                }
            }
        }

        public bool PanSelected { get { return _CurrentTool == ToolType.e_pan; } }

        private double _AppBarCompactHeight = 48;
        public double AppBarCompactHeight
        {
            get { return _AppBarCompactHeight; }
            set { Set(ref _AppBarCompactHeight, value); }
        }

        #endregion XAML properties

        private FreehandCreate _CurrentInkTool;

        public delegate void ToolSelectedDelegate();
        public event ToolSelectedDelegate ToolSelected;

        /// <summary>
        /// Creates a new AnnotationToolbar that is not associated with a ToolManager.
        /// Use the ToolManager property to make the AnnotationToolbar Work with a ToolManager.
        /// </summary>
        public AnnotationToolbarViewModel(ICloseableControl view)
        {
            _View = view;
            Init();
        }

        /// <summary>
        /// Creates a new AnnotationToolbar that works with the ToolManager.
        /// </summary>
        /// <param name="ctrl"></param>
        /// <param name="toolManager"></param>
        public AnnotationToolbarViewModel(ICloseableControl view, ToolManager toolManager)
        {
            _View = view;
            Init();
            this.ToolManager = toolManager;            
        }

        private void UndoRedoAction_OnUndoRedoStatusChanged(object sender, UndoRedoStatusEventArgs e)
        {
            CanUndo = e.CanUndo;
            CanRedo = e.CanRedo;
            UndoToolTip = e.UndoInfo;
            RedoToolTip = e.RedoInfo;
        }

        private void Init()
        {
            // set up relay commands
            ToolButtonClickedCommand = new RelayCommand(ToolButtonClickedCommandImpl);
            InkToolbarCommand = new RelayCommand(InkToolbarCommandImpl);
            InkPresetButtonCommand = new RelayCommand(InkPresetButtonCommandImpl);
            PolygonSaveToolbarCommand = new RelayCommand(PolygonSaveToolbarCommandImpl);
            UndoButtonClickedCommand = new RelayCommand(UndoButtonCommandImpl);
            RedoButtonClickedCommand = new RelayCommand(RedoButtonCommandImpl);
            ReadyInkPresets();
        }

        public void UndoButtonCommandImpl(object option)
        {
            if ((_ToolManager != null) && (_ToolManager.UndoRedoAction != null))
            {
                _ToolManager.UndoRedoAction.DoUndo();     
                _ToolManager.CreateTool(ToolType.e_pan, _ToolManager.CurrentTool, true);
            }
        }

        public void RedoButtonCommandImpl(object option)
        {
            if ((_ToolManager != null) && (_ToolManager.UndoRedoAction != null))
            {
                _ToolManager.UndoRedoAction.DoRedo();
                _ToolManager.CreateTool(ToolType.e_pan, _ToolManager.CurrentTool, true);
            }
        }

        public void ToolButtonClickedCommandImpl(object option)
        {
            Windows.UI.Xaml.Controls.Button button = option as Windows.UI.Xaml.Controls.Button;
            if (button != null)
            {
                string toolName = button.Tag as string;
                ToolType toolType = GetToolTypeFromString(toolName);
                if (CurrentTool == toolType)
                {
                    if (Settings.HasSettableProperties(CurrentTool) && CurrentTool != ToolType.e_ink_create)
                    {
                        ShowPopup(button);
                    }
                }
                else
                {
                    if ((CurrentTool == ToolType.e_polygon_create) || (CurrentTool == ToolType.e_polyline_create)
                        || (CurrentTool == ToolType.e_cloud_create))
                    {
                        CommitPolygonTools();
                    }
                    CurrentTool = toolType;
                    SetCurrentTool();
                }

                if (ToolSelected != null)
                {
                    ToolSelected();
                    IsMainAppBarOpen = false;
                }
            }
        }

        void _ToolManager_NewToolCreated(ToolType toolType)
        {
            if (toolType == _CurrentTool)
            {
                return;
            }
            _GotToolFromToolManager = true;
            CurrentTool = toolType;
            _GotToolFromToolManager = false;
        }

        ToolType FilterOutInvalidTools(ToolType toolType)
        {
            if (toolType == ToolType.e_annot_edit || toolType == ToolType.e_line_edit
                || toolType == ToolType.e_link_action || toolType == ToolType.e_none || toolType == ToolType.e_rich_media
                || toolType == ToolType.e_text_select)
            {
                return ToolType.e_pan;
            }
            return toolType;
        }

        #region Polygon Save Toolbar

        private void PolygonSaveToolbarCommandImpl(Object option)
        {
            CommitPolygonTools();
            ShowPolygonSaveToolbar = false;
            IsMainAppBarOpen = false;
            ToolManager.CreateTool(ToolType.e_pan, null);
            Utilities.AnalyticsHandlerBase.CURRENT.SendEvent("Annotation Toolbar", "Exit shape tools");
        }

        private void CommitPolygonTools()
        {
            if (_CurrentTool == ToolType.e_polyline_create)
            {
                PolylineCreate polylineTool = this.ToolManager.CurrentTool as PolylineCreate;
                if (polylineTool != null)
                {
                    polylineTool.CommitAnnotation();
                }
            }
            else if (_CurrentTool == ToolType.e_polygon_create)
            {
                PolygonCreate polygonTool = this.ToolManager.CurrentTool as PolygonCreate;
                if (polygonTool != null)
                {
                    polygonTool.CommitAnnotation();
                }
            }
            else if (_CurrentTool == ToolType.e_cloud_create)
            {
                CloudCreate cloudTool = this.ToolManager.CurrentTool as CloudCreate;
                if (cloudTool != null)
                {
                    cloudTool.CommitAnnotation();
                }
            }
        }

        #endregion

        #region InkToolbar

        /// <summary>
        /// Perform Freehand/Ink Undo
        /// </summary>
        public void PerformInkUndo()
        {
            if (_CurrentInkTool.CanUndoStroke())
                InkToolbarCommandImpl("U");
        }

        /// <summary>
        /// Perform Freehand/Ink Redo
        /// </summary>
        public void PerformInkRedo()
        {
            if (_CurrentInkTool.CanRedoStroke())
                InkToolbarCommandImpl("R");
        }

        private void CurrentInkTool_StrokeAdded()
        {
            CanSaveInk = true;
            CanRedoInk = false;
        }

        private void InkToolbarCommandImpl(Object option)
        {
            string opt = option as string;
            if (opt.Equals("S", StringComparison.OrdinalIgnoreCase))
            {
                _CurrentInkTool.CommitAnnotation();
                _ToolManager.CreateTool(ToolType.e_pan, null);
                ShowInkToolbar = false;

                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent("Annotation Toolbar", "Freehand Save button selected");
            }
            else if (opt.Equals("U", StringComparison.OrdinalIgnoreCase))
            {
                _CurrentInkTool.UndoStroke();
                if (!_CurrentInkTool.CanUndoStroke())
                {
                    CanSaveInk = false;
                }
                CanRedoInk = true;

                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent("Annotation Toolbar", "Freehand Undo button selected");
            }
            else if (opt.Equals("R", StringComparison.OrdinalIgnoreCase))
            {
                _CurrentInkTool.RedoStroke();
                if (!_CurrentInkTool.CanRedoStroke())
                {
                    CanRedoInk = false;
                }
                CanSaveInk = true;

                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent("Annotation Toolbar", "Freehand Redo button selected");
            }
            else
            {
                ShowInkToolbar = false;
                IsMainAppBarOpen = false;
                ToolManager.CreateTool(ToolType.e_pan, null);
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent("Annotation Toolbar", "Freehand Cancel button selected");
            }
        }

        private void ReadyInkPresets()
        {
            int label = Settings.GetPropertyLabel(ToolType.e_ink_create);
            _CurrentPresetLabel = label; // Don't want to edit the general ink when we initialize
            for (int i = 0; i < 3; i++)
            {
                SettingsColor color = Settings.GetShapeColor(ToolType.e_ink_create, false, i);
                _PresetBrushes[i] = new SolidColorBrush(UtilityFunctions.ConvertToColor(color));
            }
        }

        private int _CurrentPresetLabel = -1;
        public int CurrentPresetLabel
        {
            get { return _CurrentPresetLabel; }
            set
            {
                if (value != _CurrentPresetLabel)
                {
                    _CurrentPresetLabel = value;
                    if (_CurrentPresetLabel < 100)
                    {
                        Settings.SetPropertylLabel(ToolType.e_ink_create, _CurrentPresetLabel);
                        SetInkPropertiesFromLabel();
                        _CurrentInkTool.IsEraser = false;
                    }
                    else
                    {
                        Settings.SetPropertylLabel(ToolType.e_ink_eraser, _CurrentPresetLabel);
                        SetEraserProperties();
                    }
                    RaisePropertyChanged();
                }
            }
        }

        private void SetInkPropertiesFromLabel()
        {
            SettingsColor color = Settings.GetShapeColor(ToolType.e_ink_create, false, _CurrentPresetLabel);
            Settings.SetShapeColor(ToolType.e_ink_create, false, color);
            switch (CurrentPresetLabel)
            {
                case 0:
                    PresetBrush0 = new SolidColorBrush(UtilityFunctions.ConvertToColor(color));
                    break;
                case 1:
                    PresetBrush1 = new SolidColorBrush(UtilityFunctions.ConvertToColor(color));
                    break;
                case 2:
                    PresetBrush2 = new SolidColorBrush(UtilityFunctions.ConvertToColor(color));
                    break;
            }
            double thickness = Settings.GetThickness(ToolType.e_ink_create, _CurrentPresetLabel);
            Settings.SetThickness(ToolType.e_ink_create, thickness);
            double opacity = Settings.GetOpacity(ToolType.e_ink_create, _CurrentPresetLabel);
            Settings.SetOpacity(ToolType.e_ink_create, opacity);

            if (_CurrentInkTool != null)
            {
                _CurrentInkTool.SetInkProperties(thickness, opacity, UtilityFunctions.ConvertToColor(color));
            }
        }

        private void SetEraserProperties()
        {
            if (CurrentPresetLabel >= 100 && _CurrentInkTool != null)
            {
                double thickness = Settings.GetThickness(ToolType.e_ink_eraser, CurrentPresetLabel);
                Windows.UI.Color eraserColor = FreehandCreate.DEFAULT_INK_ERASER_COLOR; ;
                double opacity = FreehandCreate.DEFAULT_INK_ERASER_OPACITY;
                _CurrentInkTool.SetEraserProperties(true, thickness, opacity, eraserColor);
                _CurrentInkTool.IsEraser = true;
            }
        }

        private SolidColorBrush[] _PresetBrushes = new SolidColorBrush[3];
        public SolidColorBrush PresetBrush0
        {
            get { 
                return _PresetBrushes[0]; 
            }
            private set
            {
                _PresetBrushes[0] = value;
                RaisePropertyChanged();
            }
        }
        public SolidColorBrush PresetBrush1
        {
            get { return _PresetBrushes[1]; }
            private set
            {
                _PresetBrushes[1] = value;
                RaisePropertyChanged();
            }
        }
        public SolidColorBrush PresetBrush2
        {
            get { return _PresetBrushes[2]; }
            private set
            {
                _PresetBrushes[2] = value;
                RaisePropertyChanged();
            }
        }

        private void InkPresetButtonCommandImpl(object obj)
        {
            Button button = obj as Button;
            if (button != null)
            {
                string tag = button.Tag as string;
                if (!string.IsNullOrEmpty(tag))
                {
                    int result = -1;
                    bool success = int.TryParse(tag, out result);
                    if (success)
                    {
                        if (result == CurrentPresetLabel)
                        {
                            ShowPopup(button);
                        }
                        else
                        {
                            CurrentPresetLabel = result;
                        }
                    }
                }
            }
        }

        #endregion InkToolbar

        #region Utilities

        public static ToolType GetToolTypeFromString(string option)
        {
            if (option.Equals("Pan", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_pan;
            }
            if (option.Equals("Line", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_line_create;
            }
            if (option.Equals("Arrow", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_arrow_create;
            }
            if (option.Equals("Square", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_rect_create;
            }
            if (option.Equals("Oval", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_oval_create;
            }
            if (option.Equals("Ink", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_ink_create;
            }
            if (option.Equals("Eraser", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_ink_eraser;
            }
            if (option.Equals("Marker", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_ink_marker;
            }
            if (option.Equals("Signature", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_signature;
            }
            if (option.Equals("Stamper", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_stamper;
            }
            if (option.Equals("TextBox", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_text_annot_create;
            }
            if (option.Equals("Sticky", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_sticky_note_create;
            }
            if (option.Equals("Highlight", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_text_highlight;
            }
            if (option.Equals("Underline", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_text_underline;
            }
            if (option.Equals("Strikeout", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_text_strikeout;
            }
            if (option.Equals("Squiggly", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_text_squiggly;
            }
            if (option.Equals("Cloud", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_cloud_create;
            }
            if (option.Equals("Polygon", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_polygon_create;
            }
            if (option.Equals("Polyline", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_polyline_create;
            }
            if (option.Equals("annotation group", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_annot_edit_rect_group;
            }
            if (option.Equals("annotation group create", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_annot_edit_group_create;
            }
            if (option.Equals("None", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_none;
            }
            return ToolType.e_none;
        }

        private static string GetPropertyName<T>(System.Linq.Expressions.Expression<Func<T>> propertyExpression)
        {
            return (propertyExpression.Body as System.Linq.Expressions.MemberExpression).Member.Name;
        }

        private void SetCurrentTool()
        {
            if (_ToolManager == null)
            {
                return;
            }

            if (CurrentTool != ToolType.e_ink_create)
            {
                ShowInkToolbar = false;
            }
            bool shouldButtonStayDown = this.ButtonsStayDown;
            _ToolManager.CreateTool(CurrentTool, null, ButtonsStayDown);
            switch (CurrentTool)
            {
                case ToolType.e_annot_edit_rect_group:
                case ToolType.e_annot_edit_group_create:
                case ToolType.e_pan:
                    shouldButtonStayDown = false;
                    break;
                case ToolType.e_ink_create:
                    IsMainAppBarOpen = false;
                    shouldButtonStayDown = false;
                    _CurrentInkTool = _ToolManager.CurrentTool as FreehandCreate;
                    if (_CurrentInkTool != null)
                    {
                        _CurrentInkTool.MultiStrokeMode = true;
                        _CurrentInkTool.StrokeAdded += CurrentInkTool_StrokeAdded;
                        CanSaveInk = false;
                        CanRedoInk = false;
                        ShowInkToolbar = true;
                        _CurrentInkTool.RetainInkWhenScalingOrSizing = true;
                    }
                    break;
                case ToolType.e_polyline_create:
                case ToolType.e_cloud_create:
                case ToolType.e_polygon_create:
                    IsMainAppBarOpen = false;
                    shouldButtonStayDown = false;                    
                    ShowPolygonSaveToolbar = true;
                    break;
                case ToolType.e_ink_marker:
                    _CurrentInkTool = _ToolManager.CurrentTool as FreehandCreate;
                    if (_CurrentInkTool != null)
                    {
                        _CurrentInkTool.MultiStrokeMode = false;
                    }
                    break;
                case ToolType.e_signature:
                    shouldButtonStayDown = false;
                    break;
               case ToolType.e_stamper:
                    shouldButtonStayDown = false;
                    break;
                case ToolType.e_none:
                    IsMainAppBarOpen = false;
                    _View.CloseControl();
                    _ToolManager.CreateTool(ToolType.e_pan, null);
                    break;
            }
        }

        #endregion Utilities


        #region Annotation Properties


        private Popup _Popup;
        private FrameworkElement _Target;
        private AnnotationPropertyDialog _Dialog;
        private void ShowPopup(FrameworkElement target)
        {
            _Target = target;

            _Dialog = new AnnotationPropertyDialog();
            UIRect rect = UtilityFunctions.GetElementRect(_Target);
            if (rect.Height > AppBarCompactHeight)
            {
                rect.Height = AppBarCompactHeight;
            }

            if (CurrentTool == ToolType.e_ink_create)
            {
                _Dialog.ViewModel.Sequencenumber = _CurrentPresetLabel;
            }
            if (_CurrentPresetLabel >= 100)
            {
                _Dialog.ViewModel.AssociatedTool = ToolType.e_ink_eraser;
            }
            else
            {
                _Dialog.ViewModel.AssociatedTool = CurrentTool;
            }

            AnnotationPropertiesDisplay display = AnnotationPropertiesDisplay.CreateAnnotationPropertiesDisplay(rect, _Dialog);
            display.AnnotationPropertyDialogClosed += display_AnnotationPropertyDialogClosed;
        }

        void display_AnnotationPropertyDialogClosed(AnnotationPropertyDialog dialog)
        {
            if (CurrentTool == ToolType.e_ink_create)
            {
                if (_CurrentPresetLabel >= 100)
                {
                    SetEraserProperties();
                }
                else
                {
                    SetInkPropertiesFromLabel();
                }
            }
        }

        void _Popup_Closed(object sender, object e)
        {
            _Popup = null;
            _Dialog = null;

            if (CurrentTool == ToolType.e_ink_create)
            {
                SetInkPropertiesFromLabel();
            }
        }

        #endregion Annotation Properties

        #region Public Interface

        public bool GoBack()
        {
            if (_Popup != null)
            {
                _Popup.IsOpen = false;
                return true;
            }
            if (ShowInkToolbar)
            {
                ShowInkToolbar = false;
                ToolManager.CreateDefaultTool();
                return true;
            }
            if (ShowPolygonSaveToolbar)
            {
                ShowPolygonSaveToolbar = false;
                ToolManager.CreateDefaultTool();
                return true;
            }

            
            return false;
        }


        #endregion Public Interface
    }
}
