﻿using pdftron.Common;
using pdftron.PDF.Tools.Controls.ControlBase;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Runtime.InteropServices.WindowsRuntime;
using Windows.Graphics.Imaging;
using Windows.Storage.Streams;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Media.Imaging;
using System.Threading;
using Windows.UI.Core;

namespace pdftron.PDF.Tools.Controls.ViewModels
{
    public class CropItem : ViewModelBase
    {
        private ImageSource _Source;
        public ImageSource Source
        {
            get { return _Source; }
            set { Set(ref _Source, value); }
        }

        private double _Width;
        public double Width
        {
            get { return _Width; }
            set { Set(ref _Width, value); }

        }

        private double _Height;
        public double Height
        {
            get { return _Height; }
            set { Set(ref _Height, value); }
        }

        private int _Page;
        public int Page
        {
            get { return _Page; }
            set { Set(ref _Page, value); }
        }

        public CropItem(ImageSource source, double width, double height, int page) : this(page, width, height)
        {
            Source = source;
        }

        public CropItem(int page, double width, double height)
        {
            Page = page;
            Width = width;
            Height = height;
        }
    }

    public class CropViewViewModel : ViewModelBase
    {
        public const double RECT_COMPARE_THRESHOLD = 2.0;

        private ICloseableControl _View;

        private PDFViewCtrl _PDFViewCtrl;

        private RasterizerFactory _RasterizerFactory;

        private SemaphoreSlim _CroppingSempaphore;

        public enum MultiplePageCrop
        {
            None,
            All,
            Even,
            Odd,
        }

        private MultiplePageCrop _CurrMultiplePageCrop;

        public MultiplePageCrop CurrMultiplePageCrop
        {
            get { return _CurrMultiplePageCrop; }
            set { Set(ref _CurrMultiplePageCrop, value); }
        }

        private bool _ShowMultiplePageCropMessage;
        public bool ShowMultiplePageCropMessage
        {
            get { return _ShowMultiplePageCropMessage; }
            set { Set(ref _ShowMultiplePageCropMessage, value); }
        }

        private bool _HasRasterizedInitialPage = false;

        public delegate void DocumentEditedDelegate();
        public event DocumentEditedDelegate DocumentEdited;

        private class PageCropProperties
        {
            public Rect UserCropBoxPageSpace;
            public Rect CropBoxPageSpace;
            public Rect CropRatioRect
            {
                get
                {
                    if (UserCropBoxPageSpace != null && CropBoxPageSpace != null)
                    {
                        return CalculcateScreenRatioRect();
                    }
                    return null;
                }
                set
                {
                    if (UserCropBoxPageSpace != null && CropBoxPageSpace != null)
                    {
                        CalculateNewUserCropBox(value);
                    }
                }
            }
            public int Rotation;

            public Rect CalculcateScreenRatioRect()
            {
                Rect ratioRect = new Rect();
                switch (Rotation)
                {
                    case 1:
                        ratioRect.x1 = (UserCropBoxPageSpace.x1 - CropBoxPageSpace.x1) / CropBoxPageSpace.Width();
                        ratioRect.x2 = (CropBoxPageSpace.x2 - UserCropBoxPageSpace.x2) / CropBoxPageSpace.Width();
                        ratioRect.y1 = (UserCropBoxPageSpace.y2 - CropBoxPageSpace.y2) / CropBoxPageSpace.Height();
                        ratioRect.y2 = (CropBoxPageSpace.y1 - UserCropBoxPageSpace.y1) / CropBoxPageSpace.Height();
                        break;
                    case 2:
                        ratioRect.x1 = (CropBoxPageSpace.x1 - UserCropBoxPageSpace.x1) / CropBoxPageSpace.Width();
                        ratioRect.x2 = (UserCropBoxPageSpace.x2 - CropBoxPageSpace.x2) / CropBoxPageSpace.Width();
                        ratioRect.y1 = (UserCropBoxPageSpace.y2 - CropBoxPageSpace.y2) / CropBoxPageSpace.Height();
                        ratioRect.y2 = (CropBoxPageSpace.y1 - UserCropBoxPageSpace.y1) / CropBoxPageSpace.Height();
                        break;
                    case 3:
                        ratioRect.x1 = (CropBoxPageSpace.x1 - UserCropBoxPageSpace.x1) / CropBoxPageSpace.Width();
                        ratioRect.x2 = (UserCropBoxPageSpace.x2 - CropBoxPageSpace.x2) / CropBoxPageSpace.Width();
                        ratioRect.y1 = (CropBoxPageSpace.y2 - UserCropBoxPageSpace.y2) / CropBoxPageSpace.Height();
                        ratioRect.y2 = (UserCropBoxPageSpace.y1 - CropBoxPageSpace.y1) / CropBoxPageSpace.Height();
                        break;
                    default:
                        ratioRect.x1 = (UserCropBoxPageSpace.x1 - CropBoxPageSpace.x1) / CropBoxPageSpace.Width();
                        ratioRect.x2 = (CropBoxPageSpace.x2 - UserCropBoxPageSpace.x2) / CropBoxPageSpace.Width();
                        ratioRect.y1 = (CropBoxPageSpace.y2 - UserCropBoxPageSpace.y2) / CropBoxPageSpace.Height();
                        ratioRect.y2 = (UserCropBoxPageSpace.y1 - CropBoxPageSpace.y1) / CropBoxPageSpace.Height();
                        break;
                }
                return ratioRect;
            }

            public void CalculateNewUserCropBox(Rect cropRatio)
            {
                switch (Rotation)
                {
                    case 1:
                        UserCropBoxPageSpace.x1 = CropBoxPageSpace.x1 + cropRatio.x1 * CropBoxPageSpace.Width();
                        UserCropBoxPageSpace.x2 = CropBoxPageSpace.x2 - cropRatio.x2 * CropBoxPageSpace.Width();
                        UserCropBoxPageSpace.y1 = CropBoxPageSpace.y1 - cropRatio.y2 * CropBoxPageSpace.Height();
                        UserCropBoxPageSpace.y2 = CropBoxPageSpace.y2 + cropRatio.y1 * CropBoxPageSpace.Height();
                        break;
                    case 2:
                        UserCropBoxPageSpace.x1 = CropBoxPageSpace.x1 - cropRatio.x1 * CropBoxPageSpace.Width();
                        UserCropBoxPageSpace.x2 = CropBoxPageSpace.x2 + cropRatio.x2 * CropBoxPageSpace.Width();
                        UserCropBoxPageSpace.y1 = CropBoxPageSpace.y1 - cropRatio.y2 * CropBoxPageSpace.Height();
                        UserCropBoxPageSpace.y2 = CropBoxPageSpace.y2 + cropRatio.y1 * CropBoxPageSpace.Height();
                        break;
                    case 3:
                        UserCropBoxPageSpace.x1 = CropBoxPageSpace.x1 - cropRatio.x1 * CropBoxPageSpace.Width();
                        UserCropBoxPageSpace.x2 = CropBoxPageSpace.x2 + cropRatio.x2 * CropBoxPageSpace.Width();
                        UserCropBoxPageSpace.y1 = CropBoxPageSpace.y1 + cropRatio.y2 * CropBoxPageSpace.Height();
                        UserCropBoxPageSpace.y2 = CropBoxPageSpace.y2 - cropRatio.y1 * CropBoxPageSpace.Height();
                        break;
                    default:
                        UserCropBoxPageSpace.x1 = CropBoxPageSpace.x1 + cropRatio.x1 * CropBoxPageSpace.Width();
                        UserCropBoxPageSpace.x2 = CropBoxPageSpace.x2 - cropRatio.x2 * CropBoxPageSpace.Width();
                        UserCropBoxPageSpace.y1 = CropBoxPageSpace.y1 + cropRatio.y2 * CropBoxPageSpace.Height();
                        UserCropBoxPageSpace.y2 = CropBoxPageSpace.y2 - cropRatio.y1 * CropBoxPageSpace.Height();
                        break;
                }
            }
        }

        private PageCropProperties[] _PageProperties;

        private Rect _CropRatioRectSceenSpace = new Rect();

        private CropItem _CurrCropItem;
        public CropItem CurrCropItem
        {
            get { return _CurrCropItem; }
            set
            {
                if (Set(ref _CurrCropItem, value))
                {
                    RaisePropertyChanged("HasImage");
                }
            }
        }

        public Rect CurrCropRatioRect
        {
            get
            {
                if (CurrCropItem != null)
                {
                    PageCropProperties props = GetOrMakePageProperties(CurrCropItem.Page);
                    return props.CropRatioRect;
                }
                return null;
            }
            set
            {
                if (CurrCropItem != null)
                {
                    PageCropProperties props = GetOrMakePageProperties(CurrCropItem.Page);
                    props.CropRatioRect = value;
                }
            }
        }

        public bool HasImage { get { return CurrCropItem != null && CurrCropItem.Source != null; } }

        public bool IsPageBackEnabled
        {
            get { return CurrCropItem != null && CurrCropItem.Page > 1; }
        }

        public bool IsPageForwardEnabled
        {
            get { return CurrCropItem != null && _PDFViewCtrl != null && CurrCropItem.Page < _PDFViewCtrl.GetPageCount(); }
        }

        public bool IsApplyOddPagesEnabled
        {
            get { return CurrCropItem != null && CurrCropItem.Page % 2 != 0; }
        }

        public CropViewViewModel(ICloseableControl view, PDFViewCtrl pdfViewCtrl)
        {
            _View = view;
            _PDFViewCtrl = pdfViewCtrl;
            Init();
        }

        public void CleanUp()
        {
            _RasterizerFactory.ShutDown();
        }

        public void RasterizeInitialPage()
        {
            if (!_HasRasterizedInitialPage)
            {
                SetCropItemFromPage(_PDFViewCtrl.GetCurrentPage());
                _HasRasterizedInitialPage = true;
            }
        }

        private void Init()
        {
            CloseCommand = new RelayCommand(CloseCommandImpl);
            PageBackCommand = new RelayCommand(PageBackCommandImpl);
            PageForwardCommand = new RelayCommand(PageForwardCommandImpl);
            ApplyCropCommand = new RelayCommand(ApplyCropCommandImpl);
            MultiplePagesCropCommand = new RelayCommand(MultiplePagesCropCommandImpl);

            _CroppingSempaphore = new SemaphoreSlim(1);

            _RasterizerFactory = new RasterizerFactory(_PDFViewCtrl.GetDoc(), _PDFViewCtrl.GetPageCount());
            _RasterizerFactory.BitmapAvailable += RasterizerFactory_BitmapAvailable;

            _PageProperties = new PageCropProperties[_PDFViewCtrl.GetPageCount()];
        }

        public void SetColorPostProcessing(PDFRasterizerColorPostProcessMode mode, Windows.UI.Color? background, Windows.UI.Color? foreground)
        {
            _RasterizerFactory.SetColorPostProcessing(mode, background, foreground);
        }

        public RelayCommand CloseCommand { get; private set; }
        public RelayCommand PageBackCommand { get; private set; }
        public RelayCommand PageForwardCommand { get; private set; }
        public RelayCommand ApplyCropCommand { get; private set; }
        public RelayCommand MultiplePagesCropCommand { get; private set; }


        private void CloseCommandImpl(object command)
        {
            _View.CloseControl();
        }
        private void PageBackCommandImpl(object command)
        {
            SetCropItemFromPage(CurrCropItem.Page - 1);
        }
        private void PageForwardCommandImpl(object command)
        {
            SetCropItemFromPage(CurrCropItem.Page + 1);
        }

        private async void MultiplePagesCropCommandImpl(object command)
        {
            string val = command.ToString();
            bool handled = false;
            if (string.Equals(val, "all", StringComparison.OrdinalIgnoreCase))
            {
                handled = true;
                CurrMultiplePageCrop = MultiplePageCrop.All;
            }
            if (string.Equals(val, "even", StringComparison.OrdinalIgnoreCase))
            {
                handled = true;
                CurrMultiplePageCrop = MultiplePageCrop.Even;
            }
            if (string.Equals(val, "odd", StringComparison.OrdinalIgnoreCase))
            {
                handled = true;
                CurrMultiplePageCrop = MultiplePageCrop.Odd;
            }

            MultiplePageCrop mode = CurrMultiplePageCrop;
            _CropRatioRectSceenSpace = CurrCropRatioRect;

            if (handled)
            {
                await Task.Run(() => 
                {
                    _CroppingSempaphore.Wait();
                    SetPageCropPropertiesToMultiplePages();
                    _CroppingSempaphore.Release();
                });

                ShowMultiplePageCropMessage = true;
            }

            if (handled)
            {
                await Task.Delay(4000);

                // if user changes mode again before it gets here, don't hide it
                if (mode == CurrMultiplePageCrop)
                {
                    ShowMultiplePageCropMessage = false;
                }
            }
        }

        private async void ApplyCropCommandImpl(object command)
        {
            await Windows.ApplicationModel.Core.CoreApplication.MainView.CoreWindow.Dispatcher.RunAsync(Windows.UI.Core.CoreDispatcherPriority.High, () =>
            {
                _CroppingSempaphore.Wait();
                ApplyCropToMultiplePages();
                _CroppingSempaphore.Release();
            });
        }

        private void ApplyCropToMultiplePages()
        {
            PDFDoc pdfDoc = _PDFViewCtrl.GetDoc();
            bool hasEditedDoc = false;

            try
            {
                pdfDoc.LockRead();
                PageIterator iter = pdfDoc.GetPageIterator();
                int pageNum = 1;
                while (iter.HasNext())
                {
                    Page page = iter.Current();
                    pdftron.PDF.Rect oldUserCrop = page.GetBox(PageBox.e_user_crop);

                    if (_PageProperties[pageNum - 1] != null && !AreRectsSimilar(_PageProperties[pageNum - 1].UserCropBoxPageSpace, oldUserCrop, RECT_COMPARE_THRESHOLD))
                    {
                        if (AreRectsSimilar(_PageProperties[pageNum - 1].UserCropBoxPageSpace, _PageProperties[pageNum - 1].CropBoxPageSpace, RECT_COMPARE_THRESHOLD))
                        {
                            hasEditedDoc |= RemoveUserCropFromPage(page);
                        }
                        else
                        {
                            // at this point, the crop properties should have been set before comming here
                            UnRotateRectangle(_PageProperties[pageNum - 1].UserCropBoxPageSpace, _PageProperties[pageNum - 1].Rotation);
                            page.SetBox(PageBox.e_user_crop, _PageProperties[pageNum - 1].UserCropBoxPageSpace);
                            hasEditedDoc = true;
                        }
                    }

                    pageNum++;
                    iter.Next();
                }
            }
            catch (Exception e)
            {
                PDFNetException pdfe = new PDFNetException(e.HResult);
                if (!pdfe.IsPDFNetException)
                {
                    throw;
                }
                else
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(e));
                }
            }
            finally
            {
                pdfDoc.UnlockRead();
                _View.CloseControl();

                if (hasEditedDoc)
                {
                    DocumentEdited?.Invoke();
                }
            }
        }

        private void ApplyCropToCurrentPage()
        {
            int pageNum = CurrCropItem.Page;
            PDFDoc pdfDoc = _PDFViewCtrl.GetDoc();

            try
            {
                pdfDoc.LockRead();

                Page page = pdfDoc.GetPage(pageNum);

                // this should not happen, since it will be populated when arriving to this page
                if (_PageProperties[pageNum - 1] == null)
                    return;

                PageCropProperties cropProperties = _PageProperties[pageNum - 1];
                Rect userCropBox = cropProperties.UserCropBoxPageSpace;

                if (userCropBox != null && userCropBox.Width() > 0 && userCropBox.Height() > 0)
                {
                    UnRotateRectangle(userCropBox, cropProperties.Rotation);
                    page.SetBox(PageBox.e_user_crop, userCropBox);
                }
            }
            catch (Exception e)
            {
                PDFNetException pdfe = new PDFNetException(e.HResult);
                if (!pdfe.IsPDFNetException)
                {
                    throw;
                }
                else
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(e));
                }
            }
            finally
            {
                pdfDoc.UnlockRead();
                _View.CloseControl();
            }
        }

        private void SetPageCropPropertiesToMultiplePages()
        {
            PDFDoc pdfDoc = _PDFViewCtrl.GetDoc();

            try
            {
                pdfDoc.LockRead();
                PageIterator iter = pdfDoc.GetPageIterator();
                int pageNum = 1;
                while (iter.HasNext())
                {
                    Page page = iter.Current();

                    // skip page if mode is even and current page is odd and vice versa
                    if ((CurrMultiplePageCrop == MultiplePageCrop.Even && pageNum % 2 != 0) ||
                        (CurrMultiplePageCrop == MultiplePageCrop.Odd && pageNum % 2 == 0))
                    {
                        pageNum++;
                        iter.Next();
                        continue;
                    }

                    PageCropProperties pops = GetOrMakePageProperties(pageNum, page);

                    pops.CalculateNewUserCropBox(_CropRatioRectSceenSpace);

                    pageNum++;
                    iter.Next();
                }
            }
            catch (Exception e)
            {
                PDFNetException pdfe = new PDFNetException(e.HResult);
                if (!pdfe.IsPDFNetException)
                {
                    throw;
                }
                else
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(e));
                }
            }
            finally
            {
                pdfDoc.UnlockRead();
            }
        }

        private void UpdatePageNavigationEnabledStatus()
        {
            RaisePropertyChanged("IsPageBackEnabled");
            RaisePropertyChanged("IsPageForwardEnabled");
            RaisePropertyChanged("IsApplyOddPagesEnabled");
        }

        private void SetCropItemFromPage(int page)
        {
            if (page == 0 || page > _PDFViewCtrl.GetPageCount())
                return;

            PageCropProperties cropProperties = GetOrMakePageProperties(page);

            Rect userCropBox = _PageProperties[page - 1].UserCropBoxPageSpace;
            Rect cropBox = _PageProperties[page - 1].CropBoxPageSpace;
            int rotation = _PageProperties[page - 1].Rotation;

            // set the CropRatioRect for the image
            _CropRatioRectSceenSpace = cropProperties.CropRatioRect;

            CurrCropItem = new CropItem(page, cropProperties.CropBoxPageSpace.Width(), cropProperties.CropBoxPageSpace.Height());

            UpdatePageNavigationEnabledStatus();

            if (_RasterizerFactory.HasImage(page))
            {
                CurrCropItem.Source = _RasterizerFactory.GetImage(page);
                RaisePropertyChanged("HasImage");
            }
            _RasterizerFactory.SwitchToPage(page);
        }
        private void RasterizerFactory_BitmapAvailable(int pageNumber)
        {
            if (_RasterizerFactory.HasImage(pageNumber))
            {
                if (CurrCropItem != null && CurrCropItem.Page == pageNumber)
                {
                    CurrCropItem.Source = _RasterizerFactory.GetImage(pageNumber);
                    RaisePropertyChanged("HasImage");
                }
            }
        }

        private void RotateRectangle(Rect rect, int rotation)
        {
            double temp;
            switch (rotation)
            {
                case 0:
                    return;
                case 1:
                    temp = rect.x1;
                    rect.x1 = rect.y1;
                    rect.y1 = rect.x2;
                    rect.x2 = rect.y2;
                    rect.y2 = temp;
                    return;
                case 2:
                    temp = rect.x1;
                    rect.x1 = rect.x2;
                    rect.x2 = temp;
                    temp = rect.y1;
                    rect.y1 = rect.y2;
                    rect.y2 = temp;
                    return;
                case 3:
                    temp = rect.x1;
                    rect.x1 = rect.y2;
                    rect.y2 = rect.x2;
                    rect.x2 = rect.y1;
                    rect.y1 = temp;
                    return;
            }
        }

        private void UnRotateRectangle(Rect rect, int rotation)
        {
            double x1 = rect.x1;
            double y1 = rect.y1;
            double x2 = rect.x2;
            double y2 = rect.y2;

            switch (rotation)
            {
                case 0:
                    return;
                case 1:
                    rect.x1 = y2;
                    rect.y1 = x1;
                    rect.x2 = y1;
                    rect.y2 = x2;
                    return;
                case 2:
                    rect.x1 = x2;
                    rect.x2 = x1;
                    rect.y1 = y2;
                    rect.y2 = y1;
                    return;
                case 3:
                    rect.x1 = y1;
                    rect.y2 = x1;
                    rect.x2 = y2;
                    rect.y1 = x2;
                    return;
            }
        }

        private PageCropProperties GetOrMakePageProperties(int pageNumber)
        {
            if (_PageProperties[pageNumber - 1] == null)
            {
                _PDFViewCtrl.DocLockRead();
                try
                {
                    Page pdfPage = _PDFViewCtrl.GetDoc().GetPage(pageNumber);
                    SetPageProperties(pdfPage, pageNumber);
                }
                catch (Exception e)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(e));
                }
                finally
                {
                    _PDFViewCtrl.DocUnlockRead();
                }
            }
            return _PageProperties[pageNumber - 1];
        }

        private PageCropProperties GetOrMakePageProperties(int pageNumber, Page pdfPage)
        {
            if (_PageProperties[pageNumber - 1] == null)
            {
                SetPageProperties(pdfPage, pageNumber);
            }
            return _PageProperties[pageNumber - 1];
        }

        private void SetPageProperties(Page pdfPage, int page)
        {
            _PageProperties[page - 1] = new PageCropProperties();
            _PageProperties[page - 1].CropBoxPageSpace = pdfPage.GetBox(PageBox.e_crop); ;
            _PageProperties[page - 1].UserCropBoxPageSpace = pdfPage.GetBox(PageBox.e_user_crop);
            _PageProperties[page - 1].Rotation = (int)pdfPage.GetRotation();
            RotateRectangle(_PageProperties[page - 1].UserCropBoxPageSpace, _PageProperties[page - 1].Rotation);
            RotateRectangle(_PageProperties[page - 1].CropBoxPageSpace, _PageProperties[page - 1].Rotation);
        }

        public static bool AreRectsSimilar(pdftron.PDF.Rect rect1, pdftron.PDF.Rect rect2, double thresh)
        {
            return Math.Abs(rect1.x1 - rect2.x1) <= thresh
                        && Math.Abs(rect1.y1 - rect2.y1) <= thresh
                        && Math.Abs(rect1.x2 - rect2.x2) <= thresh
                        && Math.Abs(rect1.y2 - rect2.y2) <= thresh;
        }

        private const String CROPPING_OBJECT_STRING = "TRN_UserCrop";
        public static bool RemoveUserCropFromPage(Page page)
        {
            pdftron.SDF.Obj pageRootObj = page.GetSDFObj();
            if (pageRootObj.FindObj(CROPPING_OBJECT_STRING) != null)
            {
                pageRootObj.Erase(CROPPING_OBJECT_STRING);
                return true;
            }
            return false;
        }

        public bool GoBack()
        {
            _View.CloseControl();
            return true;
        }


        #region Bitmaps
        
        private class RasterizerFactory
        {
            private class ImageProperties
            {
                public int Width;
                public int Height;
                public ByteArrayRef ImageBuffer;
                public WriteableBitmap Bitmap;
            }


            public delegate void BitmapAvailableDelegate(int pageNumber);
            public event BitmapAvailableDelegate BitmapAvailable;

            private PDFDoc _Doc;

            private Dictionary<int, ImageProperties> _Images;
            private int _TotalPages;
            private int _CurrentPage;

            private int _CurrentJobPage;
            private PDFRasterizer _Rasterizer;
            private bool _IsWorking = false;

            private const int MAX_RENDERED_IMAGES_PER_DIRECTION = 4;

            private bool _Destroyed = false;

            public RasterizerFactory(PDFDoc doc, int totalPages)
            {
                _Doc = doc;
                _TotalPages = totalPages;
                _CurrentPage = 1;
                _CurrentJobPage = 0;

                _Images = new Dictionary<int, ImageProperties>();
                _Rasterizer = new PDFRasterizer();
                _Rasterizer.SetDrawAnnotations(true);
            }

            public void ShutDown()
            {
                _Destroyed = true;
                _Rasterizer.Cancel = true;
                _Images.Clear();
            }

            public void SetColorPostProcessing(PDFRasterizerColorPostProcessMode mode, Windows.UI.Color? background, Windows.UI.Color? foreground)
            {
                if (mode == PDFRasterizerColorPostProcessMode.e_postprocess_gradient_map && background != null && foreground != null)
                {
                    _Rasterizer.SetColorPostProcessColors(background.Value, foreground.Value);
                    _Rasterizer.SetColorPostProcessColors(foreground.Value, background.Value);
                }
                else
                {
                    _Rasterizer.SetColorPostProcessMode(mode);
                }
                _Rasterizer.SetBackgroundColor(Windows.UI.Colors.White);
            }

            public bool HasImage(int pageNumber)
            {
                if (_Images.ContainsKey(pageNumber) && _Images[pageNumber].Bitmap != null)
                {
                    return true;
                }
                return false;
            }

            public WriteableBitmap GetImage(int pageNumber)
            {
                if (HasImage(pageNumber))
                {
                    return _Images[pageNumber].Bitmap;
                }
                return null;
            }

            public void SwitchToPage(int pageNumber)
            {
                if (!HasImage(pageNumber) && _CurrentJobPage != pageNumber)
                {
                    _Rasterizer.Cancel = true;
                }
                _CurrentPage = pageNumber;
                StartWork();
            }

            private async void StartWork()
            {
                if (_IsWorking)
                {
                    return;
                }
                _IsWorking = true;
                while (true)
                {
                    if (_Destroyed)
                    {
                        return;
                    }
                    try
                    {
                        _CurrentJobPage = GetNextJobPage();
                        string pagesWithImages = "";
                        foreach (int key in _Images.Keys)
                        {
                            pagesWithImages += " " + key;
                        }

                        if (_CurrentJobPage >= 0 && _CurrentJobPage <= _TotalPages)
                        {
                            _Rasterizer.Cancel = false;
                            ImageProperties props = await GetBitMapFromPageAsync(_CurrentJobPage, _Doc);
                            if (props != null && !_Destroyed)
                            {
                                Windows.UI.Xaml.Media.Imaging.WriteableBitmap wb = new Windows.UI.Xaml.Media.Imaging.WriteableBitmap(props.Width, props.Height);
                                System.IO.Stream pixelStream = wb.PixelBuffer.AsStream();
                                await Task<WriteableBitmap>.Run(() =>
                                {
                                    pixelStream.Seek(0, System.IO.SeekOrigin.Begin);
                                    pixelStream.Write(props.ImageBuffer.Value, 0, props.ImageBuffer.Value.Length);
                                });
                                if (!_Destroyed)
                                {
                                    props.Bitmap = wb;
                                    _Images[_CurrentJobPage] = props;
                                    RemoveFurthestBitmapIfNeeded(_CurrentPage);
                                    BitmapAvailable?.Invoke(_CurrentJobPage);
                                }
                            }
                        }
                        else
                        {
                            _IsWorking = false;
                            return;
                        }
                    }
                    catch (Exception e)
                    {
                        Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(e));
                    }
                    finally
                    {

                    }
                }
            }

            private int GetNextJobPage()
            {
                int nextJobPage = _CurrentPage;
                while (Math.Abs(nextJobPage - _CurrentPage) <= MAX_RENDERED_IMAGES_PER_DIRECTION)
                {
                    if (nextJobPage < 1 || nextJobPage > _TotalPages || _Images.ContainsKey(nextJobPage))
                    {
                        nextJobPage = _CurrentPage + (_CurrentPage - nextJobPage);
                        if (nextJobPage >= _CurrentPage)
                        {
                            ++nextJobPage;
                        }
                    }
                    else
                    {
                        return nextJobPage;
                    }
                }
                return -1;
            }

            private void RemoveFurthestBitmapIfNeeded(int pageNumber)
            {
                if (_Images.Count > MAX_RENDERED_IMAGES_PER_DIRECTION * 2 + 1)
                {
                    int furthestKey = pageNumber;
                    foreach (int key in _Images.Keys)
                    {
                        int dist = Math.Abs(key - pageNumber);
                        if (dist > MAX_RENDERED_IMAGES_PER_DIRECTION
                            && dist > Math.Abs(furthestKey - pageNumber))
                        {
                            furthestKey = key;
                        }
                    }
                    if (furthestKey != pageNumber)
                    {
                        _Images.Remove(furthestKey);
                    }
                }
            }

            private Windows.Foundation.IAsyncOperation<ImageProperties> GetBitMapFromPageAsync(int pageNumber, PDFDoc doc)
            {
                Task<ImageProperties> t = new Task<ImageProperties>(() =>
                {
                    return GetBitMapFromPage(pageNumber, doc);
                });
                t.Start();
                return t.AsAsyncOperation<ImageProperties>();
            }

            private ImageProperties GetBitMapFromPage(int pageNum, PDFDoc pdfDoc)
            {
                try
                {
                    ImageProperties props = new ImageProperties();

                    pdfDoc.LockRead();
                    Page page = pdfDoc.GetPage(pageNum);

                    double scaleFactor = 1;
                    Rect bbox = page.GetBox(PageBox.e_crop);
                    bbox.Normalize();

                    double pageWidth = page.GetPageWidth();
                    double pageHeight = page.GetPageHeight();

                    double pageScaleFactorX = 1;
                    double pageScaleFactorY = 1;

                    scaleFactor *= Math.Max(pageScaleFactorX, pageScaleFactorY);

                    if (scaleFactor <= 0)
                    {
                        return null;
                    }

                    Matrix2D mtx = new Matrix2D(scaleFactor, 0, 0, scaleFactor, 0, 0);
                    mtx = Matrix2D.Mult(mtx, page.GetDefaultMatrix(true, PageBox.e_crop, PageRotate.e_0));

                    int width = (int)(pageWidth * scaleFactor);
                    int height = (int)(pageHeight * scaleFactor);

                    int comps = 4; // BGRA
                    int stride = ((width * comps + 3) / 4) * 4;

                    ByteArrayRef img = new ByteArrayRef(stride * height);

                    _Rasterizer.Rasterize(page, img, width, height, stride, comps, false, mtx, bbox);
                    if (_Rasterizer.Cancel)
                    {
                        return null;
                    }
                    props.Width = width;
                    props.Height = height;
                    props.ImageBuffer = img;

                    return props;
                }
                catch (Exception e)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(e));
                }

                finally
                {
                    pdfDoc.UnlockRead();
                }

                return null;
            }
        }

        #endregion Bitmaps
    }
}
