﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Windows.UI.Input;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Media.Animation;
using Windows.UI.Xaml.Shapes;

using UIPoint = Windows.Foundation.Point;
using UIRect = Windows.Foundation.Rect;

using PDFPoint = pdftron.PDF.Point;
using PDFRect = pdftron.PDF.Rect;
using PDFPage = pdftron.PDF.Page;
using PDFDouble = pdftron.Common.DoubleRef;
using pdftron.PDF;
using pdftron.Common;

namespace pdftron.PDF.Tools
{
    class PolylineCreate : AdvancedShapeCreate
    {
        #region Members

        private bool mShapeHasBeenCreated = false;
        private Path mShape;
        private PathGeometry mPolygonGeometry = new PathGeometry();
        private PathFigure mFigure = new PathFigure();
        private UIPoint mFirstPoint = new UIPoint();
        private List<PDFPoint> mPdfPointList = new List<PDFPoint>();
        private int mPreviousDrawPointIndex = -1;
        private bool mCanSelectNewPoint = false;

        #endregion

        #region Constructor

        public PolylineCreate(PDFViewCtrl pdfViewControl, ToolManager toolManager) 
            : base(pdfViewControl, toolManager)
        {
            mNextToolMode = ToolType.e_polyline_create;
            mToolMode = ToolType.e_polyline_create;
        }

        #endregion

        #region Event Handlers

        internal override bool PointerPressedHandler(object sender, PointerRoutedEventArgs e)
        {
            try
            {
                base.PointerPressedHandler(sender, e);

                if (mSimpleShapeCanceledAtPress)
                {
                    return false;
                }

                mShapeHasBeenCreated = false;

                if (mStartDrawing)
                {
                    if (mShape == null)
                    {
                        CreatePath();

                        mPreviousDrawPointIndex++;

                        LineSegment segment = new LineSegment();
                        segment.Point = new UIPoint(0, 0);
                        mFigure.Segments.Add(segment);

                        TrimToPageBounds(m_cnPt1, mDrawThickness);
                        mPdfPointList.Add(new PDFPoint(m_cnPt1.x, m_cnPt1.y));
                        mFirstPoint = new UIPoint(m_cnPt1.x, m_cnPt1.y);
                    }
                    else
                    {
                        if (mCanSelectNewPoint)
                        {
                            //TrimToPageBounds already called for m_cnPt2 in PointerMovedHandler
                            double nextPointX = m_cnPt2.x - mFirstPoint.X;
                            double nextPointY = m_cnPt2.y - mFirstPoint.Y;

                            mPreviousDrawPointIndex++;
                            mPdfPointList.Add(new PDFPoint(m_cnPt2.x, m_cnPt2.y));

                            LineSegment segment = new LineSegment();
                            segment.Point = new UIPoint(nextPointX, nextPointY);
                            mFigure.Segments.Add(segment);

                            mCanSelectNewPoint = false;
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }

            return false;
        }      

        internal override bool PointerMovedHandler(object sender, PointerRoutedEventArgs e)
        {
            try
            {
                if (mSimpleShapeCanceledAtPress)
                {
                    return false;
                }

                base.PointerMovedHandler(sender, e);

                if (mStartDrawing) // Shape is ready to draw
                {
                    TrimToPageBounds(m_cnPt2, mDrawThickness);

                    double nextPointX = m_cnPt2.x - mFirstPoint.X;
                    double nextPointY = m_cnPt2.y - mFirstPoint.Y;

                    if (mPointerDown || (e.Pointer.PointerDeviceType != Windows.Devices.Input.PointerDeviceType.Touch))
                    {
                        if (mPdfPointList.Count == (mPreviousDrawPointIndex + 1))
                        {
                            mPdfPointList.Add(new PDFPoint(m_cnPt2.x, m_cnPt2.y));
                        }
                        else
                        {
                            mPdfPointList[mPdfPointList.Count - 1].x = m_cnPt2.x;
                            mPdfPointList[mPdfPointList.Count - 1].y = m_cnPt2.y;
                        }

                        if (mFigure.Segments.Count == (mPreviousDrawPointIndex + 1))
                        {
                            LineSegment segment = new LineSegment();
                            segment.Point = new UIPoint(nextPointX, nextPointY);
                            mFigure.Segments.Add(segment);
                        }
                        else
                        {
                            (mFigure.Segments[mFigure.Segments.Count - 1] as LineSegment).Point = new UIPoint(nextPointX, nextPointY);
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }

            return false;
        }

        internal override bool PointerReleasedHandler(object sender, PointerRoutedEventArgs e)
        {
            mCanSelectNewPoint = true;
            base.PointerReleasedHandler(sender, e);
            return true;
        }

        internal override bool DoubleTappedHandler(object sender, DoubleTappedRoutedEventArgs e)
        {
            if (e.PointerDeviceType != Windows.Devices.Input.PointerDeviceType.Touch)
            {
                mSimpleShapeCanceledAtPress = true;
            }
            return false;
        }

        internal override bool RightTappedHandler(Object sender, RightTappedRoutedEventArgs e)
        {
            if (e.PointerDeviceType != Windows.Devices.Input.PointerDeviceType.Touch)
            {
                mSimpleShapeCanceledAtPress = true;
            }
            return false;
        }

        internal override bool TappedHandler(object sender, TappedRoutedEventArgs e)
        {
            return base.TappedHandler(sender, e);
        }

        #endregion

        #region Methods

        public void CommitAnnotation()
        {
            try
            {
                if (mStartDrawing) // don't want to add any annotations resulting from a tap
                {
                    EnableScrolling();
                    Finish();
                    mToolManager.RaiseAnnotationAddedEvent(mAnnot, mDownPageNumber);
                }

                EndCurrentTool(ToolType.e_pan);
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
        }

        private void CreatePath()
        {
            try
            { 
                if (!mShapeHasBeenCreated)
                {
                    mShapeHasBeenCreated = true;
                    if (mDrawThickness == 0)
                    {
                        mDrawThickness = 1;
                    }

                    mShape = new Path();
                    mShape.Data = mPolygonGeometry;
                    mFigure.StartPoint = new UIPoint(0, 0);
                    mPolygonGeometry.Figures.Add(mFigure);
                    mShape.StrokeThickness = mDrawThickness;
                    mShape.Stroke = mStrokeBrush;
                    this.Children.Add(mShape);
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
        }

        internal override void Finish()
        {
            //creates a polyline annotation and adds it to the PDF page
            try
            {
                mPDFView.DocLock(true);

                PDFRect rect = GetAdvancedShapeBBox(mPdfPointList);
                if (rect != null)
                {
                    rect.Inflate(mDrawThickness);
                    Annots.PolyLine pdfPolyline = Annots.PolyLine.Create(mPDFView.GetDoc().GetSDFDoc(), rect);

                    int vertexIdx = 0;
                    foreach (PDFPoint point in mPdfPointList)
                    {
                        // The Rect is normalized, so we need to recalculate the points if we don't want it to start from 0,0
                        PDFPoint pdfPoint = ConvertFromCanvasPointToPagePoint(point, mDownPageNumber);
                        pdfPolyline.SetVertex(vertexIdx++, pdfPoint);
                    }

                    SetStyle(pdfPolyline);
                    pdfPolyline.RefreshAppearance();

                    UtilityFunctions.SetUniqueId(pdfPolyline, Guid.NewGuid().ToString());

                    PDFPage page = mPDFView.GetDoc().GetPage(mDownPageNumber);
                    page.AnnotPushBack(pdfPolyline);

                    mAnnotPageNum = mDownPageNumber;
                    mAnnot = pdfPolyline;
                    mPDFView.UpdateWithAnnot(pdfPolyline, mAnnotPageNum);
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                mPDFView.DocUnlock();
            }
        }

        #endregion
    }
}
