﻿using pdftron.PDF.Tools.Controls.ViewModels.Common;
using pdftron.PDF.Tools.Utilities;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.Data.Json;
using Windows.UI.Popups;

namespace pdftron.PDF.Tools.ToolSettings
{
    public class TextAnnotationFontItem : ViewModelBase
    {
        private string _FontName;
        public string FontName
        {
            get { return _FontName; }
        }

        public TextAnnotationFontItem(string fontName)
        {
            _FontName = fontName;
        }
    }

    public class FreeTextFontViewModel : ViewModelBase
    {
        public FreeTextFontViewModel()
        {
            InitCommands();
            _FontItems = new ObservableCollection<TextAnnotationFontItem>();
            _SelectedItems = new List<TextAnnotationFontItem>();

            InitFonts();
        }

        public event Windows.UI.Xaml.RoutedEventHandler ExitRequested;

        public void CleanUp()
        {
            if (_CurrentMessage != null && _CurrentMessage.Status == Windows.Foundation.AsyncStatus.Started)
            {
                _CurrentMessage.Cancel();
            }
        }

        #region Members / Properties

        private ObservableCollection<TextAnnotationFontItem> _FontItems = null;
        public ObservableCollection<TextAnnotationFontItem> FontItems { get { return _FontItems; } }

        public delegate List<TextAnnotationFontItem> SelectiedItemProviderDeleagate();
        private SelectiedItemProviderDeleagate _SelectedItemProvider;

        public SelectiedItemProviderDeleagate SelectedItemProvider
        {
            get { return _SelectedItemProvider; }
            set { _SelectedItemProvider = value; }
        }

        #endregion Properties


        #region Commands

        private void InitCommands()
        {
            FontItemsSelectionChangedCommand = new RelayCommand(FontItemsSelectionChangedCommandImpl);
            SaveFontListCommand = new RelayCommand(SaveFontListCommandImpl);
            CancelFontListCommand = new RelayCommand(CancelFontListCommandImpl);
        }

        public RelayCommand FontItemsSelectionChangedCommand { get; private set; }
        public RelayCommand SaveFontListCommand { get; private set; }
        public RelayCommand CancelFontListCommand { get; private set; }

        private void FontItemsSelectionChangedCommandImpl(object changeArgs)
        {

        }

        private void SaveFontListCommandImpl(object parameter)
        {
            bool saved = SaveFontList();
            if (saved)
            {
                Exit();
            }
        }

        private void CancelFontListCommandImpl(object parameter)
        {
            Exit();
        }

        #endregion Commands


        #region Impl

        private List<TextAnnotationFontItem> _SelectedItems = null;

        private void InitFonts()
        {
            _FontItems = new ObservableCollection<TextAnnotationFontItem>();
            try
            {
                string fontListJson = UtilityFunctions.FontBundle.FontListJson;

                JsonObject jobj = JsonObject.Parse(fontListJson);

                JsonArray fontArr = jobj.GetNamedArray(Constants.FONT_JSON_NAME);
                string lastDisplayName = "";

                for (uint i = 0; i < fontArr.Count; ++i)
                {
                    JsonObject fontObj = fontArr.GetObjectAt(i);
                    string displayName = fontObj.GetNamedString(Constants.FONT_JSON_DISPLAY_NAME);
                    if (!string.IsNullOrWhiteSpace(displayName) && !displayName.Equals(lastDisplayName))
                    {
                        _FontItems.Add(new TextAnnotationFontItem(displayName));
                        lastDisplayName = displayName;
                    }
                }
            }
            catch (Exception e)
            {
                AnalyticsHandlerBase.CURRENT.SendEvent(AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, AnalyticsHandlerBase.GetExceptionMessage(e));
            }
            finally
            {
            }
        }

        private bool SaveFontList()
        {
            if (SelectedItemProvider != null)
            {
                List<TextAnnotationFontItem> selectedItems = SelectedItemProvider();
                if (selectedItems != null)
                {
                    if (selectedItems.Count > 0)
                    {
                        HashSet<string> whiteListedFonts = new HashSet<string>();
                        foreach (TextAnnotationFontItem item in selectedItems)
                        {
                            whiteListedFonts.Add(item.FontName);
                        }
                        Settings.FontWhiteList = whiteListedFonts;
                    }
                    else
                    {
                        RaiseNoSelectedFontErrorMessage();
                        return false;
                    }
                }
            }
            return true;
        }

        private Windows.Foundation.IAsyncOperation<Windows.UI.Popups.IUICommand> _CurrentMessage;
        private async void RaiseNoSelectedFontErrorMessage()
        {
            try
            {
                MessageDialog message = new MessageDialog(ResourceHandler.GetString("TextAnnotationFonts_NoFontsError"));
                _CurrentMessage = message.ShowAsync();
                await _CurrentMessage;
            }
            catch (Exception)
            { 
            }
        }

        private void Exit()
        {
            ExitRequested?.Invoke(this, new Windows.UI.Xaml.RoutedEventArgs());
        }

        #endregion Impl
    }
}
