﻿using System;
using System.Text.RegularExpressions;

namespace pdftron.PDF.Tools
{

    public class ActionUtils
    {
        public delegate void PageChangedDelegate(int prevPage, int page);
        public static PageChangedDelegate PageChanged;

        public static bool IsEmail(string url)
        {
            string pattern = @"^(?!\.)(""([^""\r\\]|\\[""\r\\])*""|"
              + @"([-a-z0-9!#$%&'*+/=?^_`{|}~]|(?<!\.)\.)*)(?<!\.)"
              + @"@[a-z0-9][\w\.-]*[a-z0-9]\.[a-z][a-z\.]*[a-z]$";

            Regex regex = new Regex(pattern, RegexOptions.IgnoreCase);
            return regex.IsMatch(url);
        }
        public static async void LaunchBrowser(string uristring)
        {
            if (string.IsNullOrWhiteSpace(uristring))
            {
                return;
            }

            try
            {
                uristring = Annots.Link.GetNormalizedURL(uristring);

                // for now, we ignore emails.
                if (IsEmail(uristring))
                {
                    return;
                }

                var uri = new Uri(uristring);
                var success = await Windows.System.Launcher.LaunchUriAsync(uri);
                if (!success)
                {
                    // Take appropriate action if desirable
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
        }
        public static void ExecuteAction(ActionParameter action_param, PDFViewCtrl mPDFView, NavigationStack navigationStack = null)
        {
            Action action = action_param.GetAction();
            if (action != null)
            {
                ActionType at = action.GetType();
                if (at == ActionType.e_URI)
                {
                    pdftron.SDF.Obj o = action.GetSDFObj();
                    o = o.FindObj("URI");
                    if (o != null)
                    {
                        String uristring = o.GetAsPDFText();

                        LaunchBrowser(uristring);
                    }
                }
                else
                {
                    int prevPage = mPDFView.GetCurrentPage();
                    if (navigationStack != null)
                    {
                        navigationStack.ExecuteAction(action_param);
                    }
                    else
                    {
                        mPDFView.ExecuteAction(action_param);
                    }
                    int page = mPDFView.GetCurrentPage();
                    PageChanged?.Invoke(prevPage, page);
                }
            }
        }
    }
}
