﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.Storage.Streams;
using System.Runtime.InteropServices.WindowsRuntime;

using UIPoint = Windows.Foundation.Point;
using PDFPoint = pdftron.PDF.Point;
using PDFRect = pdftron.PDF.Rect;
using PDFDouble = pdftron.Common.DoubleRef;
using UIRect = Windows.Foundation.Rect;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Media;
using Windows.UI;
using pdftron.Common;
using Windows.Data.Json;
using pdftron.SDF;
using Windows.UI.Text;

namespace pdftron.PDF.Tools
{
    public class UtilityLogger
    {
        public class LogItem
        {
            public string LogText { get; private set; }
            public LogItem(string txt)
            {
                LogText = txt;
            }
        }


        private static UtilityLogger _Current;
        public static UtilityLogger Current
        {
            get
            {
                if (_Current == null)
                {
                    _Current = new UtilityLogger();
                }
                return _Current;
            }
        }

        public UtilityLogger()
        {
            _LogList = new System.Collections.ObjectModel.ObservableCollection<LogItem>();
            _Timings = new Dictionary<string, DateTime>();
            _Current = this;
        }

        private System.Collections.ObjectModel.ObservableCollection<LogItem> _LogList;
        
        public System.Collections.ObjectModel.ObservableCollection<LogItem> LogList
        {
            get { return _LogList; }
        }

        public void LogString(string log)
        {
            DateTime dt = DateTime.Now;
            string logString = string.Format("{0}:{1}:{2}.{3:000} - {4}", dt.Hour, dt.Minute, dt.Second, dt.Millisecond, log);
            _LogList.Insert(0, new LogItem(logString));
        }

        private Dictionary<string, DateTime> _Timings;
        public bool LogString(string log, string key)
        {
            if (_Timings.ContainsKey(key))
            {
                DateTime dt = DateTime.Now;
                TimeSpan timeTaken = dt - _Timings[key];
                _Timings.Remove(key);

                string logString = string.Format("{0}:{1}:{2}.{3:000} ({5}ms)- {4}", dt.Hour, dt.Minute, dt.Second, dt.Millisecond, 
                    log, (int)timeTaken.TotalMilliseconds);
                _LogList.Insert(0, new LogItem(logString));

                if (timeTaken.TotalMilliseconds > 100)
                {
                    return true;
                }
            }
            else
            {
                _Timings[key] = DateTime.Now;
                LogString(log);
            }

            return false;
        }
        
        public FrameworkElement GetScrollableLogStack()
        {
            Windows.UI.Xaml.Controls.ScrollViewer sv = new Windows.UI.Xaml.Controls.ScrollViewer();
            sv.ZoomMode = Windows.UI.Xaml.Controls.ZoomMode.Disabled;
            sv.HorizontalScrollMode = Windows.UI.Xaml.Controls.ScrollMode.Disabled;
            sv.VerticalScrollMode = Windows.UI.Xaml.Controls.ScrollMode.Auto;

            Windows.UI.Xaml.Controls.StackPanel sp = new Windows.UI.Xaml.Controls.StackPanel();
            sp.Orientation = Windows.UI.Xaml.Controls.Orientation.Vertical;
            sp.Width = 500;
            sv.Content = sp;
            sp.Background = new SolidColorBrush(Colors.Black);

            foreach (LogItem item in LogList)
            {
                Windows.UI.Xaml.Controls.TextBox tb = new Windows.UI.Xaml.Controls.TextBox();
                tb.IsReadOnly = true;
                tb.AcceptsReturn = true;
                tb.HorizontalAlignment = HorizontalAlignment.Stretch;
                tb.Text = item.LogText;
                tb.Foreground = new SolidColorBrush(Colors.White);

                sp.Children.Add(tb);
            }

            return sv;
        }

        public void ShowPopup()
        {
            Windows.UI.Xaml.Controls.Primitives.Popup pup = new Windows.UI.Xaml.Controls.Primitives.Popup();
            pup.IsLightDismissEnabled = true;

            pup.Child = GetScrollableLogStack();
            (pup.Child as FrameworkElement).Height = Window.Current.Bounds.Height - 80;
            pup.Height = Window.Current.Bounds.Height - 80;
            pup.HorizontalOffset = 200;
            pup.VerticalOffset = 80;
            pup.IsOpen = true;
        }
    }


    public class UtilityFunctions
    {
        public static string KeyIRT = "IRT";
        public static string KeyRT = "RT";
        public static string KeyNM = "NM";
        public static string ValueGroup = "Group";
        public static string PDFTronId = "pdftron";
        public static double PointToLineDistance(PDFPoint basePoint, PDFPoint tipPoint, PDFPoint target)
        {
            double dx;
            double dy;
            double dist;

            double lineXDist = tipPoint.x - basePoint.x;
            double lineYDist = tipPoint.y - basePoint.y;

            double squaredDist = (lineXDist * lineXDist) + (lineYDist * lineYDist);

            if (double.IsNaN(squaredDist) || Math.Abs(squaredDist) < 0.001)
            {
                dx = target.x - basePoint.x;
                dy = target.y - basePoint.y;
                dist = (dx * dx) + (dy * dy);
                return dist;
            }

            double distRatio = ((target.x - basePoint.x) * lineXDist + (target.y - basePoint.y) * lineYDist) / squaredDist;

            if (distRatio < 0)
            {
                distRatio = 0; // This way, we will compare against mBasePoint
            }
            if (distRatio > 1)
            {
                distRatio = 0; // This way, we will compare against mTipPoint
            }

            dx = basePoint.x - target.x + distRatio * lineXDist;
            dy = basePoint.y - target.y + distRatio * lineYDist;
            dist = (dx * dx) + (dy * dy);

            return dist;
        }


        /// <summary>
        /// Checks if point is inside quad, 
        /// Assumes the quad is convex
        /// </summary>
        /// <param name="quad">A convex quad</param>
        /// <param name="point"></param>
        /// <param name="inflate"></param>
        /// <returns></returns>
        public static bool IsPointInSideQuad(QuadPoint quad, PDFPoint point, double inflate)
        {
            bool l0 = IsLeftTurn(quad.p1, quad.p2, point);
            bool l1 = IsLeftTurn(quad.p2, quad.p3, point);
            bool l2 = IsLeftTurn(quad.p3, quad.p4, point);
            bool l3 = IsLeftTurn(quad.p4, quad.p1, point);
            if (inflate == 0)
            {
                return (l0 && l1 && l2 && l3) == (l0 || l1 || l2 || l3);
            }
            double thresh = inflate * inflate;
            return PointToLineDistance(quad.p1, quad.p2, point) < thresh
                || PointToLineDistance(quad.p2, quad.p3, point) < thresh
                || PointToLineDistance(quad.p3, quad.p4, point) < thresh
                || PointToLineDistance(quad.p4, quad.p1, point) < thresh;
        }

        public static bool IsLeftTurn(PDFPoint p0, PDFPoint p1, PDFPoint p2)
        {
            return ((p1.x - p0.x) * (p2.y - p0.y)
            - (p2.x - p0.x) * (p1.y - p0.y)) >= 0;
        }

        public static UIPoint GetPointCloserToLine(UIPoint basePoint, UIPoint tipPoint, UIPoint target)
        {
            UIPoint returnPoint = new UIPoint(target.X, target.Y);

            double lineXDist = tipPoint.X - basePoint.X;
            double lineYDist = tipPoint.Y - basePoint.Y;

            double squaredDist = (lineXDist * lineXDist) + (lineYDist * lineYDist);

            double distRatio = ((target.X - basePoint.X) * lineXDist + (target.Y - basePoint.Y) * lineYDist) / squaredDist;

            if (double.IsNaN(squaredDist) || Math.Abs(squaredDist) < 0.001)
            {
                returnPoint.X = (target.X + basePoint.X) / 2;
                returnPoint.Y = (target.Y + basePoint.Y) / 2;
                return returnPoint;
            }

            if (distRatio < 0)
            {
                distRatio = 0; // This way, we will compare against mBasePoint
            }
            if (distRatio > 1)
            {
                distRatio = 0; // This way, we will compare against mTipPoint
            }

            double dx = basePoint.X - target.X + distRatio * lineXDist;
            double dy = basePoint.Y - target.Y + distRatio * lineYDist;

            returnPoint = new UIPoint(target.X + (dx / 2), target.Y + (dy / 2));

            return returnPoint;
        }

        public static UIRect GetElementRect(Windows.UI.Xaml.FrameworkElement element, Windows.UI.Xaml.FrameworkElement target = null)
        {
            Windows.UI.Xaml.Media.GeneralTransform elementtransform = element.TransformToVisual(target);
            UIRect rect = elementtransform.TransformBounds(new UIRect(new UIPoint(0, 0), new Windows.Foundation.Size(element.ActualWidth, element.ActualHeight)));
            return rect;
        }

        public static T FindVisualChild<T>(DependencyObject depObj) where T : DependencyObject
        {
            if (depObj != null)
            {
                for (int i = 0; i < VisualTreeHelper.GetChildrenCount(depObj); i++)
                {
                    DependencyObject child = VisualTreeHelper.GetChild(depObj, i);
                    if (child != null && child is T)
                    {
                        return (T)child;
                    }

                    T childItem = FindVisualChild<T>(child);
                    if (childItem != null) return childItem;
                }
            }
            return null;
        }
        public static IEnumerable<T> FindAllVisualChildren<T>(DependencyObject depObj) where T : DependencyObject
        {
            if (depObj != null)
            {
                for (int i = 0; i < VisualTreeHelper.GetChildrenCount(depObj); i++)
                {
                    DependencyObject child = VisualTreeHelper.GetChild(depObj, i);
                    if (child != null && child is T)
                    {
                        yield return (T)child;
                    }

                    foreach (T childOfChild in FindAllVisualChildren<T>(child))
                    {
                        yield return childOfChild;
                    }
                }
            }
        }

        public static void CopySelectedTextToClipBoard(string text)
        {
#if !WINDOWS_PHONE_APP
            Windows.ApplicationModel.DataTransfer.DataPackage dataPackage = new Windows.ApplicationModel.DataTransfer.DataPackage();
            dataPackage.RequestedOperation = Windows.ApplicationModel.DataTransfer.DataPackageOperation.Copy;
            dataPackage.SetText(text);
            Windows.ApplicationModel.DataTransfer.Clipboard.SetContent(dataPackage);
#endif
        }

        public static void SetCursor(Windows.UI.Core.CoreCursorType cursorType)
        {
#if !WINDOWS_PHONE_APP
            Windows.UI.Xaml.Window.Current.CoreWindow.PointerCursor =
                new Windows.UI.Core.CoreCursor(cursorType, 1);
#endif
        }

        public static SettingsColor ConvertToSettingsColor(Color color)
        {
            return new SettingsColor(color.R, color.G, color.B, color.A != 0);
        }

        public static Color ConvertToColor(SettingsColor color)
        {
            byte alpha = 0;
            if (color.Use)
            {
                alpha = 255;
            }
            return Color.FromArgb(alpha, color.R, color.G, color.B);
        }

        public static ToolType GetToolTypeFromAnnotType(IAnnot annot)
        {
            switch (annot.GetAnnotType())
            {
                case AnnotType.e_Line:
                    Annots.Line line = new Annots.Line(annot.GetSDFObj());
                    if (IsArrow(line))
                    {
                        return ToolType.e_arrow_create;
                    }
                    return ToolType.e_line_create;
                case AnnotType.e_Square:
                    return ToolType.e_rect_create;
                case AnnotType.e_Circle:
                    return ToolType.e_oval_create;
                case AnnotType.e_Ink:
                    Annots.Ink ink = new Annots.Ink(annot.GetSDFObj());
                    if (ink.GetBlendMode() == GStateBlendMode.e_bl_multiply)
                    {
                        return ToolType.e_ink_marker;
                    }
                    return ToolType.e_ink_create;
                case AnnotType.e_Polyline:
                    return ToolType.e_polyline_create;
                case AnnotType.e_Polygon:
                    Annots.Polygon polygon = new Annots.Polygon(annot.GetSDFObj());
                    if (polygon.GetBorderEffect() == Annots.MarkupBorderEffect.e_Cloudy)
                    {
                        return ToolType.e_cloud_create;
                    }
                    return ToolType.e_polygon_create;
                case AnnotType.e_Text:
                    return ToolType.e_sticky_note_create;
                case AnnotType.e_FreeText:
                    return ToolType.e_text_annot_create;
                case AnnotType.e_Highlight:
                    return ToolType.e_text_highlight;
                case AnnotType.e_Underline:
                    return ToolType.e_text_underline;
                case AnnotType.e_StrikeOut:
                    return ToolType.e_text_strikeout;
                case AnnotType.e_Squiggly:
                    return ToolType.e_text_squiggly;
            }
            return ToolType.e_none;
        }

        public static bool IsArrow(pdftron.PDF.Annots.Line line)
        {
            Annots.LineEndingStyle endingStyle = line.GetStartStyle();
            if (endingStyle == Annots.LineEndingStyle.e_ClosedArrow || endingStyle == Annots.LineEndingStyle.e_OpenArrow
                || endingStyle == Annots.LineEndingStyle.e_RClosedArrow || endingStyle == Annots.LineEndingStyle.e_ROpenArrow)
            {
                return true;
            }
            endingStyle = line.GetEndStyle();
            if (endingStyle == Annots.LineEndingStyle.e_ClosedArrow || endingStyle == Annots.LineEndingStyle.e_OpenArrow
                || endingStyle == Annots.LineEndingStyle.e_RClosedArrow || endingStyle == Annots.LineEndingStyle.e_ROpenArrow)
            {
                return true;
            }
            return false;
        }

        public static bool AreFilesEqual(Windows.Storage.StorageFile file1, Windows.Storage.StorageFile file2)
        {
            if (file1 == null && file2 == null)
            {
                return true;
            }
            else if (file1 != null && file2 != null)
            {
#if WINDOWS_PHONE_APP
                string path1 = file1.Path;
                string path2 = file2.Path;
                if (!string.IsNullOrEmpty(path1) && !string.IsNullOrEmpty(path2))
                {
                    return path1.Equals(path2);
                }
#else 
                return file1.IsEqual(file2);
#endif
            }
            return false;
        }

        public static Windows.Foundation.Size GetUniformStretch(double maxWidth, double maxHeight, double width, double height)
        {
            double scale = Math.Max(width / maxWidth, height / maxHeight);
            return new Windows.Foundation.Size(width / scale, height / scale);
        }

        public static void HideElement(FrameworkElement element)
        {
            element.Opacity = 0;
            element.IsHitTestVisible = false;
            element.Width = 0;
        }

        public static void ShowElement(FrameworkElement element, double width)
        {
            element.Opacity = 1;
            element.IsHitTestVisible = true;
            element.Width = width;
        }

        public static Color GetPostProcessedColor(Color inColor, PDFViewCtrl ctrl)
        {
            double r = inColor.R / 255;
            double g = inColor.G / 255;
            double b = inColor.B / 255;
            ColorPt cp = new ColorPt(r, g, b);
            cp = ctrl.GetPostProcessedColor(cp);
            return Color.FromArgb(inColor.A, (byte)(cp.Get(0) * 255), (byte)(cp.Get(1) * 255), (byte)(cp.Get(2) * 255));
        }

        public static int GetAnnotIdexForAddingTextMarkup(Page page)
        {
            int insertionIndex = page.GetNumAnnots() - 1;
            for (; insertionIndex >= 0; --insertionIndex)
            {
                try
                {
                    var annot = page.GetAnnot(insertionIndex);
                    if (annot == null || !annot.IsValid()) continue;

                    AnnotType annotType = annot.GetAnnotType();  
                    if (annotType == AnnotType.e_Highlight ||
                        annotType == AnnotType.e_Underline ||
                        annotType == AnnotType.e_StrikeOut ||
                        annotType == AnnotType.e_Squiggly ||
                        annotType == AnnotType.e_Link ||
                        annotType == AnnotType.e_Widget)
                    {
                        break;
                    }
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
            }
            return insertionIndex + 1;
        }

        /// <summary>
        /// Adds a popup to a markup annotation.
        /// Assumes we already have a write lock, and that the annot does not have a popup already.
        /// </summary>
        /// <param name="doc">the document the annotation is in</param>
        /// <param name="markup">the annotation</param>
        /// <param name="text">the text to add</param>
        public static Annots.Popup AddPopupToAnnot(PDFDoc doc, Annots.IMarkup markup, string text)
        {
            Rect rect = markup.GetRect();
            // This rect is offset a bit. We do this because Adobe uses this to position the popup when you open it.
            Annots.Popup p = Annots.Popup.Create(doc.GetSDFDoc(), new Rect(rect.x2 + 50, rect.y2 + 50, rect.x2 + 200, rect.y2 + 200));
            p.SetParent(markup);
            markup.SetPopup(p);
            p.SetContents(text);
            return p;
        }

        public static string GetAnnotContent(Annots.IMarkup markup)
        {
            Annots.Popup pop = markup.GetPopup();
            string content = string.Empty;
            if (pop != null && pop.IsValid())
            {
                content = pop.GetContents();
            }
            else
            {
                content = markup.GetContents();
            }
            return content;
        }

        public static bool DoesAnnotHaveContent(Annots.IMarkup markup)
        {
            return !string.IsNullOrEmpty(GetAnnotContent(markup));
        }

        public static bool DoesAnnotHavePopup(Annots.IMarkup markup)
        {
            Annots.Popup pop = markup.GetPopup();
            return pop != null && pop.IsValid();
        }

        public static String GetSelectedTextOnPage(PDFViewCtrl ctrl, int pageNum)
        {
            String text = "";
            if (ctrl.HasSelectionOnPage(pageNum))
            {
                try
                {
                    ctrl.DocLockRead();
                    PDFViewCtrlSelection sel = ctrl.GetSelection(pageNum);
                    text = sel.GetAsUnicode();
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
                finally
                {
                    ctrl.DocUnlockRead();
                }
            }
            return text;
        }

        private static string _HelpIconString = new string((char)0x00a5, 1);
        private static string _HelpIconBGString = new string((char)0x017a, 1);
        public static string IconNameToFontIconOutline(string name)
        {
            switch (name)
            {
                case "Comment":
                    return "ś";
                case "Help":
                    return _HelpIconString;
                case "InsertCaret":
                    return "˘";
                case "Star":
                    return "Ş";
                case "LightBulb":
                    return "¤";
                case "NewParagraph":
                    return "¨";
                case "CrossMark":
                    return "ž";
                case "CheckMark":
                    return "š";
                case "LocationPin":
                    return "¦";
                case "ThumbsUp":
                    return "¬";
                default:
                    return name;
            }
        }

        public static string IconNameToFontIconFill(string name)
        {
            switch (name)
            {
                case "Comment":
                    return "›";
                case "Help":
                    return _HelpIconBGString;
                case "InsertCaret":
                    return "ˇ";
                case "Star":
                    return "©";
                case "LightBulb":
                    return "Ł";
                case "NewParagraph":
                    return "§";
                case "CrossMark":
                    return "ť";
                case "CheckMark":
                    return "™";
                case "LocationPin":
                    return "Ą";
                case "ThumbsUp":
                    return "«";
                default:
                    return name;
            }
        }

        /// <summary>
        /// This returns a magic number to multiply the font size of an icon in order to
        /// make it match the icon for the sticky note pushed into the document.
        /// </summary>
        /// <param name="name"></param>
        /// <returns></returns>
        public static double IconNameToFontSizeMultiplier(string name)
        {
            switch (name)
            {
                case "Comment":
                    return 1.0;
                case "Help":
                    return 1.04;
                case "InsertCaret":
                    return 0.92;
                case "Star":
                    return 1.04;
                case "LightBulb":
                    return 1.0;
                case "NewParagraph":
                    return 1.08;
                case "CrossMark":
                    return 1.08;
                case "CheckMark":
                    return 1.08;
                case "LocationPin":
                    return 1.04;
                case "ThumbsUp":
                    return 1.0;
                default:
                    return 1.0;
            }
        }

        public static Rect ConvertFromScreenRectToAnnotationCanvasRect(PDFViewCtrl ctrl, Rect rect)
        {
            DoubleRef x1 = new DoubleRef(rect.x1);
            DoubleRef y1 = new DoubleRef(rect.y1);
            DoubleRef x2 = new DoubleRef(rect.x2);
            DoubleRef y2 = new DoubleRef(rect.y2);
            ctrl.ConvScreenPtToAnnotationCanvasPt(x1, y1);
            ctrl.ConvScreenPtToAnnotationCanvasPt(x2, y2);
            return new Rect(x1.Value, y1.Value, x2.Value, y2.Value);
        }

        // 1 x 1 png
        static readonly byte[] PngBytes = {
            137, 80, 78, 71, 13, 10, 26, 10, 0, 0, 0, 13, 73, 72, 68, 82, 0, 0, 0, 1, 0, 0, 0, 1,
            8, 6, 0, 0, 0, 31, 21, 196, 137, 0, 0, 0, 1, 115, 82, 71, 66, 0, 174, 206, 28, 233, 0,
            0, 0, 4, 103, 65, 77, 65, 0, 0, 177, 143, 11, 252, 97, 5, 0, 0, 0, 9, 112, 72, 89, 115,
            0, 0, 14, 195, 0, 0, 14, 195, 1, 199, 111, 168, 100, 0, 0, 0, 24, 116, 69, 88, 116, 83,
            111, 102, 116, 119, 97, 114, 101, 0, 112, 97, 105, 110, 116, 46, 110, 101, 116, 32, 52,
            46, 48, 46, 51, 140, 230, 151, 80, 0, 0, 0, 13, 73, 68, 65, 84, 24, 87, 99, 248, 255,
            255, 255, 127, 0, 9, 251, 3, 253, 5, 67, 69, 202, 0, 0, 0, 0, 73, 69, 78, 68, 174, 66,
            96, 130
        };

        public static async Task<InMemoryRandomAccessStream> ConvertToPngStreamAsync(byte[] buffer, int w, int h)
        {
            InMemoryRandomAccessStream memoryStream = new InMemoryRandomAccessStream();
            await memoryStream.WriteAsync(PngBytes.AsBuffer()).AsTask().ConfigureAwait(false);
            memoryStream.Seek(0);

            Windows.Graphics.Imaging.BitmapEncoder encode = await Windows.Graphics.Imaging.BitmapEncoder.CreateAsync(Windows.Graphics.Imaging.BitmapEncoder.PngEncoderId,
                memoryStream).AsTask().ConfigureAwait(false);
            encode.SetPixelData(Windows.Graphics.Imaging.BitmapPixelFormat.Bgra8, Windows.Graphics.Imaging.BitmapAlphaMode.Straight,
                (uint)w, (uint)h, 96.0, 96.0, buffer);
            await encode.FlushAsync().AsTask().ConfigureAwait(false);
            memoryStream.Seek(0);
            return memoryStream;
        }

        public static void CloseAllOpenPopups()
        {
            Controls.ViewModels.AnnotationPropertiesFontNameSelectorViewModel.ClosePopups();
            Controls.ViewModels.AnnotationPropertiesDisplay.CloseCurrentPopup();
        }

        public static bool IsContinuous(PDFViewCtrlPagePresentationMode presentationMode)
        {
            return (presentationMode == PDFViewCtrlPagePresentationMode.e_facing_continuous ||
            presentationMode == PDFViewCtrlPagePresentationMode.e_facing_continuous_cover ||
            presentationMode == PDFViewCtrlPagePresentationMode.e_single_continuous);
        }

        /// <summary>
        /// Checks if the annotation is an annotation reply
        /// </summary>
        /// <param name="annot">the annotation</param>
        /// <returns>true if it is an annotation reply, false otherwise</returns>
        public static bool IsReplyTypeReply(IAnnot annot)
        {
            try
            {
                if (annot == null)
                {
                    return false;
                }

                Obj annotSDFObj = annot.GetSDFObj();
                if (annotSDFObj != null)
                {
                    Obj irt = annotSDFObj.FindObj(KeyIRT);
                    // if annotation contains IRT and does not contain RT of "Group"
                    // then it is a reply
                    return ((irt != null) && (!IsReplyTypeGroup(annot)));
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            return false;
        }

        /// <summary>
        /// Gets the list of annotations on a given page.
        /// </summary>
        /// <param name="pdfViewCtrl">PDFViewCtrl object which contains the current document.</param>
        /// <param name="pageNumber">Page number on which to search for annotations.</param>
        /// <returns>List of annotations on a given page.</returns>
        public static List<IAnnot> GetAnnotationsOnPage(PDFViewCtrl pdfViewCtrl, int pageNumber)
        {
            List<IAnnot> annotationList = new List<IAnnot>();
            try
            {
                PDFDoc document = pdfViewCtrl.GetDoc();
                if (document != null)
                {
                    Page page = document.GetPage(pageNumber);
                    
                    if (page != null)
                    {
                        int annotationCount = page.GetNumAnnots();
                        for (int i = 0; i < annotationCount; i++)
                        {
                            var annot = page.GetAnnot(i);
                            if (annot == null || !annot.IsValid()) continue;

                            annotationList.Add(annot);
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(
                    Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                    Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }            

            return annotationList;
        }

        /// <summary>
        /// Gets the unique Id of the IRT parent annotation of this annotation
        /// </summary>
        /// <param name="annotation">The annotation</param>
        /// <returns>The unique id of the IRT parent annotation of this annotation</returns>
        public static string GetIRTAsString(IAnnot annotation)
        {
            if (annotation == null)
            {
                return null;
            }

            Obj annotSDFObj = annotation.GetSDFObj();
            if (annotSDFObj != null)
            {
                Obj irt = annotSDFObj.FindObj(KeyIRT);
                if (irt != null)
                {
                    // first check string, for backwards compatibility
                    if (irt.IsString())
                    {
                        return irt.GetAsPDFText();
                    }
                    // IRT is dict link
                    if (irt.IsDict())
                    {
                        Obj nm = irt.FindObj(KeyNM);
                        if ((nm != null) && nm.IsString())
                        {
                            return nm.GetAsPDFText();
                        }
                    }
                }
            }
            return null;
        }

        /// <summary>
        /// Checks if the annotation is part of a group annotation.
        /// </summary>
        /// <param name="annotation">The annotation.</param>
        /// <returns>True if it is part of a group annotation, false otherwise.</returns>
        public static bool IsReplyTypeGroup(IAnnot annotation)
        {
            if (annotation == null)
            {
                return false;
            }

            // group annotation
            // RT is only meaningful if IRT is present
            // valid values are "R" and "Group"
            Obj annotSDFObj = annotation.GetSDFObj();
            if (annotSDFObj != null)
            {
                Obj irt = annotSDFObj.FindObj(KeyIRT);
                Obj rt = annotSDFObj.FindObj(KeyRT);
                if ((irt != null) && (rt != null) && rt.IsName())
                {
                    string rtVal = rt.GetName();
                    return ValueGroup.Equals(rtVal);
                }
            }
            return false;
        }

        /// <summary>
        /// Returns all annotations in this group if the annotation passed in is in an annotation group
        /// </summary>
        /// <param name="pdfViewCtrl">PDFViewCtrl object which contains the current document</param>
        /// <param name="annotation">the annotation.</param>
        /// <param name="pageNumber">page number of page that contains the annotations</param>
        /// <returns>returns all annotations in this group if the annotation passed in is in an annotation group</returns>
        public static List<IAnnot> GetAnnotationsInGroup(PDFViewCtrl pdfViewCtrl, IAnnot annotation, int pageNumber) 
        {
            if ((pdfViewCtrl == null) || (annotation == null))
            {
                return null;
            }

            if (!annotation.IsValid())
            {
                return null;
            }

            bool shouldUnlockRead = false;
            try
            {
                List<IAnnot> annotationsInGroup = new List<IAnnot>();
                pdfViewCtrl.DocLockRead();
                shouldUnlockRead = true;

                string mainAnnotationId = null;

                if (IsReplyTypeGroup(annotation))
                {
                    // this is a sub annot in a group
                    // its IRT will be the primary annot in the group
                    mainAnnotationId = GetIRTAsString(annotation);
                }
                else if (annotation.GetUniqueID() != null)
                {
                    // this is a primary annot
                    mainAnnotationId = annotation.GetUniqueID().GetAsPDFText();
                }

                if (String.IsNullOrEmpty(mainAnnotationId))
                {
                    return null;
                }

                List<IAnnot> annotationsInPage = GetAnnotationsOnPage(pdfViewCtrl, pageNumber);
                foreach (IAnnot annotationItem in annotationsInPage)
                {
                    if ((annotationItem != null) && annotationItem.IsValid() && (annotationItem.GetUniqueID() != null))
                    {
                        // find all annotations in the group
                        string id = annotationItem.GetUniqueID().GetAsPDFText();
                        if ((id != null) && id.Equals(mainAnnotationId))
                        {
                            // add primary annot
                            annotationsInGroup.Add(annotationItem);
                        }
                        string irt = GetIRTAsString(annotationItem);
                        if (IsReplyTypeGroup(annotationItem) && (irt != null) && irt.Equals(mainAnnotationId))
                        {
                            // add subordinate annot
                            annotationsInGroup.Add(annotationItem);
                        }
                    }
                }

                return annotationsInGroup;
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                if (shouldUnlockRead)
                {
                    pdfViewCtrl.DocUnlockRead();
                }
            }

            return null;
        }

        /// <summary>
        /// Snaps a point to the specified rectangle
        /// </summary>
        /// <param name="point">the resulted point snapped to the rectangle</param>
        /// <param name="rect">the rectangle</param>        
        public static void SnapPointToRect(PDFPoint point, PDFRect rect)
        {
            if (rect != null)
            {
                if (point.x < rect.x1)
                {
                    point.x = rect.x1;
                }
                else if (point.x > rect.x2)
                {
                    point.x = rect.x2;
                }
                if (point.y < rect.y2)
                {
                    point.y = rect.y2;
                }
                else if (point.y > rect.y1)
                {
                    point.y = rect.y1;
                }
            }
        }

        /// <summary>
        /// Gets the annotation creation date in local date time; if it fails then UTC time is returned
        /// </summary>
        /// <param name="annotation">the annotation</param>
        /// <returns>the date</returns>
        public static DateTime GetAnnotationLocalCreationDate(IAnnot annotation)
        {
            pdftron.PDF.Date date = annotation.GetDate();
            if (annotation.IsMarkup())
            {
                date = (annotation as Annots.Markup).GetCreationDates();
            }

            DateTime annotTime = new DateTime(date.year, date.month, date.day, date.hour, date.minute, date.second);
            DateTime timeUtc = annotTime.ToUniversalTime();

            try
            {                 
                TimeZoneInfo localZone = TimeZoneInfo.Local;
                DateTime localTime = TimeZoneInfo.ConvertTime(timeUtc, localZone);
                return localTime;
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }

            return timeUtc;
        }

        /// <summary>
        /// Returns true if the annotation was created with PDFTron SDK
        /// </summary>
        /// <param name="annotation">the annotation</param>
        /// <returns>true if created by PDFTron SDK; false otherwise</returns>
        public static bool IsMadeByPDFTron(IAnnot annotation)
        {
            string[] supportedTag = new string[] { PDFTronId, "pdftronlink" };
            if (annotation != null && annotation.GetSDFObj() != null)
            {
                Obj sdfObj = annotation.GetSDFObj();
                foreach (String tag in supportedTag)
                {
                    Object selfMadeObj = sdfObj.FindObj(tag);
                    if (selfMadeObj != null)
                    {
                        return true;
                    }
                }
            }
            return false;
        }

        /// <summary>
        /// Ungroups the annotations
        /// </summary>
        /// <param name="pdfViewCtrl">PDFViewCtrl object which contains the current document</param>
        /// <param name="allAnnotationsInGroup">the annotations in the group</param>
        public static void UngroupAnnotations(PDFViewCtrl pdfViewCtrl, List<IAnnot> allAnnotationsInGroup)
        {
            if ((pdfViewCtrl == null) || (allAnnotationsInGroup == null))
            {
                return;
            }
            bool shouldUnlock = false;
            try
            {
                pdfViewCtrl.DocLock(true);
                shouldUnlock = true;
                foreach (IAnnot annotation in allAnnotationsInGroup)
                {
                    // remove all related keys
                    annotation.GetSDFObj().Erase(KeyRT);
                    annotation.GetSDFObj().Erase(KeyIRT);

                    annotation.SetDateToNow(); // Update Date in case of using ExternalAnnotManager
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                if (shouldUnlock)
                {
                    pdfViewCtrl.DocUnlock();
                }
            }
        }

        /// <summary>
        /// Creates a group of annotations
        /// </summary>
        /// <param name="pdfViewCtrl">PDFViewCtrl object which contains the current document</param>
        /// <param name="primaryAnnotation">the primary annotation</param>
        /// <param name="allAnnotationsInGroup">all annotations to be added to this group</param>
        public static void CreateAnnotationGroup(PDFViewCtrl pdfViewCtrl, IAnnot primaryAnnotation, List<IAnnot> allAnnotationsInGroup)
        {
            if ((pdfViewCtrl == null) || (primaryAnnotation == null) || (allAnnotationsInGroup == null))
            {
                return;
            }

            bool shouldUnlock = false;

            try
            {
                pdfViewCtrl.DocLock(true);
                shouldUnlock = true;
                foreach (IAnnot annotation in allAnnotationsInGroup)
                {
                    // first set Guid if not exist yet
                    if (annotation.GetUniqueID() == null)
                    {
                        SetUniqueId(annotation, Guid.NewGuid().ToString());
                    }
                    if (!annotation.Equals(primaryAnnotation))
                    {
                        annotation.GetSDFObj().PutName(KeyRT, ValueGroup);
                        annotation.GetSDFObj().Put(KeyIRT, primaryAnnotation.GetSDFObj());
                    }
                    annotation.SetDateToNow(); // Update Date in case of using ExternalAnnotManager
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(
                    Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, 
                    Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                if (shouldUnlock)
                {
                    pdfViewCtrl.DocUnlock();
                }
            }
        }

        /// <summary>
        /// Sets the unique identifier for the specified annotation, if possible.
        /// This method does not lock the document, so a write lock should be acquired outside.
        /// </summary>
        /// <param name="annotation">the annotation</param>
        /// <param name="id">the unique identifier</param>
        public static void SetUniqueId(IAnnot annotation, string id)
        {
            try
            {
                if (annotation != null)
                {
                    annotation.SetUniqueID(id);
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
        }

        #region FreeText fonts
        public class FontBundle
        {
            /// <summary>
            /// Returns a string intended for use by system UI widgets like TextBoxes.
            /// </summary>
            public string FontUri
            {
                get
                {
                    // TODO - Figure out how this works with imported fonts.
                    string localFolder = Windows.Storage.ApplicationData.Current.LocalFolder.Path;
                    localFolder = localFolder.Replace("\\", "/");
                    if (FontPath.Contains(localFolder))
                    {
                        string fontUri = FontPath.Replace(localFolder, "ms-appdata:///local") + "#" + DisplayName;
                        return fontUri;
                    }
                    return FontFamily;
                }
            }

            public string FontPath { get; private set; }
            public string DisplayName { get; private set; }
            public string PDFTronName { get; private set; }

            public string FontFamily { get; private set; }

            public FontWeight FontWeight { get; private set; }

            public FontStyle FontStyle { get; private set; }

            public static string FontListJson { get; private set; }

            public FontBundle(string fontPath, string displayName, string pdfTronName, string family, FontWeight weight, FontStyle style)
            {
                FontPath = fontPath;
                DisplayName = displayName;
                PDFTronName = pdfTronName;
                FontFamily = family;
                FontWeight = weight;
                FontStyle = style;
            }

            private enum CompareProperties
            {
                DisplayName,
                PDFTronName,
            }

            public static FontBundle CreateNullFont()
            {
                return new FontBundle("", "", "", "", FontWeights.Normal, FontStyle.Normal);
            }

            public static bool IsNullFont(FontBundle font)
            {
                return string.IsNullOrWhiteSpace(font.DisplayName) || string.IsNullOrWhiteSpace(font.FontPath) || string.IsNullOrWhiteSpace(font.PDFTronName);
            }

            public static FontBundle Create(ToolType toolType)
            {
                return Create(Settings.GetFontName(toolType), CompareProperties.DisplayName);
            }

            public static FontBundle CreateFromDisplayName(string displayName)
            {
                return FontBundle.Create(displayName, CompareProperties.DisplayName);
            }

            public static FontBundle CreateFromPDFTronName(string pdfTronName)
            {
                return FontBundle.Create(pdfTronName, CompareProperties.PDFTronName);
            }

            private static FontBundle Create(string fontName, CompareProperties comparer)
            {
                return Create(fontName, comparer, false);
            }

            public static FontBundle CreateDefaultFontBundle()
            {
                return Create(ToolSettings.Constants.FONT_DEFAULT_NAME, CompareProperties.DisplayName, true);
            }

            public static async Task<string> GetSystemFontListAsync()
            {
                try
                {
                    FontListJson = await Task.Run(() => pdftron.PDFNet.GetSystemFontListJSON());
                }
                catch (Exception exception)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                            Utilities.AnalyticsHandlerBase.GetExceptionMessage(exception));
                }
                return FontListJson;
            }

            private static FontBundle Create(string fontName, CompareProperties comparer, bool pickFistIfFails)
            {
                if (String.IsNullOrEmpty(FontListJson))
                {
                    return null;
                }

                FontBundle firstFontbundle = null;
                try
                {
                    string fontListJson = FontListJson;

                    JsonObject jobj = JsonObject.Parse(fontListJson);

                    JsonArray fontArr = jobj.GetNamedArray(ToolSettings.Constants.FONT_JSON_NAME);

                    for (uint i = 0; i < fontArr.Count; ++i)
                    {
                        JsonObject fontObj = fontArr.GetObjectAt(i);
                        string displayName = fontObj.GetNamedString(ToolSettings.Constants.FONT_JSON_DISPLAY_NAME);
                        string pdfTronName = fontObj.GetNamedString(ToolSettings.Constants.FONT_JSON_PDFTRON_NAME);
                        if ((comparer == CompareProperties.DisplayName && displayName.Equals(fontName)) ||
                            (comparer == CompareProperties.PDFTronName && pdfTronName.Equals(fontName)))
                        {
                            FontBundle fontBundle = new FontBundle(fontObj.GetNamedString(ToolSettings.Constants.FONT_JSON_FILE_PATH), displayName, pdfTronName,
                                fontObj.GetNamedString(ToolSettings.Constants.FONT_JSON_FAMILY_NAME),
                                StringToFontWeight(fontObj.GetNamedString(ToolSettings.Constants.FONT_JSON_FONT_WEIGTH)),
                                StringToFontStyle(fontObj.GetNamedString(ToolSettings.Constants.FONT_JSON_FONT_STYLE)));
                            return fontBundle;
                        }
                        if (pickFistIfFails)
                        {
                            firstFontbundle = new FontBundle(fontObj.GetNamedString(ToolSettings.Constants.FONT_JSON_FILE_PATH), displayName, pdfTronName,
                                fontObj.GetNamedString(ToolSettings.Constants.FONT_JSON_FAMILY_NAME),
                                StringToFontWeight(fontObj.GetNamedString(ToolSettings.Constants.FONT_JSON_FONT_WEIGTH)),
                                StringToFontStyle(fontObj.GetNamedString(ToolSettings.Constants.FONT_JSON_FONT_STYLE)));
                            pickFistIfFails = false;
                        }
                    }
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }

                return firstFontbundle;
            }

            public override bool Equals(object obj)
            {
                FontBundle other = obj as FontBundle;
                if (other != null)
                {
                    return this.DisplayName.Equals(other.DisplayName);
                }
                return base.Equals(obj);
            }

            public override int GetHashCode()
            {
                return this.DisplayName.GetHashCode();
            }

            public static FontWeight StringToFontWeight(string weightString)
            {
                FontWeight weight = FontWeights.Normal;
                if (!string.IsNullOrWhiteSpace(weightString))
                {
                    if (weightString.Equals("Thin", StringComparison.OrdinalIgnoreCase))
                    {
                        weight = FontWeights.Thin;
                    }
                    else if (weightString.Equals("ExtraLight", StringComparison.OrdinalIgnoreCase))
                    {
                        weight = FontWeights.ExtraLight;
                    }
                    else if (weightString.Equals("Light", StringComparison.OrdinalIgnoreCase))
                    {
                        weight = FontWeights.Light;
                    }
                    else if (weightString.Equals("Medium", StringComparison.OrdinalIgnoreCase))
                    {
                        weight = FontWeights.Medium;
                    }
                    else if (weightString.Equals("SemiBold", StringComparison.OrdinalIgnoreCase))
                    {
                        weight = FontWeights.SemiBold;
                    }
                    else if (weightString.Equals("Bold", StringComparison.OrdinalIgnoreCase))
                    {
                        weight = FontWeights.Bold;
                    }
                    else if (weightString.Equals("Heavy", StringComparison.OrdinalIgnoreCase))
                    {
                        weight = FontWeights.ExtraBold;
                    }
                    else if (weightString.Equals("Black", StringComparison.OrdinalIgnoreCase))
                    {
                        weight = FontWeights.Black;
                    }
                    else if (weightString.Equals("UltraBlack", StringComparison.OrdinalIgnoreCase))
                    {
                        weight = FontWeights.ExtraBlack;
                    }
                }
                return weight;
            }

            public static FontStyle StringToFontStyle(string styleString)
            {
                FontStyle style = FontStyle.Normal;
                if (!string.IsNullOrWhiteSpace(styleString))
                {
                    if (styleString.Equals("Italic", StringComparison.OrdinalIgnoreCase))
                    {
                        style = FontStyle.Italic;
                    }
                    else if (styleString.Equals("Oblique", StringComparison.OrdinalIgnoreCase))
                    {
                        style = FontStyle.Oblique;
                    }
                }
                return style;
            }
        }

        public static List<FontBundle> GetListOfSystemFonts()
        {
            List<FontBundle> fontList = new List<FontBundle>();

            if (String.IsNullOrEmpty(FontBundle.FontListJson))
            {
                return fontList;
            }

            try
            {
                string fontListJson = FontBundle.FontListJson;

                JsonObject jobj = JsonObject.Parse(fontListJson);

                JsonArray fontArr = jobj.GetNamedArray(ToolSettings.Constants.FONT_JSON_NAME);
                string lastDisplayName = "";

                for (uint i = 0; i < fontArr.Count; ++i)
                {
                    JsonObject fontObj = fontArr.GetObjectAt(i);
                    string displayName = fontObj.GetNamedString(ToolSettings.Constants.FONT_JSON_DISPLAY_NAME);
                    if (!string.IsNullOrWhiteSpace(displayName) && !displayName.Equals(lastDisplayName))
                    {
                        fontList.Add(new FontBundle(fontObj.GetNamedString(ToolSettings.Constants.FONT_JSON_FILE_PATH), displayName,
                            fontObj.GetNamedString(ToolSettings.Constants.FONT_JSON_PDFTRON_NAME),
                                fontObj.GetNamedString(ToolSettings.Constants.FONT_JSON_FAMILY_NAME),
                                FontBundle.StringToFontWeight(fontObj.GetNamedString(ToolSettings.Constants.FONT_JSON_FONT_WEIGTH)),
                                FontBundle.StringToFontStyle(fontObj.GetNamedString(ToolSettings.Constants.FONT_JSON_FONT_STYLE))));

                        lastDisplayName = displayName;
                    }
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            return fontList;
        }

        public static void SetFontFromBundle(Annots.FreeText freeText, PDFDoc doc, FontBundle bundle)
        {
            string pdftronFontName = bundle.PDFTronName;

            // Create a DR entry for embedding the font
            Obj annotObj = freeText.GetSDFObj();
            Obj drDict = annotObj.PutDict("DR");

            // Embed the font
            Obj fontDict = drDict.PutDict("Font");
            Font font = Font.Create(doc, pdftronFontName, freeText.GetContents());
            fontDict.Put(ToolSettings.Constants.FONT_DR_NAME, font.GetSDFObj());
        }

        public static void SetFreeTextDefaultAppearanceWithoutFont(Annots.FreeText freeText)
        {
            SetFreeTextDefaultAppearance(freeText, ToolSettings.Constants.FONT_FREETEXT_DA_DEFAULT);
        }

        public static void SetFreeTextDefaultAppearanceWithFont(Annots.FreeText freeText)
        {
            SetFreeTextDefaultAppearance(freeText, ToolSettings.Constants.FONT_DR_NAME);
        }

        private static void SetFreeTextDefaultAppearance(Annots.FreeText freeText, string fontName)
        {
            // Set DA string
            String DA = freeText.GetDefaultAppearance();
            int slashPosition = DA.IndexOf("/", 0);

            // if DR string contains '/' which it always should.
            if (slashPosition > 0)
            {
                String beforeSlash = DA.Substring(0, slashPosition);
                String afterSlash = DA.Substring(slashPosition);
                String afterFont = afterSlash.Substring(afterSlash.IndexOf(" "));
                String updatedDA = beforeSlash + "/" + fontName + afterFont;

                freeText.SetDefaultAppearance(updatedDA);
            }
        }

        public static FontBundle GetFontFromAnnot(Annots.FreeText freeText)
        {
            string fontName = GetFontNameFromAnnot(freeText);
            if (fontName != null)
            {
                return UtilityFunctions.FontBundle.CreateFromPDFTronName(fontName);
            }
            return null;
        }

        public static string GetFontNameFromAnnot(Annots.FreeText freeText)
        {
            Obj ftObj = freeText.GetSDFObj();
            Obj drDict = ftObj.FindObj("DR");
            if (drDict != null && drDict.IsDict())
            {
                Obj fontDict = drDict.FindObj("Font");
                if (fontDict != null && fontDict.IsDict())
                {
                    DictIterator fItr = fontDict.GetDictIterator();
                    if (fItr.HasNext())
                    {
                        Font f = new Font(fItr.Value());
                        Obj fontDescriptor = f.GetDescriptor();
                        if (fontDescriptor != null && fontDescriptor.IsDict())
                        {
                            Obj fontNameObj = fontDescriptor.FindObj("FontName");
                            if (fontNameObj != null && fontNameObj.IsName())
                            {
                                return fontNameObj.GetName();
                            }
                        }
                        string fontName = f.GetName();
                        return fontName;
                    }
                }
            }
            return null;
        }

        public static void GenerateDSEntry(Annots.FreeText freeText, FontBundle fontBundle = null)
        {
            if (fontBundle == null)
            {
                fontBundle = GetFontFromAnnot(freeText);
            }
            ColorPt textColor = freeText.GetTextColor();
            int r = (int)(textColor.Get(0) * 255 + 0.5);
            int g = (int)(textColor.Get(1) * 255 + 0.5);
            int b = (int)(textColor.Get(2) * 255 + 0.5);
            string colorString = Windows.UI.Color.FromArgb(255, (byte)r, (byte)g, (byte)b).ToString();
            colorString = "#" + colorString.Substring(3);

            double fs = freeText.GetFontSize();

            FontStyle style = fontBundle.FontStyle;
            if (style == FontStyle.Oblique) // oblique is not a valid value
            {
                style = FontStyle.Italic;
            }

            string formatString = string.Format("font-family: {0}; font-style: {1}; font-weight: {2}; font-size: {3}pt; color: {4}", 
                fontBundle.FontFamily, style, fontBundle.FontWeight.Weight, fs, colorString);
            freeText.GetSDFObj().PutString("DS", formatString);
        }

        #endregion FreeText fonts

        #region Rectangle Conversions

        public static PDFPoint ConvertFromScreenPointToCanvasPoint(PDFPoint point, PDFViewCtrl ctrl)
        {
            PDFDouble x = new PDFDouble(point.x);
            PDFDouble y = new PDFDouble(point.y);

            ctrl.ConvScreenPtToAnnotationCanvasPt(x, y);

            return new PDFPoint(x.Value, y.Value);
        }

        public static PDFPoint ConvertFromScreenPointToCanvasPoint(double x, double y, PDFViewCtrl ctrl)
        {
            PDFDouble px = new PDFDouble(x);
            PDFDouble py = new PDFDouble(y);

            ctrl.ConvScreenPtToAnnotationCanvasPt(px, py);

            return new PDFPoint(px.Value, py.Value);
        }

        public static PDFPoint ConvertFromCanvasPointToScreenPoint(PDFPoint point, PDFViewCtrl ctrl)
        {
            PDFDouble x = new PDFDouble(point.x);
            PDFDouble y = new PDFDouble(point.y);

            ctrl.ConvAnnotationCanvasPtToScreenPt(x, y);

            return new PDFPoint(x.Value, y.Value);
        }

        public static PDFPoint ConvertFromCanvasPointToScreenPoint(double x, double y, PDFViewCtrl ctrl)
        {
            PDFDouble px = new PDFDouble(x);
            PDFDouble py = new PDFDouble(y);

            ctrl.ConvAnnotationCanvasPtToScreenPt(px, py);

            return new PDFPoint(px.Value, py.Value);
        }

        public static PDFPoint ConvertFromCanvasPointToPagePoint(PDFPoint point, int pageNum, PDFViewCtrl ctrl)
        {
            PDFDouble x = new PDFDouble(point.x);
            PDFDouble y = new PDFDouble(point.y);

            ctrl.ConvAnnotationCanvasPtToScreenPt(x, y);
            ctrl.ConvScreenPtToPagePt(x, y, pageNum);

            return new PDFPoint(x.Value, y.Value);
        }

        public static PDFPoint ConvertFromPagePointToCanvasPoint(PDFPoint point, int pageNum, PDFViewCtrl ctrl)
        {
            PDFDouble x = new PDFDouble(point.x);
            PDFDouble y = new PDFDouble(point.y);

            ctrl.ConvPagePtToScreenPt(x, y, pageNum);
            ctrl.ConvScreenPtToAnnotationCanvasPt(x, y);

            return new PDFPoint(x.Value, y.Value);
        }

        public static PDFRect ConvertFromScreenRectToPageRect(PDFRect rect, int pageNum, PDFViewCtrl ctrl)
        {
            PDFDouble x1 = new PDFDouble(rect.x1);
            PDFDouble y1 = new PDFDouble(rect.y1);
            PDFDouble x2 = new PDFDouble(rect.x2);
            PDFDouble y2 = new PDFDouble(rect.y2);

            ctrl.ConvScreenPtToPagePt(x1, y1, pageNum);
            ctrl.ConvScreenPtToPagePt(x2, y2, pageNum);

            PDFRect retRect = new PDFRect(x1.Value, y1.Value, x2.Value, y2.Value);

            return retRect;
        }

        public static PDFRect ConvertFromPageRectToScreenRect(PDFRect rect, int pageNum, PDFViewCtrl ctrl)
        {
            PDFDouble x1 = new PDFDouble(rect.x1);
            PDFDouble y1 = new PDFDouble(rect.y1);
            PDFDouble x2 = new PDFDouble(rect.x2);
            PDFDouble y2 = new PDFDouble(rect.y2);

            ctrl.ConvPagePtToScreenPt(x1, y1, pageNum);
            ctrl.ConvPagePtToScreenPt(x2, y2, pageNum);

            PDFRect retRect = new PDFRect(x1.Value, y1.Value, x2.Value, y2.Value);

            return retRect;
        }

        public static PDFRect ConvertFromCanvasRectToPageRect(PDFRect rect, int pageNum, PDFViewCtrl ctrl)
        {
            PDFDouble x1 = new PDFDouble(rect.x1);
            PDFDouble y1 = new PDFDouble(rect.y1);
            PDFDouble x2 = new PDFDouble(rect.x2);
            PDFDouble y2 = new PDFDouble(rect.y2);

            ctrl.ConvAnnotationCanvasPtToScreenPt(x1, y1);
            ctrl.ConvAnnotationCanvasPtToScreenPt(x2, y2);

            ctrl.ConvScreenPtToPagePt(x1, y1, pageNum);
            ctrl.ConvScreenPtToPagePt(x2, y2, pageNum);

            PDFRect retRect = new PDFRect(x1.Value, y1.Value, x2.Value, y2.Value);
            retRect.Normalize();

            return retRect;
        }

        public static PDFRect ConvertFromPageRectToCanvasRect(PDFRect rect, int pageNum, PDFViewCtrl ctrl)
        {
            PDFDouble x1 = new PDFDouble(rect.x1);
            PDFDouble y1 = new PDFDouble(rect.y1);
            PDFDouble x2 = new PDFDouble(rect.x2);
            PDFDouble y2 = new PDFDouble(rect.y2);

            ctrl.ConvPagePtToScreenPt(x1, y1, pageNum);
            ctrl.ConvPagePtToScreenPt(x2, y2, pageNum);

            ctrl.ConvScreenPtToAnnotationCanvasPt(x1, y1);
            ctrl.ConvScreenPtToAnnotationCanvasPt(x2, y2);

            PDFRect retRect = new PDFRect(x1.Value, y1.Value, x2.Value, y2.Value);

            return retRect;
        }

        public static PDFRect ConvertFromCanvasRectToScreenRect(PDFRect rect, PDFViewCtrl ctrl)
        {
            PDFDouble x1 = new PDFDouble(rect.x1);
            PDFDouble y1 = new PDFDouble(rect.y1);
            PDFDouble x2 = new PDFDouble(rect.x2);
            PDFDouble y2 = new PDFDouble(rect.y2);

            ctrl.ConvAnnotationCanvasPtToScreenPt(x1, y1);
            ctrl.ConvAnnotationCanvasPtToScreenPt(x2, y2);

            PDFRect retRect = new PDFRect(x1.Value, y1.Value, x2.Value, y2.Value);

            return retRect;
        }

        public static PDFRect ConvertFromScreenRectToCanvasRect(PDFRect rect, PDFViewCtrl ctrl)
        {
            PDFDouble x1 = new PDFDouble(rect.x1);
            PDFDouble y1 = new PDFDouble(rect.y1);
            PDFDouble x2 = new PDFDouble(rect.x2);
            PDFDouble y2 = new PDFDouble(rect.y2);

            ctrl.ConvScreenPtToAnnotationCanvasPt(x1, y1);
            ctrl.ConvScreenPtToAnnotationCanvasPt(x2, y2);

            PDFRect retRect = new PDFRect(x1.Value, y1.Value, x2.Value, y2.Value);

            return retRect;
        }

        #endregion Rectangle Conversions


        #region Color Space

        public struct RGB
        {
            public int R; // 0 - 255
            public int G;
            public int B;

            public RGB(int r, int g, int b)
            {
                R = r;
                G = g;
                B = b;
            }
        }

        public struct HSV
        {
            public double H; // 0 - 1
            public double S; // 0 - 1
            public double V; // 0 - 1

            public HSV(double h, double s, double v)
            {
                H = h;
                S = s;
                V = v;
            }
        }

        public static RGB HSVtoRGB(HSV hsv)
        {
            // HSV contains values scaled as in the color wheel:
            // that is, all from 0 to 255. 

            // for ( this code to work, HSV.Hue needs
            // to be scaled from 0 to 360 (it//s the angle of the selected
            // point within the circle). HSV.Saturation and HSV.value must be 
            // scaled to be between 0 and 1.

            double h;
            double s;
            double v;

            double r = 0;
            double g = 0;
            double b = 0;

            // Scale Hue to be between 0 and 360. Saturation
            // and value scale to be between 0 and 1.
            h = ((double)hsv.H * 360) % 360;
            s = (double)hsv.S;
            v = (double)hsv.V;

            if (s == 0)
            {
                // If s is 0, all colors are the same.
                // This is some flavor of gray.
                r = v;
                g = v;
                b = v;
            }
            else
            {
                double p;
                double q;
                double t;

                double fractionalSector;
                int sectorNumber;
                double sectorPos;

                // The color wheel consists of 6 sectors.
                // Figure out which sector you//re in.
                sectorPos = h / 60;
                sectorNumber = (int)(Math.Floor(sectorPos));

                // get the fractional part of the sector.
                // That is, how many degrees into the sector
                // are you?
                fractionalSector = sectorPos - sectorNumber;

                // Calculate values for the three axes
                // of the color. 
                p = v * (1 - s);
                q = v * (1 - (s * fractionalSector));
                t = v * (1 - (s * (1 - fractionalSector)));

                // Assign the fractional colors to r, g, and b
                // based on the sector the angle is in.
                switch (sectorNumber)
                {
                    case 0:
                        r = v;
                        g = t;
                        b = p;
                        break;

                    case 1:
                        r = q;
                        g = v;
                        b = p;
                        break;

                    case 2:
                        r = p;
                        g = v;
                        b = t;
                        break;

                    case 3:
                        r = p;
                        g = q;
                        b = v;
                        break;

                    case 4:
                        r = t;
                        g = p;
                        b = v;
                        break;

                    case 5:
                        r = v;
                        g = p;
                        b = q;
                        break;
                }
            }
            // return an RGB structure, with values scaled
            // to be between 0 and 255.
            return new RGB((int)(r * 255), (int)(g * 255), (int)(b * 255));
        }

        public static HSV RGBtoHSV(RGB rgb)
        {
            // In this function, R, G, and B values must be scaled 
            // to be between 0 and 1.
            // HSV.Hue will be a value between 0 and 360, and 
            // HSV.Saturation and value are between 0 and 1.
            // The code must scale these to be between 0 and 255 for
            // the purposes of this application.

            double min;
            double max;
            double delta;

            double r = (double)rgb.R / 255;
            double g = (double)rgb.G / 255;
            double b = (double)rgb.B / 255;

            double h;
            double s;
            double v;

            min = Math.Min(Math.Min(r, g), b);
            max = Math.Max(Math.Max(r, g), b);
            v = max;
            delta = max - min;
            if (max == 0 || delta == 0)
            {
                // R, G, and B must be 0, or all the same.
                // In this case, S is 0, and H is undefined.
                // Using H = 0 is as good as any...
                s = 0;
                h = 0;
            }
            else
            {
                s = delta / max;
                if (r == max)
                {
                    // Between Yellow and Magenta
                    h = (g - b) / delta;
                }
                else if (g == max)
                {
                    // Between Cyan and Yellow
                    h = 2 + (b - r) / delta;
                }
                else
                {
                    // Between Magenta and Cyan
                    h = 4 + (r - g) / delta;
                }

            }
            // Scale h to be between 0 and 360. 
            // This may require adding 360, if the value
            // is negative.
            h *= 60;
            if (h < 0)
            {
                h += 360;
            }

            // Scale to the requirements of this 
            // application. All values are between 0 and 1.
            return new HSV((h / 360), s, v);
        }

        #endregion Color Space

        #region Enum Helpers

        /// <summary>
        /// Get annotation description from Annotation Enum Type
        /// </summary>
        /// <param name="annotType"> Annotation Type </param>
        /// <returns> Description in case it exists otherwise "Unknown" </returns>
        public static string GetAnnotDescriptionFromEnum(AnnotType annotType)
        {
            switch (annotType)
            {
                case AnnotType.e_3D:
                    return "3D";
                case AnnotType.e_Caret:
                    return "Caret";
                case AnnotType.e_Circle:
                    return "Circle";
                case AnnotType.e_FileAttachment:
                    return "File Attachment";
                case AnnotType.e_FreeText:
                    return "Free Text";
                case AnnotType.e_Highlight:
                    return "Highlight";
                case AnnotType.e_Ink:
                    return "Ink";
                case AnnotType.e_Line:
                    return "Line";
                case AnnotType.e_Movie:
                    return "Movie";
                case AnnotType.e_Polygon:
                    return "Polygon";
                case AnnotType.e_Polyline:
                    return "Polyline";
                case AnnotType.e_Popup:
                    return "Popup";
                case AnnotType.e_PrinterMark:
                    return "Printer Mark";
                case AnnotType.e_Projection:
                    return "Projection";
                case AnnotType.e_Redact:
                    return "Redact";
                case AnnotType.e_RichMedia:
                    return "Rich Media";
                case AnnotType.e_Screen:
                    return "Screen";
                case AnnotType.e_Sound:
                    return "Sound";
                case AnnotType.e_Square:
                    return "Square";
                case AnnotType.e_Squiggly:
                    return "Squiggly";
                case AnnotType.e_Stamp:
                    return "Stamp";
                case AnnotType.e_StrikeOut:
                    return "Strikeout";
                case AnnotType.e_Text:
                    return "Sticky";
                case AnnotType.e_TrapNet:
                    return "Trap Net";
                case AnnotType.e_Underline:
                    return "Underline";
                case AnnotType.e_Watermark:
                    return "Watermark";
                case AnnotType.e_Widget:
                    return "Widget";
                    
                case AnnotType.e_Unknown:
                default:
                    return "Unknown";
            }
        }
        #endregion
    }
}
