using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.UI.ViewManagement;
using Windows.UI.Xaml;

namespace pdftron.PDF.Tools
{

    public delegate void InputPaneShowingHandler(object sender, InputPaneVisibilityEventArgs e);
    public delegate void InputPaneHidingHandler(InputPane input, InputPaneVisibilityEventArgs e);

    class KeyboardHelper
    {


        private Dictionary<UIElement, InputPaneShowingHandler> handlerMap;
        private UIElement lastFocusedElement = null;
        private InputPaneHidingHandler hidingHandlerDelegate = null;

        private InputPaneHidingHandler expectingHidingHandlerDelegate = null;
        private bool expectedHidingShouldUnsubscribe = false;

        public KeyboardHelper()
        {
            handlerMap = new Dictionary<UIElement, InputPaneShowingHandler>();
        }

        public void SubscribeToKeyboard(bool subscribe)
        {
            InputPane input = InputPane.GetForCurrentView();
            if (subscribe)
            {
                input.Showing += ShowingHandler;
                input.Hiding += HidingHandler;
            }
            else
            {
                input.Showing -= ShowingHandler;
                if (expectingHidingHandlerDelegate == null)
                {
                    input.Hiding -= HidingHandler;
                }
                else
                {
                    expectedHidingShouldUnsubscribe = true;
                }
            }
        }

        public void AddShowingHandler(UIElement element, InputPaneShowingHandler handler)
        {
            if (handlerMap.ContainsKey(element))
            {
                throw new System.Exception("A handler is already registered!");
            }
            else
            {
                handlerMap.Add(element, handler);
                element.GotFocus += GotFocusHandler;
                element.LostFocus += LostFocusHandler;
            }
        }

        // It's important to tell which element got focus in order to decide if
        // the keyboard handler should be called
        private void GotFocusHandler(object sender, RoutedEventArgs e)
        {
            lastFocusedElement = (UIElement)sender;
        }

        private void LostFocusHandler(object sender, RoutedEventArgs e)
        {
            if (lastFocusedElement == (UIElement)sender)
            {
                lastFocusedElement = null;
            }
        }

        private void ShowingHandler(InputPane sender, InputPaneVisibilityEventArgs e)
        {
            System.Diagnostics.Debug.WriteLine("Showing Handler is happening");
            if (lastFocusedElement != null && handlerMap.Count > 0)
            {
                handlerMap[lastFocusedElement](lastFocusedElement, e);
                expectingHidingHandlerDelegate = hidingHandlerDelegate;
            }
            //lastFocusedElement = null;
        }

        private void HidingHandler(InputPane sender, InputPaneVisibilityEventArgs e)
        {
            System.Diagnostics.Debug.WriteLine("Hiding Handler is happening");
            if (expectingHidingHandlerDelegate != null)
            {
                expectingHidingHandlerDelegate(sender, e);
                expectingHidingHandlerDelegate = null;
                if (expectedHidingShouldUnsubscribe)
                {
                    expectedHidingShouldUnsubscribe = false;
                    InputPane input = InputPane.GetForCurrentView();
                    input.Hiding -= HidingHandler;
                }
            }
            else if (hidingHandlerDelegate != null)
            {
                hidingHandlerDelegate(sender, e);
            }
            //lastFocusedElement = null;
        }

        public void SetHidingHandler(InputPaneHidingHandler handler)
        {
            this.hidingHandlerDelegate = handler;
        }

        public void RemoveShowingHandler(UIElement element)
        {
            handlerMap.Remove(element);
            element.GotFocus -= GotFocusHandler;
            element.LostFocus -= LostFocusHandler;
            if (lastFocusedElement != null && lastFocusedElement.Equals(element))
            {
                lastFocusedElement = null;
            }
        }
    }
}
